/********************************************************************************
 * Copyright (c) 2016-2018 ITK Engineering GmbH
 *               2017-2020 in-tech GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
//! @file  roadGeometryInterface.h
//! @brief This file contains a abstract class representing the geometry of a
//!        road, i.e. the layout of the road's reference line in the in the x/y
//!        plane.
//-----------------------------------------------------------------------------

#pragma once

#include "common/vector2d.h"

//-----------------------------------------------------------------------------
//! Abstract class representing the geometry of a road, i.e. the layout of the
//! road's reference line in the in the x/y plane.
//-----------------------------------------------------------------------------
class RoadGeometryInterface
{
public:
  RoadGeometryInterface() = default;
  RoadGeometryInterface(const RoadGeometryInterface&) = delete;
  RoadGeometryInterface(RoadGeometryInterface&&) = delete;
  RoadGeometryInterface& operator=(const RoadGeometryInterface&) = delete;
  RoadGeometryInterface& operator=(RoadGeometryInterface&&) = delete;
  virtual ~RoadGeometryInterface() = default;

  //-----------------------------------------------------------------------------
  //! Calculates the x/y coordinates as vector.
  //!
  //! @param[in]  geometryOffset    offset within geometry section
  //! @param[in]  laneOffset        offset to the lane
  //! @return                coordinates with the x/y coordinates
  //-----------------------------------------------------------------------------
  virtual Common::Vector2d<units::length::meter_t> GetCoord(units::length::meter_t geometryOffset,
                                                            units::length::meter_t laneOffset) const
      = 0;

  //-----------------------------------------------------------------------------
  //! Calculates the heading.
  //!
  //! @param[in]  geometryOffset    offset within geometry section
  //! @return                       heading
  //-----------------------------------------------------------------------------
  virtual units::angle::radian_t GetDir(units::length::meter_t geometryOffset) const = 0;

  //-----------------------------------------------------------------------------
  //! Retrieves the offset within the OpenDRIVE road.
  //!
  //! @return                         offset
  //-----------------------------------------------------------------------------
  virtual units::length::meter_t GetS() const = 0;

  //-----------------------------------------------------------------------------
  //! Retrieves the initial direction of the geometry section
  //!
  //! @return                         initial direction of geometry
  //-----------------------------------------------------------------------------
  virtual units::angle::radian_t GetHdg() const = 0;

  //-----------------------------------------------------------------------------
  //! Retrieves the length of the geometry section
  //!
  //! @return                         length of geometry section
  //-----------------------------------------------------------------------------
  virtual units::length::meter_t GetLength() const = 0;
};
