/********************************************************************************
 * Copyright (c) 2016-2018 ITK Engineering GmbH
 *               2017-2020 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
//! @file  roadInterface.h
//! @brief This file contains a interface representing a road.
//-----------------------------------------------------------------------------

#pragma once

#include <algorithm>
#include <cassert>
#include <cmath>
#include <map>
#include <memory>
#include <vector>

#include "common/globalDefinitions.h"
#include "roadElementTypes.h"
#include "roadElevation.h"
#include "roadGeometryInterface.h"
#include "roadLaneInterface.h"
#include "roadLaneOffset.h"
#include "roadLaneSectionInterface.h"
#include "roadLaneWidth.h"
#include "roadLinkInterface.h"
#include "roadObjectInterface.h"
#include "roadSignalInterface.h"

//-----------------------------------------------------------------------------
//! Struct containing values for parametric cubic polynomial geometry
//-----------------------------------------------------------------------------
struct ParamPoly3Parameters
{
  units::length::meter_t aU;                                               ///< constant factor of the polynomial for u
  units::dimensionless::scalar_t bU;                                       ///< linear factor of the polynomial for u
  units::curvature::inverse_meter_t cU;                                    ///< quadratic factor of the polynomial for u
  units::unit_t<units::inverse<units::squared<units::length::meter>>> dU;  ///< cubic factor of the polynomial for u
  units::length::meter_t aV;                                               ///< constant factor of the polynomial for v
  units::dimensionless::scalar_t bV;                                       ///< linear factor of the polynomial for v
  units::curvature::inverse_meter_t cV;                                    ///< quadratic factor of the polynomial for v
  units::unit_t<units::inverse<units::squared<units::length::meter>>> dV;  ///< cubic factor of the polynomial for v
};

//-----------------------------------------------------------------------------
//! Class representing a road.
//-----------------------------------------------------------------------------
class RoadInterface
{
public:
  RoadInterface() = default;
  RoadInterface(const RoadInterface &) = delete;
  RoadInterface(RoadInterface &&) = delete;
  RoadInterface &operator=(const RoadInterface &) = delete;
  RoadInterface &operator=(RoadInterface &&) = delete;
  virtual ~RoadInterface() = default;

  //-----------------------------------------------------------------------------
  //! Adds a line geometry to a road by creating a new RoadGeometryLine object and
  //! adding it to the stored list of geometries.
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  x                   start position x inertial
  //! @param[in]  y                   start position y inertial
  //! @param[in]  hdg                 start orientation (inertial heading)
  //! @param[in]  length              length of the element's reference line
  //-----------------------------------------------------------------------------
  virtual void AddGeometryLine(units::length::meter_t s,
                               units::length::meter_t x,
                               units::length::meter_t y,
                               units::angle::radian_t hdg,
                               units::length::meter_t length)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds an arc geometry to a road by creating a new RoadGeometryArc object and
  //! adding it to the stored list of geometries.
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  x                   start position x inertial
  //! @param[in]  y                   start position y inertial
  //! @param[in]  hdg                 start orientation (inertial heading)
  //! @param[in]  length              length of the element's reference line
  //! @param[in]  curvature           constant curvature throughout the element
  //-----------------------------------------------------------------------------
  virtual void AddGeometryArc(units::length::meter_t s,
                              units::length::meter_t x,
                              units::length::meter_t y,
                              units::angle::radian_t hdg,
                              units::length::meter_t length,
                              units::curvature::inverse_meter_t curvature)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds a spiral geometry to a road by creating a new RoadGeometrySpiral object and
  //! adding it to the stored list of geometries.
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  x                   start position x inertial
  //! @param[in]  y                   start position y inertial
  //! @param[in]  hdg                 start orientation (inertial heading)
  //! @param[in]  length              length of the element's reference line
  //! @param[in]  curvStart           curvature at the start of the element
  //! @param[in]  curvEnd             curvature at the end of the element
  //-----------------------------------------------------------------------------
  virtual void AddGeometrySpiral(units::length::meter_t s,
                                 units::length::meter_t x,
                                 units::length::meter_t y,
                                 units::angle::radian_t hdg,
                                 units::length::meter_t length,
                                 units::curvature::inverse_meter_t curvStart,
                                 units::curvature::inverse_meter_t curvEnd)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds a cubic polynomial geometry to a road by creating a new RoadGeometryPoly3
  //! object and adding it to the stored list of geometries.
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  x                   start position x inertial
  //! @param[in]  y                   start position y inertial
  //! @param[in]  hdg                 start orientation (inertial heading)
  //! @param[in]  length              length of the element's reference line
  //! @param[in]  a                   constant factor of the polynomial
  //! @param[in]  b                   linear factor of the polynomial
  //! @param[in]  c                   quadratic factor of the polynomial
  //! @param[in]  d                   cubic factor of the polynomial
  //-----------------------------------------------------------------------------
  virtual void AddGeometryPoly3(units::length::meter_t s,
                                units::length::meter_t x,
                                units::length::meter_t y,
                                units::angle::radian_t hdg,
                                units::length::meter_t length,
                                units::length::meter_t a,
                                double b,
                                units::unit_t<units::inverse<units::length::meter>> c,
                                units::unit_t<units::inverse<units::squared<units::length::meter>>> d)
      = 0;
  //-----------------------------------------------------------------------------
  //! Adds a parametric cubic polynomial geometry to a road by creating a new
  //! RoadGeometryparamPoly3 object and adding it to the stored list of geometries.
  //! Each coordinate is calculated in a local (u,v) coordinate system
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  x                   start position x inertial
  //! @param[in]  y                   start position y inertial
  //! @param[in]  hdg                 start orientation (inertial heading)
  //! @param[in]  length              length of the element's reference line
  //! @param[in]  parameters          Factors for the polynomials describing the road
  //-----------------------------------------------------------------------------
  virtual void AddGeometryParamPoly3(units::length::meter_t s,
                                     units::length::meter_t x,
                                     units::length::meter_t y,
                                     units::angle::radian_t hdg,
                                     units::length::meter_t length,
                                     ParamPoly3Parameters parameters)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds an elevation profile defined via a cubic polynomial to a road by creating
  //! a new RoadElevation object and adding it to the stored list of elevations.
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  a                   constant factor of the polynomial
  //! @param[in]  b                   linear factor of the polynomial
  //! @param[in]  c                   quadratic factor of the polynomial
  //! @param[in]  d                   cubic factor of the polynomial
  //-----------------------------------------------------------------------------
  virtual void AddElevation(units::length::meter_t s,
                            units::length::meter_t a,
                            double b,
                            units::unit_t<units::inverse<units::length::meter>> c,
                            units::unit_t<units::inverse<units::squared<units::length::meter>>> d)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds a lane offset defined via a cubic polynomial to a road by creating a new
  //! RoadLaneOffset object and adding it to the stored list of lane offsets.
  //!
  //! @param[in]  s                   start position s-coordinate
  //! @param[in]  a                   constant factor of the polynomial
  //! @param[in]  b                   linear factor of the polynomial
  //! @param[in]  c                   quadratic factor of the polynomial
  //! @param[in]  d                   cubic factor of the polynomial
  //-----------------------------------------------------------------------------
  virtual void AddLaneOffset(units::length::meter_t s,
                             units::length::meter_t a,
                             double b,
                             units::unit_t<units::inverse<units::length::meter>> c,
                             units::unit_t<units::inverse<units::squared<units::length::meter>>> d)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds a new link to a road by creating a new RoadLink object and adding it
  //! to the stored list of links.
  //!
  //! @param[in]  type                type of link
  //! @param[in]  elementType         element type of the link
  //! @param[in]  elementId           ID of the linked element
  //! @param[in]  contactPoint        contact point type of link on the linked element
  //-----------------------------------------------------------------------------
  virtual void AddLink(RoadLinkType type,
                       RoadLinkElementType elementType,
                       const std::string &elementId,
                       ContactPointType contactPoint)
      = 0;

  //-----------------------------------------------------------------------------
  //! Adds a new lane section to a road by creating a new RoadLaneSection object
  //! and adding it to the stored list of lane sections.
  //!
  //! @param[in]  start               start position s-coordinate
  //-----------------------------------------------------------------------------
  virtual void AddRoadLaneSection(units::length::meter_t start) = 0;

  //-----------------------------------------------------------------------------
  //! Adds a new lane section to a road by creating a new RoadLaneSection object
  //! and adding it to the stored list of lane sections.
  //!
  //! @param[in]  signal               road signal
  //-----------------------------------------------------------------------------
  virtual void AddRoadSignal(const RoadSignalSpecification &signal) = 0;

  /**
   * Adds a new road from road specifications
   *
   * @param object road object specification
   */
  virtual void AddRoadObject(const RoadObjectSpecification &object) = 0;

  //-----------------------------------------------------------------------------
  //! Returns the ID of the road.
  //!
  //! @return                         ID of the road
  //-----------------------------------------------------------------------------
  virtual const std::string GetId() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the length of the road.
  //!
  //! @return                         length of the road
  //-----------------------------------------------------------------------------
  virtual double GetLength() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list of elevation profiles.
  //!
  //! @return                         list of elevation profiles
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadElevation>> &GetElevations() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list of lane offsets.
  //!
  //! @return                         list of lane offsets
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadLaneOffset>> &GetLaneOffsets() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list of road geometries.
  //!
  //! @return                         list of road geometries
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadGeometryInterface>> &GetGeometries() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list of road links.
  //!
  //! @return                         list of road links
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadLinkInterface>> &GetRoadLinks() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list of lane sections.
  //!
  //! @return                         list of lane sections
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadLaneSectionInterface>> &GetLaneSections() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list signals
  //!
  //! @return                         list of signals
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadSignalInterface>> &GetRoadSignals() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns the stored list roads
  //!
  //! @return                         list of roads
  //-----------------------------------------------------------------------------
  virtual const std::vector<std::unique_ptr<RoadObjectInterface>> &GetRoadObjects() const = 0;

  //-----------------------------------------------------------------------------
  //! Sets the flag, if the road is in the reference direction or not.
  //!
  //! @param[in]  inDirection         Flag, if the road is in the reference
  //!                                 direction or not
  //-----------------------------------------------------------------------------
  virtual void SetInDirection(bool inDirection) = 0;

  //-----------------------------------------------------------------------------
  //! Returns if the road is in the reference direction or not.
  //!
  //! @return                         Flag, if the road is in the reference
  //!                                 direction or not
  //-----------------------------------------------------------------------------
  virtual bool GetInDirection() const = 0;

  /**
   * @brief Add the type of road from road type specifications
   *
   * @param info road type specification
   */
  virtual void AddRoadType(const RoadTypeSpecification &info) = 0;

  /**
   * @brief Get the Road Type object
   *
   * @param start  start position of the road
   * @return information on road type
   */
  virtual RoadTypeInformation GetRoadType(units::length::meter_t start) const = 0;

  /**
   * @brief Set the Junction Id object
   *
   * @param junctionId id of the junction
   */
  virtual void SetJunctionId(const std::string &junctionId) = 0;

  /**
   * @brief Get the Junction Id object
   *
   * @return junction id
   */
  virtual const std::string GetJunctionId() const = 0;
};
