/********************************************************************************
 * Copyright (c) 2017 ITK Engineering GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
//! @file  roadLaneWidth.h
//! @brief This file contains a class representing the polynomial defining the
//!        lane width.
//!        The polynomial has the form width = a + b*ds + c*ds^2 + d*ds^3.
//-----------------------------------------------------------------------------

#ifndef ROADLANEWIDTH
#define ROADLANEWIDTH

//-----------------------------------------------------------------------------
//! Class representing the polynomial defining the lane width. The polynomial has
//! the form width = a + b*ds + c*ds^2 + d*ds^3.
//-----------------------------------------------------------------------------
class RoadLaneWidth
{
public:
  /**
   * Generate Road lane width object
   * @param sOffset   Offset relative to the preceding lane section
   * @param a         Constant factor from the polynomial
   * @param b         Linear factor from the polynomial
   * @param c         Quadratic factor from the polynomial
   * @param d         Cubic factor from the polynomial
   */
  RoadLaneWidth(units::length::meter_t sOffset,
                units::length::meter_t a,
                double b,
                units::unit_t<units::inverse<units::length::meter>> c,
                units::unit_t<units::inverse<units::squared<units::length::meter>>> d)
      : sOffset(sOffset), a(a), b(b), c(c), d(d)
  {
  }
  RoadLaneWidth(const RoadLaneWidth&) = delete;
  RoadLaneWidth(RoadLaneWidth&&) = delete;
  RoadLaneWidth& operator=(const RoadLaneWidth&) = delete;
  RoadLaneWidth& operator=(RoadLaneWidth&&) = delete;
  virtual ~RoadLaneWidth() = default;

  //-----------------------------------------------------------------------------
  //! Returns the starting position relative to the position of the preceding
  //! lane section.
  //!
  //! @return                         Offset relative to the preceding lane section
  //-----------------------------------------------------------------------------
  units::length::meter_t GetSOffset() const { return sOffset; }

  //-----------------------------------------------------------------------------
  //! Returns the constant factor from the polynomial.
  //!
  //! @return                         Constant factor from the polynomial
  //-----------------------------------------------------------------------------
  units::length::meter_t GetA() const { return a; }

  //-----------------------------------------------------------------------------
  //! Returns the linear factor from the polynomial.
  //!
  //! @return                         Linear factor from the polynomial
  //-----------------------------------------------------------------------------
  double GetB() const { return b; }

  //-----------------------------------------------------------------------------
  //! Returns the quadratic factor from the polynomial.
  //!
  //! @return                         Quadratic factor from the polynomial
  //-----------------------------------------------------------------------------
  units::unit_t<units::inverse<units::length::meter>> GetC() const { return c; }

  //-----------------------------------------------------------------------------
  //! Returns the cubic factor from the polynomial.
  //!
  //! @return                         Cubic factor from the polynomial
  //-----------------------------------------------------------------------------
  units::unit_t<units::inverse<units::squared<units::length::meter>>> GetD() const { return d; }

private:
  units::length::meter_t sOffset;
  units::length::meter_t a;
  double b;
  units::unit_t<units::inverse<units::length::meter>> c;
  units::unit_t<units::inverse<units::squared<units::length::meter>>> d;
};

#endif  // ROADLANEWIDTH
