/********************************************************************************
 * Copyright (c) 2019-2020 in-tech GmbH
 *               2022-2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/
#pragma once

#include <string>

#include "roadElementTypes.h"

//! Represents a RoadSignal in OpenDRIVE
class RoadSignalInterface
{
public:
  RoadSignalInterface() = default;
  RoadSignalInterface(const RoadSignalInterface&) = delete;
  RoadSignalInterface(RoadSignalInterface&&) = delete;
  RoadSignalInterface& operator=(const RoadSignalInterface&) = delete;
  RoadSignalInterface& operator=(RoadSignalInterface&&) = delete;
  virtual ~RoadSignalInterface() = default;

  //-----------------------------------------------------------------------------
  /// @brief Returns the unique identification string of the signal
  /// @return id
  //-----------------------------------------------------------------------------
  virtual std::string GetId() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the s coordinate of the signal
  ///
  /// @return s [m]
  //-----------------------------------------------------------------------------
  virtual units::length::meter_t GetS() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the t coordinate of the signal
  ///
  /// @return s [m]
  //-----------------------------------------------------------------------------
  virtual units::length::meter_t GetT() const = 0;

  //-----------------------------------------------------------------------------
  //! Returns if the signal is valid for a given lane
  //! @param[in] laneId   lane Id of interest
  //!
  //! @return             true if signal is valid
  //-----------------------------------------------------------------------------
  virtual bool IsValidForLane(int64_t laneId) const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the country of the signal
  /// @return country
  //-----------------------------------------------------------------------------
  virtual std::string GetCountry() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the revision specification of the country signal
  /// @return country_revision
  //-----------------------------------------------------------------------------
  virtual std::string GetCountryRevision() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the type of the signal (e.g. code according to StVO)
  /// @return type
  //-----------------------------------------------------------------------------
  virtual std::string GetType() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the subtype of the signal (e.g. code according to StVO)
  /// @return type
  //-----------------------------------------------------------------------------
  virtual std::string GetSubType() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the value
  /// @return value
  //-----------------------------------------------------------------------------
  virtual std::optional<double> GetValue() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the unit
  /// @return unit
  //-----------------------------------------------------------------------------
  virtual RoadSignalUnit GetUnit() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the text on the signal
  /// @return text
  //-----------------------------------------------------------------------------
  virtual std::string GetText() const = 0;

  /**
   * @brief Get the Dependencies object
   *
   * @return list of dependencies
   */
  virtual std::vector<std::string> GetDependencies() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns true if the signal is dynamic
  ///
  /// @return true if dynamic == "yes"
  //-----------------------------------------------------------------------------
  virtual bool GetIsDynamic() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the width of the signal
  ///
  /// @return height [m]
  //-----------------------------------------------------------------------------
  virtual units::length::meter_t GetWidth() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the height of the signal
  ///
  /// @return height [m]
  //-----------------------------------------------------------------------------
  virtual units::length::meter_t GetHeight() const = 0;

  //-----------------------------------------------------------------------------
  /// @brief Returns the pitch of the signal
  ///
  /// @return height [m]
  //-----------------------------------------------------------------------------
  virtual units::angle::radian_t GetPitch() const = 0;
  //-----------------------------------------------------------------------------
  /// @brief Returns the roll of the signal
  ///
  /// @return height [m]
  //-----------------------------------------------------------------------------
  virtual units::angle::radian_t GetRoll() const = 0;

  /**
   * @brief Get offset in z direction
   *
   * @return offset in z direction
   */
  virtual units::length::meter_t GetZOffset() const = 0;

  /**
   * @brief Get the Orientation of road signal
   *
   * @return
   */
  virtual bool GetOrientation() const = 0;

  /**
   * @brief Get H Offset
   *
   * @return
   */
  virtual units::angle::radian_t GetHOffset() const = 0;
};
