/********************************************************************************
 * Copyright (c) 2020 in-tech GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
//! @brief This file contains all functions for class SpeedActionSignal
//!
//! This class contains all functionality of the module.
//-----------------------------------------------------------------------------

#pragma once

#include "include/modelInterface.h"

//-----------------------------------------------------------------------------
//! Signal class
//-----------------------------------------------------------------------------
class SpeedActionSignal : public ComponentStateSignalInterface
{
public:
  /// constructor
  static constexpr const char* COMPONENTNAME{"SpeedActionSignal"};

  //-----------------------------------------------------------------------------
  //! Constructor
  //-----------------------------------------------------------------------------
  SpeedActionSignal() { componentState = ComponentState::Disabled; }

  /**
   * @brief Construct a new Speed Action Signal object
   *
   * @param other another speed action signal
   */
  SpeedActionSignal(SpeedActionSignal& other)
      : SpeedActionSignal(other.componentState, other.targetSpeed, other.acceleration)
  {
  }

  /**
   * @brief Construct a new Speed Action Signal object
   *
   * @param componentState State of the component
   * @param targetSpeed    Desired speed
   * @param acceleration   Acceleration of the vehicle
   */
  SpeedActionSignal(ComponentState componentState,
                    const units::velocity::meters_per_second_t targetSpeed,
                    const units::acceleration::meters_per_second_squared_t acceleration)
      : ComponentStateSignalInterface(componentState), targetSpeed(targetSpeed), acceleration(acceleration)
  {
  }

  ~SpeedActionSignal() override = default;

  //-----------------------------------------------------------------------------
  //! Returns the content/payload of the signal as an std::string
  //!
  //! @return     Content/payload of the signal as an std::string
  //-----------------------------------------------------------------------------
  virtual operator std::string() const override
  {
    std::ostringstream stream;
    stream << COMPONENTNAME << std::endl;
    return stream.str();
  }

  /// target speed
  units::velocity::meters_per_second_t targetSpeed{0.0};

  /// acceleration Range [0..inf[
  units::acceleration::meters_per_second_squared_t acceleration{0.0};
};
