/********************************************************************************
 * Copyright (c) 2020-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * https://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include "AlgorithmSspWrapper.h"

#include <utility>

#include "AlgorithmSspWrapperImplementation.h"

/// version of this current module
const std::string Version = "0.1.0";
static const CallbackInterface *Callbacks = nullptr;  // NOLINT[cppcoreguidelines-avoid-non-const-global-variables]

//-----------------------------------------------------------------------------
//! dll-function to obtain the version of the current module
//!
//! @return                       Version of the current module
//-----------------------------------------------------------------------------
const std::string &OpenPASS_GetVersion()
{
  return Version;
}

/// @brief Create an instance of lateral algorithm
/// @param componentName      Name of the component
/// @param isInit             If instance initialzed
/// @param priority           Priority of the instance
/// @param offsetTime         Offset time
/// @param responseTime       Response time
/// @param cycleTime          Cycle time
/// @param stochastics        Reference to the stochastics interface
/// @param world              Reference to the world interface
/// @param parameters         Reference to the parameter interface
/// @param publisher          Reference to the publisher interface
/// @param agent              Reference to the agent interface
/// @param callbacks          Reference to the callback interface
/// @param scenarioControl    scenarioControl of entity
/// @return
ModelInterface *OpenPASS_CreateInstance(std::string componentName,  // NOLINT(performance-unnecessary-value-param)
                                        bool isInit,
                                        int priority,
                                        int offsetTime,
                                        int responseTime,
                                        int cycleTime,
                                        StochasticsInterface *stochastics,
                                        WorldInterface *world,
                                        const ParameterInterface *parameters,
                                        PublisherInterface *const publisher,
                                        AgentInterface *agent,
                                        const CallbackInterface *callbacks,
                                        std::shared_ptr<ScenarioControlInterface> scenarioControl)
{
  SspLogger::SetLogger(callbacks, agent->GetId(), componentName);
  Callbacks = callbacks;
  try
  {
    return new (std::nothrow) AlgorithmSspWrapperImplementation(componentName,
                                                                isInit,
                                                                priority,
                                                                offsetTime,
                                                                responseTime,
                                                                cycleTime,
                                                                stochastics,
                                                                world,
                                                                parameters,
                                                                publisher,
                                                                agent,
                                                                callbacks,
                                                                std::move(scenarioControl));
  }
  catch (const std::exception &ex)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, ex.what());
    }
    return nullptr;
  }
  catch (...)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, "unexpected exception");
    }

    return nullptr;
  }
}
//-----------------------------------------------------------------------------
//! dll-function to destroy/delete an instance of the module.
//!
//! @param[in]     implementation    The instance that should be freed
//-----------------------------------------------------------------------------
void OpenPASS_DestroyInstance(ModelInterface *implementation)
{
  delete implementation;  // NOLINT(cppcoreguidelines-owning-memory)
}

bool OpenPASS_UpdateInput(ModelInterface *implementation,
                          int localLinkId,
                          const std::shared_ptr<SignalInterface const> &data,
                          int time)
{
  try
  {
    implementation->UpdateInput(localLinkId, data, time);
  }
  catch (const std::exception &ex)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, ex.what());
    }

    return false;
  }
  catch (...)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, "unexpected exception");
    }

    return false;
  }

  return true;
}

bool OpenPASS_UpdateOutput(ModelInterface *implementation,
                           int localLinkId,
                           std::shared_ptr<SignalInterface const> &data,
                           int time)
{
  try
  {
    implementation->UpdateOutput(localLinkId, data, time);
  }
  catch (const std::exception &ex)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, ex.what());
    }

    return false;
  }
  catch (...)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, "unexpected exception");
    }

    return false;
  }

  return true;
}

bool OpenPASS_Trigger(ModelInterface *implementation, int time)
{
  try
  {
    implementation->Trigger(time);
  }
  catch (const std::exception &ex)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, ex.what());
    }

    return false;
  }
  catch (...)
  {
    if (Callbacks != nullptr)
    {
      Callbacks->Log(CbkLogLevel::Error, __FILE__, __LINE__, "unexpected exception");
    }

    return false;
  }

  return true;
}
