/********************************************************************************
 * Copyright (c) 2019 in-tech GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

/** \addtogroup ComponentController
 * @{
 * \brief This module
 *
 * @} */

#pragma once

#include <map>
#include <memory>
#include <stdexcept>
#include <string>
#include <utility>

#include "common/compCtrlToAgentCompSignal.h"
#include "components/ComponentController/src/condition.h"
#include "condition.h"
#include "include/callbackInterface.h"
#include "include/modelInterface.h"
#include "stateManager.h"

class AgentInterface;
class ParameterInterface;
class PublisherInterface;
class ScenarioControlInterface;
class SignalInterface;
class StochasticsInterface;
class WorldInterface;

using namespace ComponentControl;

/// Class representing implementation of component controller
class ComponentControllerImplementation : public UnrestrictedControllStrategyModelInterface
{
public:
  /// Name of this component
  const std::string COMPONENTNAME = "ComponentController";

  /**
   * @brief Construct a new Component Controller Implementation object
   *
   * @param[in]     componentName     Name of the component
   * @param[in]     isInit            Corresponds to "init" of "Component"
   * @param[in]     priority          Corresponds to "priority" of "Component"
   * @param[in]     offsetTime        Corresponds to "offsetTime" of "Component"
   * @param[in]     responseTime      Corresponds to "responseTime" of "Component"
   * @param[in]     cycleTime         Corresponds to "cycleTime" of "Component"
   * @param[in]     stochastics       Pointer to the stochastics class loaded by the framework
   * @param[in]     world             Pointer to the world
   * @param[in]     parameters        Pointer to the parameters of the module
   * @param[in]     publisher         Pointer to the publisher instance
   * @param[in]     callbacks         Pointer to the callbacks
   * @param[in]     agent             Pointer to agent instance
   * @param[in]     scenarioControl   scenarioControl of entity
   */
  ComponentControllerImplementation(std::string componentName,
                                    bool isInit,
                                    int priority,
                                    int offsetTime,
                                    int responseTime,
                                    int cycleTime,
                                    StochasticsInterface *stochastics,
                                    WorldInterface *world,
                                    const ParameterInterface *parameters,
                                    PublisherInterface *const publisher,
                                    const CallbackInterface *callbacks,
                                    AgentInterface *agent,
                                    std::shared_ptr<ScenarioControlInterface> scenarioControl);
  ComponentControllerImplementation(const ComponentControllerImplementation &) = delete;
  ComponentControllerImplementation(ComponentControllerImplementation &&) = delete;
  ComponentControllerImplementation &operator=(const ComponentControllerImplementation &) = delete;
  ComponentControllerImplementation &operator=(ComponentControllerImplementation &&) = delete;
  virtual ~ComponentControllerImplementation() = default;

  /*!
   * \brief Update Inputs
   *
   * Function is called by framework when another component delivers a signal over
   * a channel to this component (scheduler calls update taks of other component).
   *
   * Refer to module description for input channels and input ids.
   *
   * @param[in]     localLinkId    Corresponds to "id" of "ComponentInput"
   * @param[in]     data           Referenced signal (copied by sending component)
   * @param[in]     time           Current scheduling time
   */
  void UpdateInput(int localLinkId, const std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Update outputs.
   *
   * Function is called by framework when this component has to deliver a signal over
   * a channel to another component (scheduler calls update task of this component).
   *
   * Refer to module description for output channels and output ids.
   *
   * @param[in]     localLinkId    Corresponds to "id" of "ComponentOutput"
   * @param[out]    data           Referenced signal (copied by this component)
   * @param[in]     time           Current scheduling time
   */
  void UpdateOutput(int localLinkId, std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Process data within component.
   *
   * Function is called by framework when the scheduler calls the trigger task
   * of this component
   *
   * @param[in]     time           Current scheduling time
   */
  void Trigger(int time) override;

private:
  template <typename T, typename Signal>
  const std::shared_ptr<T const> SignalCast(Signal &&signal, int linkId)
  {
    if (const auto castedSignal = std::dynamic_pointer_cast<T const>(std::forward<Signal>(signal)))
    {
      return castedSignal;
    }

    const std::string msg = COMPONENTNAME + " invalid signaltype on link " + std::to_string(linkId);
    LOG(CbkLogLevel::Debug, msg);
    throw std::runtime_error(msg);
  }

  std::map<int, std::string> driverInputChannels;
  std::map<int, std::string> driverOutputChannels;

protected:
  /// driver warinings
  Warnings driverWarnings;

  /// state manager
  StateManager stateManager;
};
