/********************************************************************************
 * Copyright (c) 2020-2021 ITK Engineering GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include "dynamics_collision_implementation.h"

#include <MantleAPI/Common/orientation.h>
#include <ostream>
#include <stdexcept>
#include <utility>
#include <vector>

#include "common/globalDefinitions.h"
#include "common/vector2d.h"
#include "include/agentInterface.h"
#include "include/callbackInterface.h"

class ParameterInterface;
class PublisherInterface;
class SignalInterface;
class StochasticsInterface;
class WorldInterface;

Dynamics_Collision_Implementation::Dynamics_Collision_Implementation(std::string componentName,
                                                                     bool isInit,
                                                                     int priority,
                                                                     int offsetTime,
                                                                     int responseTime,
                                                                     int cycleTime,
                                                                     StochasticsInterface *stochastics,
                                                                     WorldInterface *world,
                                                                     const ParameterInterface *parameters,
                                                                     PublisherInterface *const publisher,
                                                                     const CallbackInterface *callbacks,
                                                                     AgentInterface *agent)
    : DynamicsInterface(std::move(componentName),
                        isInit,
                        priority,
                        offsetTime,
                        responseTime,
                        cycleTime,
                        stochastics,
                        world,
                        parameters,
                        publisher,
                        callbacks,
                        agent)
{
  LOGINFO("Constructing Dynamics_Collision");
  cycleTimeSec = units::time::millisecond_t(cycleTime);
  LOGINFO("Constructing Dynamics_Collision successful");
}

void Dynamics_Collision_Implementation::UpdateInput(int localLinkId,
                                                    const std::shared_ptr<SignalInterface const> &data,
                                                    [[maybe_unused]] int time)
{
  std::stringstream log;
  log << COMPONENTNAME << " UpdateInput";
  LOG(CbkLogLevel::Debug, log.str());
  log.str(std::string());

  bool success = inputPorts.at(localLinkId)->SetSignalValue(data);

  if (success)
  {
    log << COMPONENTNAME << " UpdateInput successful";
    LOG(CbkLogLevel::Debug, log.str());
  }
  else
  {
    log << COMPONENTNAME << " UpdateInput failed";
    LOG(CbkLogLevel::Error, log.str());
  }
}

void Dynamics_Collision_Implementation::UpdateOutput([[maybe_unused]] int localLinkId,
                                                     [[maybe_unused]] std::shared_ptr<SignalInterface const> &data,
                                                     [[maybe_unused]] int time)
{
  throw std::runtime_error("invalid link");
}

void Dynamics_Collision_Implementation::Trigger([[maybe_unused]] int time)
{
  CollisionState nextState = CollisionState::NOCOLLISION;
  bool in_collisionOccured = collisionOccured.GetValue();
  switch (collisionState)
  {
    case CollisionState::NOCOLLISION:
      if (in_collisionOccured)
      {
        nextState = CollisionState::COLLISIONIMPULS;
      }
      else
      {
        nextState = CollisionState::NOCOLLISION;
      }
      break;
    case CollisionState::COLLISIONIMPULS:
    case CollisionState::COLLISION:
      if (in_collisionOccured)
      {
        nextState = CollisionState::COLLISION;
      }
      else
      {
        nextState = CollisionState::NOCOLLISION;
      }
      break;
  }

  collisionState = nextState;  // immediate transition

  if (collisionState == CollisionState::COLLISIONIMPULS)
  {
    auto collisionPartners = GetAgent()->GetCollisionPartners();

    if (collisionPartners.size() > 0)
    {
      PostCrashVelocity postCrashVel = GetAgent()->GetPostCrashVelocity();

      GetAgent()->SetPositionX(GetAgent()->GetPositionX() + GetAgent()->GetVelocity().x * cycleTimeSec);  // global CS
      GetAgent()->SetPositionY(GetAgent()->GetPositionY() + GetAgent()->GetVelocity().y * cycleTimeSec);  // global CS
      GetAgent()->SetYaw(GetAgent()->GetYaw() + GetAgent()->GetYawRate() * cycleTimeSec);

      Common::Vector2d<units::velocity::meters_per_second_t> velocityImpact(postCrashVel.velocityAbsolute, 0.0_mps);
      velocityImpact.Rotate(postCrashVel.velocityDirection);
      GetAgent()->SetVelocityVector({velocityImpact.x, velocityImpact.y, 0.0_mps});
      GetAgent()->SetYawRate(postCrashVel.yawVelocity);
      LOGDEBUG("Override Velocity by Dynamics_CollisionPCM for agent " + std::to_string(GetAgent()->GetId()) + ": "
               + std::to_string(velocityImpact.x.value()) + ", " + std::to_string(velocityImpact.y.value()) + ", "
               + std::to_string(postCrashVel.yawVelocity.value()));
    }
    Common::Vector2d accelerationImpact(0.0, 0.0);
    units::angular_acceleration::radians_per_second_squared_t yawAccel{0.0};
    GetAgent()->SetTangentialAcceleration(units::acceleration::meters_per_second_squared_t(accelerationImpact.x));
    GetAgent()->SetCentripetalAcceleration(units::acceleration::meters_per_second_squared_t(accelerationImpact.y));
    GetAgent()->SetYawAcceleration(yawAccel);

    // for logging purposes: only longitudinal acceleration in vehicle CS
    GetAgent()->SetAcceleration(units::acceleration::meters_per_second_squared_t(accelerationImpact.x));

    // Application of acceleration instead of setting velocities is not possible:
    // The acceleration values are far too high for a single simulation step (200g and above)

    LOGDEBUG("Override Acceleration by Dynamics_CollisionPCM for agent " + std::to_string(GetAgent()->GetId()) + ": "
             + std::to_string(accelerationImpact.x) + ", " + std::to_string(accelerationImpact.y) + ", "
             + std::to_string(yawAccel.value()));
  }
}
