/********************************************************************************
 * Copyright (c) 2018-2019 AMFD GmbH
 *               2020 HLRS, University of Stuttgart
 *               2016-2017 ITK Engineering GmbH
 *               2017-2019 in-tech GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/
#pragma once

#include <MantleAPI/Traffic/entity_properties.h>
#include <math.h>
#include <memory>
#include <stdexcept>
#include <string>
#include <units.h>

#include "common/dynamicsSignal.h"
#include "include/callbackInterface.h"
#include "include/modelInterface.h"
#include "include/signalInterface.h"

class AgentInterface;
class ParameterInterface;
class PublisherInterface;
class StochasticsInterface;
class WorldInterface;

/** \addtogroup Dynamics_RegularDriving
* @{
* \brief models the longitudinal and lateral dynamics of the vehicle corresponding to the pedal and steering inputs
*
* This component models the reaction of the vehicle based on the pedal and steering wheel operation of the driver.
* It calculates velocity and position of the agent according to his acceleration and his collision state.
* If no collision took place in the previous time step where the agent was
* the rear opponent, the velocity and position of the agent is calculated
* according to his acceleration wish.
* If the agent was the rear opponent in a collision, its velocity is set to
* the one of the front agent. The position is calculated according to this
* velocity. Additionaly the maximum change of steering wheel angle according to the
* desired steering wheel angle is calculated.
*
* \section Dynamics_RegularDriving_Inputs Inputs
* Input variables:
* name | meaning
* -----|------
* in_accPedalPos        | Position of the gas pedal in percent
* in_brakePedalPos      | Position of the brake pedal in percent
* in_gear               | Current gear
* in_steeringWheelAngle | Steering wheel angle of the agent
*
*
* Input channel IDs:
* Input ID  | signal class                              | contained variables
* ----------|-------------------------------------------|-------------
* 1         | LongitudinalSignal                        | in_accPedalPos, in_brakePedalPos, in_gear
* 2         | SteeringSignal                            | in_steeringWheelAngle
*
*
** \section Dynamics_RegularDriving_InitInputs Init Inputs
* Init input variables:
* name          | meaning
* --------------|------
* vehicleModelParameters | mantle_api::VehicleProperties
*
*
* * Init input channel IDs:
* Input Id      | signal class                | contained variables
* --------------|-----------------------------|------------
* 100           | ParametersVehicleSignal     | vehicleModelParameters
*
*
*
* * \section Dynamics_RegularDriving_Outputs Outputs
* Output variables:
* name                       | meaning
* ---------------------------|------
* acceleration               | Acceleration of the current agent corresponding to the pedalpositions [m/s²]
* velocity                   | The driven curvature of the car corresponding to the steering wheel angle based on
Ackermann [radian]
* positionX                  | new inertial x-position [m]
* positionY                  | new inertial y-position [m]
* yaw                        | new yaw angle in current time step [radian]
* yawRate                    | change of yaw angle due to ds and curvature [radian]
* steeringWheelAngle         | new angle of the steering wheel angle [radian]
* travelDistance             | distance traveled by the agent during this timestep [m]
*
*
* Output channel IDs:
* Output Id                  | signal class    | contained variables
* ---------------------------|-----------------|-------------
* 0                          | DynamicsSignal  | acceleration, velocity, positionX, positionY, yaw, yawRate,
steeringWheelAngle, travelDistance
*

* \section Dynamics_RegularDriving_ExternalParameters External parameters
* none
*
* \section Dynamics_RegularDriving_InternalParameters Internal paramters
* name                  | value  | meaning
* ----------------------|--------|-------------
* VLowerLimit           |      0 | The minimal velocity of the agent in m/s.
*
* @} */

/// \ingroup Dynamics_RegularDriving
/// Class representing implementation of regular drving dynamics
class DynamicsRegularDrivingImplementation : public DynamicsInterface
{
public:
  //! Name of the current component
  const std::string COMPONENTNAME = "DynamicRegularDriving";

  /**
   * @brief Construct a new Dynamics Regular Driving Implementation object
   *
   * @param[in]     componentName  Name of the component
   * @param[in]     isInit         Corresponds to "init" of "Component"
   * @param[in]     priority       Corresponds to "priority" of "Component"
   * @param[in]     offsetTime     Corresponds to "offsetTime" of "Component"
   * @param[in]     responseTime   Corresponds to "responseTime" of "Component"
   * @param[in]     cycleTime      Corresponds to "cycleTime" of "Component"
   * @param[in]     stochastics    Pointer to the stochastics class loaded by the framework
   * @param[in]     world          Pointer to the world
   * @param[in]     parameters     Pointer to the parameters of the module
   * @param[in]     publisher      Pointer to the publisher instance
   * @param[in]     callbacks      Pointer to the callbacks
   * @param[in]     agent          Pointer to agent instance
   */
  DynamicsRegularDrivingImplementation(std::string componentName,
                                       bool isInit,
                                       int priority,
                                       int offsetTime,
                                       int responseTime,
                                       int cycleTime,
                                       StochasticsInterface *stochastics,
                                       WorldInterface *world,
                                       const ParameterInterface *parameters,
                                       PublisherInterface *const publisher,
                                       const CallbackInterface *callbacks,
                                       AgentInterface *agent)
      : DynamicsInterface(componentName,
                          isInit,
                          priority,
                          offsetTime,
                          responseTime,
                          cycleTime,
                          stochastics,
                          world,
                          parameters,
                          publisher,
                          callbacks,
                          agent),
        dynamicsSignal{ComponentState::Acting}
  {
    if (GetPublisher() == nullptr)
    {
      std::string msg = "DynamicsRegularDriving requires a publisher";
      LOG(CbkLogLevel::Error, msg);
      throw std::runtime_error(msg);
    }
  }
  DynamicsRegularDrivingImplementation(const DynamicsRegularDrivingImplementation &) = delete;
  DynamicsRegularDrivingImplementation(DynamicsRegularDrivingImplementation &&) = delete;
  DynamicsRegularDrivingImplementation &operator=(const DynamicsRegularDrivingImplementation &) = delete;
  DynamicsRegularDrivingImplementation &operator=(DynamicsRegularDrivingImplementation &&) = delete;
  virtual ~DynamicsRegularDrivingImplementation() = default;

  /*!
   * \brief Update Inputs
   *
   * Function is called by framework when another component delivers a signal over
   * a channel to this component (scheduler calls update taks of other component).
   *
   * Refer to module description for input channels and input ids.
   *
   * @param[in]     localLinkId    Corresponds to "id" of "ComponentInput"
   * @param[in]     data           Referenced signal (copied by sending component)
   * @param[in]     time           Current scheduling time
   */
  void UpdateInput(int localLinkId, const std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Update outputs.
   *
   * Function is called by framework when this component has to deliver a signal over
   * a channel to another component (scheduler calls update task of this component).
   *
   * Refer to module description for output channels and output ids.
   *
   * @param[in]     localLinkId    Corresponds to "id" of "ComponentOutput"
   * @param[out]    data           Referenced signal (copied by this component)
   * @param[in]     time           Current scheduling time
   */
  void UpdateOutput(int localLinkId, std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Process data within component.
   *
   * Function is called by framework when the scheduler calls the trigger task
   * of this component
   *
   * @param[in]     time           Current scheduling time
   */
  void Trigger(int time) override;

private:
  //! Applies limit to incoming geer
  void ApplyGearLimit();

  //! Applies limits to the incoming steering wheel angle and incoming pedal position
  void ApplyPedalPositionLimits();

  //! Get the acceleration of the vehicle from pedal position and gear.
  //! @param [in] accPedalPos         current Gas pedal position [%]
  //! @param [in] brakePedalPos       current Brake pedal position [%]
  //! @param [in] gear                current Gear
  //! @return current acceleration
  units::acceleration::meters_per_second_squared_t GetAccVehicle(double accPedalPos, double brakePedalPos, int gear);

  //! Get the acceleration (negative) caused by the air resistance of the vehicle.
  //! @param [in] velocity        absolute vehicle speed [m/s]
  //! @return acceleration due to rolling resistance [m/s^2]
  //-----------------------------------------------------------------------------
  units::acceleration::meters_per_second_squared_t GetAccelerationFromAirResistance(
      units::velocity::meters_per_second_t velocity);

  //! Get the acceleration (negative) caused by the rolling resistance of the wheels.
  //! @return acceleration due to rolling resistance [m/s^2]
  //-----------------------------------------------------------------------------
  units::acceleration::meters_per_second_squared_t GetAccelerationFromRollingResistance();

  //! Get the moment of the engine from pedal position and gear.
  //! @param [in] gasPedalPos         current Gas pedal position. [%]
  //! @param [in] lastGear            the last chosen gear.
  //! @param [in] carParameters       parameters of the car
  //! @return engine moment [Nm]
  //-----------------------------------------------------------------------------
  units::torque::newton_meter_t GetEngineMoment(double gasPedalPos, int lastGear);

  //! Get the drag moment of the engine corresponding to its speed
  //! @param [in] engineSpeed         the speed of the engine
  //! @param [in] carParameters       parameters of the car
  //! @return drag moment [Nm]
  units::torque::newton_meter_t GetEngineMomentMin(units::angular_velocity::revolutions_per_minute_t engineSpeed);

  //! Get the maximum possible torque of the engine in the current state
  //! @param [in] engineSpeed         the speed of the engine
  //! @param [in] carParameters       parameters of the car
  //! @return maximum moment in the current state [Nm]
  units::torque::newton_meter_t GetEngineMomentMax(units::angular_velocity::revolutions_per_minute_t engineSpeed);

  //! Calculate the engine speed corresponding to the current speed and gear of the car
  //! @param [in] xVel               longitudinal speed of the car
  //! @param [in] gear               current gear
  //! @param [in] carParameters      parameters of the car
  //! @return engine speed [1/min]
  units::angular_velocity::revolutions_per_minute_t GetEngineSpeedByVelocity(units::velocity::meters_per_second_t xVel,
                                                                             int gear);

  //! Calculate the resulting acceleration of the car with the delivered moment of the engine
  //! @param [in] xVel               longitudinal speed of the car (unused)
  //! @param [in] engineMoment       delivered moment of the engine
  //! @param [in] chosenGear         the curent gear
  //! @param [in] carParameters      parameters of the car
  //! @param [in] cycleTime          Cycle time of this components trigger task [ms] (unused)
  //! @return resulting acceleration [m/s²]
  units::acceleration::meters_per_second_squared_t GetAccFromEngineMoment(units::velocity::meters_per_second_t xVel,
                                                                          units::torque::newton_meter_t engineMoment,
                                                                          int chosenGear,
                                                                          int cycleTime);

  //! Returns the friction coefficient according to enviroment and car conditions
  //! @return friction coefficient (currently ALWAYS 1)
  double GetFrictionCoefficient();

  //! Returns the property with given name in the mantle_api::VehicleProperties
  //! or throws an error if the property is missing
  double GetVehicleProperty(const std::string &propertyName);

  /** @name Private Variables
   * @{
   * */
  /** @name Internal Parameters
   * @{
   */

  //! The minimal velocity of the agent [m/s].
  const units::velocity::meters_per_second_t VLowerLimit{0};

  /**
   *  @} */ // End of Internal Parameters
  /** @name External Parameters
   * @{
   */

  // --- Inputs

  //! Position of the accecaleration pedal position [%].
  double in_accPedalPos = 0;

  //! Position of the brake pedal position [%].
  double in_brakePedalPos = 0;

  //! Number of gears and position of gear.
  int in_gear = 0;

  //! The steering wheel angle [rad].
  units::angle::radian_t in_steeringWheelAngle{0};

  // --- Outputs

  //! Output Signal containing (aLong,v,x,y,dpsi,psi)
  DynamicsSignal dynamicsSignal;

  // --- Init Inputs

  //! Containing the vehicle parameters e.g. double carMass; double rDyn and more.
  mantle_api::VehicleProperties vehicleModelParameters;

  // Constants

  //! PI.
  units::angle::radian_t _twoPi = 2_rad * M_PI;

  //! value of earth gravity [m/s²].
  units::acceleration::meters_per_second_squared_t _oneG{9.81};

  //! air density.
  units::density::kilograms_per_cubic_meter_t _rho{1.23};

  units::angular_velocity::radians_per_second_t yawRatePrevious{0.0};
  /**
   *  @} */ // End of External Parameters
  /**
   *  @} */ // End of Private Variables
};
