/********************************************************************************
 * Copyright (c) 2018-2020 in-tech GmbH
 *               2022-2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

/*!
 * \addtogroup Dynamics_TrajectoryFollower
 * @{
 * \brief makes an agent strictly follow a predefined path
 *
 * The path is defined in a file that is loaded in the importer into a map.
 * The module simply reads the coordinates defined in the map every step and forces the agent to move to that position.
 * The vehicle is in this mode very simplified (there is no driver, no logic or any intelligence). It's basically a
 * "playback" of the trajectory.
 *
 * \section Dynamics_TrajectoryFollower_Inputs Inputs
 * Input variables:
 * name | meaning
 * -----|------
 * inputAcceleration | acceleration from some vehicle component, if any
 * velocity | current velocity
 * maxReachableState | max ComponentState for this module is given by ComponentController
 *
 * Input channel IDs:
 * | Input ID   | signal class              | contained variables               |
 * |------------|---------------------------|-----------------------------------|
 * | 0          | SteeringSignal            | componentState                    |
 * | 1          | AccelerationSignal        | inputAcceleration, componentState |
 * | 2          | TrajectorySignal          | trajectory                        |
 *
 * * \section Dynamics_TrajectoryFollower_Outputs Outputs
 * Output variables:
 * name                       | meaning
 * ---------------------------|------
 * acceleration               | Acceleration of the current agent corresponding to the pedalpositions [m/s²]
 * velocity                   | The driven curvature of the car corresponding to the steering wheel angle based on
 * Ackermann [radian] positionX                  | new inertial x-position [m] positionY                  | new inertial
 * y-position [m] yaw                        | new yaw angle in current time step [radian] yawRate                    |
 * change of yaw angle due to ds and curvature [radian] steeringWheelAngle         | new angle of the steering wheel
 * angle [radian] travelDistance             | distance traveled by the agent during this timestep [m]
 *
 *
 * Output channel IDs:
 * | Output Id                  | signal class                              | contained variables |
 * |----------------------------|-------------------------------------------|------------------------------------------------------------------------------------------------|
 * | 0                          | DynamicsSignal                            | acceleration, velocity, positionX,
 * positionY, yaw, yawRate, steeringWheelAngle, travelDistance |
 * @} */

#pragma once

#include <MantleAPI/Common/poly_line.h>
#include <MantleAPI/Common/trajectory.h>
#include <memory>
#include <string>
#include <units.h>
#include <vector>

#include "common/dynamicsSignal.h"
#include "common/vector2d.h"
#include "include/modelInterface.h"
#include "include/signalInterface.h"

class AgentInterface;
class CallbackInterface;
class ParameterInterface;
class PublisherInterface;
class ScenarioControlInterface;
class StochasticsInterface;
class WorldInterface;

using mantle_api::PolyLine;
using mantle_api::PolyLinePoint;
using mantle_api::Trajectory;

/*!
 * \brief Makes an agent strictly follow a predefined path.
 *
 * \ingroup Dynamics_TrajectoryFollower
 */
class TrajectoryFollowerImplementation : public UnrestrictedControllStrategyModelInterface
{
public:
  /// Name of this component
  static constexpr const char *COMPONENTNAME{"Dynamics_TrajectoryFollower"};

  /**
   * @brief Construct a new Trajectory Follower Implementation object
   *
   * @param[in]     componentName     Name of the component
   * @param[in]     isInit            Corresponds to "init" of "Component"
   * @param[in]     priority          Corresponds to "priority" of "Component"
   * @param[in]     offsetTime        Corresponds to "offsetTime" of "Component"
   * @param[in]     responseTime      Corresponds to "responseTime" of "Component"
   * @param[in]     cycleTime         Corresponds to "cycleTime" of "Component"
   * @param[in]     stochastics       Pointer to the stochastics class loaded by the framework
   * @param[in]     world             Pointer to the world
   * @param[in]     parameters        Pointer to the parameters of the module
   * @param[in]     publisher         Pointer to the publisher instance
   * @param[in]     callbacks         Pointer to the callbacks
   * @param[in]     agent             Pointer to agent instance
   * @param[in]     scenarioControl   scenarioControl of entity
   */
  TrajectoryFollowerImplementation(std::string componentName,
                                   bool isInit,
                                   int priority,
                                   int offsetTime,
                                   int responseTime,
                                   int cycleTime,
                                   StochasticsInterface *stochastics,
                                   WorldInterface *world,
                                   const ParameterInterface *parameters,
                                   PublisherInterface *const publisher,
                                   const CallbackInterface *callbacks,
                                   AgentInterface *agent,
                                   std::shared_ptr<ScenarioControlInterface> scenarioControl);

  /*!
   * \brief Update Inputs
   *
   * Function is called by framework when another component delivers a signal over
   * a channel to this component (scheduler calls update taks of other component).
   *
   * Refer to module description for input channels and input ids.
   *
   * @param[in]     localLinkId    Corresponds to "id" of "ComponentInput"
   * @param[in]     data           Referenced signal (copied by sending component)
   * @param[in]     time           Current scheduling time
   */
  void UpdateInput(int localLinkId, const std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Update outputs.
   *
   * Function is called by framework when this component has to deliver a signal over
   * a channel to another component (scheduler calls update task of this component).
   *
   * Refer to module description for output channels and output ids.
   *
   * @param[in]     localLinkId    Corresponds to "id" of "ComponentOutput"
   * @param[out]    data           Referenced signal (copied by this component)
   * @param[in]     time           Current scheduling time
   */
  void UpdateOutput(int localLinkId, std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Process data within component.
   *
   * Function is called by framework when the scheduler calls the trigger task
   * of this component
   *
   * @param[in]     time           Current scheduling time
   */
  void Trigger(int time) override;

  /*!
   * \brief Calculates dynamics for next timestep.
   */
  void CalculateNextTimestep();

private:
  const units::time::second_t cycleTimeInSeconds{0.0};

  bool initialization{true};
  bool enforceTrajectory{false};
  bool automaticDeactivation{false};
  bool inputAccelerationActive{false};

  units::time::millisecond_t currentTime{0};

  units::acceleration::meters_per_second_squared_t inputAcceleration{0.0};

  DynamicsSignal dynamicsOutputSignal{};

  Trajectory trajectory{};
  PolyLine polyLine;
  PolyLine::iterator previousTrajectoryIterator{};
  PolyLine::iterator nextTrajectoryIterator{};

  units::time::millisecond_t lastCoordinateTimestamp{0};
  units::velocity::meters_per_second_t lastVelocity{0.0};
  units::angular_velocity::radians_per_second_t lastYawVelocity{0.0};
  PolyLinePoint lastWorldPosition;
  double percentageTraveledBetweenCoordinates{0};

  ComponentState componentState{ComponentState::Disabled};
  bool canBeActivated{true};

  void Init();

  [[noreturn]] void ThrowCouldNotInstantiateSignalError();
  [[noreturn]] void ThrowInvalidSignalTypeError();

  ComponentState GetState() const;
  void UpdateState(const ComponentState newState);

  bool CheckEndOfTrajectory();

  void HandleEndOfTrajectory();

  void ParseParameters(const ParameterInterface *parameters);

  Common::Vector2d<units::length::meter_t> CalculateScaledVector(const PolyLinePoint &previousPosition,
                                                                 const PolyLinePoint &nextPosition,
                                                                 const double &factor);
  units::angle::radian_t CalculateScaledDeltaYawAngle(const PolyLinePoint &previousPosition,
                                                      const PolyLinePoint &nextPosition,
                                                      const double &factor);

  units::length::meter_t CalculateDistanceBetweenWorldCoordinates(PolyLinePoint previousPosition,
                                                                  PolyLinePoint nextPosition);

  void TriggerWithActiveAccelerationInput();
  void TriggerWithInactiveAccelerationInput();

  void SetComponentState(const ComponentState newState);

  void UpdateDynamics(const PolyLinePoint &previousPosition,
                      const Common::Vector2d<units::length::meter_t> &direction,
                      units::velocity::meters_per_second_t velocity,
                      Common::Vector2d<units::length::meter_t> movementDirection,
                      units::angle::radian_t deltaYawAngle);
};
