/********************************************************************************
 * Copyright (c) 2020-2021 ITK Engineering GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#ifndef SENSOR_COLLISION_IMPLEMENTATION_H
#define SENSOR_COLLISION_IMPLEMENTATION_H

#include <map>
#include <memory>
#include <new>
#include <stddef.h>
#include <string>

#include "common/componentPorts.h"
#include "common/primitiveSignals.h"
#include "include/modelInterface.h"

class AgentInterface;
class CallbackInterface;
class ParameterInterface;
class PublisherInterface;
class SignalInterface;
class StochasticsInterface;
class WorldInterface;

/**
 * \addtogroup Components_PCM openPASS components pcm
 * @{
 * \addtogroup Sensor_Collision
 *
 * \brief Sensor component to sense a collision
 *
 * \details This module provides information about whether a collision occured.\n
 * At every timestep it checks the AgentInterface function #AgentInterface::GetCollisionPartners()
 * whether the current count of collisions has changed. In this case a collision occured. \n
 * To maintain a specific time the agents can penetrate eachother, the output signal collisionOccured
 * stays false till the penetrationTime is over or if the collision is already over.
 * After that condition the component signals that a collision occured to other components
 * as long as the collision is still detected.
 *
 * image html @cond Collision_Integrated.png @endcond "State machine of collision handling"
 *
 * @}
 */

/*!
 * \copydoc Sensor_Collision
 * \ingroup Sensor_Collision
 */
class Sensor_Collision_Implementation : public SensorInterface
{
public:
  /// Name of this component
  const std::string COMPONENTNAME = "Sensor_Collision";

  //! Constructor
  //!
  //! @param[in]     componentName  Name of the component
  //! @param[in]     isInit         Corresponds to "init" of "Component"
  //! @param[in]     priority       Corresponds to "priority" of "Component"
  //! @param[in]     offsetTime     Corresponds to "offsetTime" of "Component"
  //! @param[in]     responseTime   Corresponds to "responseTime" of "Component"
  //! @param[in]     cycleTime      Corresponds to "cycleTime" of "Component"
  //! @param[in]     stochastics    Pointer to the stochastics class loaded by the framework
  //! @param[in]     world          Pointer to the world interface
  //! @param[in]     parameters     Pointer to the parameters of the module
  //! @param[in]     publisher      Pointer to the publisher instance
  //! @param[in]     callbacks      Pointer to the callbacks
  //! @param[in]     agent          Pointer to agent instance
  Sensor_Collision_Implementation(std::string componentName,
                                  bool isInit,
                                  int priority,
                                  int offsetTime,
                                  int responseTime,
                                  int cycleTime,
                                  StochasticsInterface *stochastics,
                                  WorldInterface *world,
                                  const ParameterInterface *parameters,
                                  PublisherInterface *const publisher,
                                  const CallbackInterface *callbacks,
                                  AgentInterface *agent);

  //! Destructor
  virtual ~Sensor_Collision_Implementation();

  /*!
   * \brief Update Inputs
   *
   * Function is called by framework when another component delivers a signal over
   * a channel to this component (scheduler calls update taks of other component).
   *
   * Refer to module description for input channels and input ids.
   *
   * \param[in]     localLinkId    Corresponds to "id" of "ComponentInput"
   * \param[in]     data           Referenced signal (copied by sending component)
   * \param[in]     time           Current scheduling time
   */
  void UpdateInput(int localLinkId, const std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Update outputs.
   *
   * Function is called by framework when this component has to deliver a signal over
   * a channel to another component (scheduler calls update task of this component).
   *
   * Refer to module description for output channels and output ids.
   *
   * \param[in]     localLinkId    Corresponds to "id" of "ComponentOutput"
   * \param[out]    data           Referenced signal (copied by this component)
   * \param[in]     time           Current scheduling time
   */
  void UpdateOutput(int localLinkId, std::shared_ptr<SignalInterface const> &data, int time) override;

  /*!
   * \brief Process data within component.
   *
   * Function is called by framework when the scheduler calls the trigger task
   * of this component.
   *
   * Refer to module description for information about the module's task.
   *
   * \param[in]     time           Current scheduling time
   */
  void Trigger(int time) override;

private:
  std::map<int, ComponentPort *> outputPorts;  //!< map for all OutputPort
  /** \addtogroup Sensor_Collision
   *  @{
   *      \name OutputPorts
   *      All output ports with PortId
   *      @{
   */
  OutputPort<BoolSignal, bool> collisionOccured{0, &outputPorts};  //!< true if collision occured
  /**
   *      @}
   *  @}
   */

  // std::map<std::string, externalParameter<int>*> parameterMapInt;
  /** \ingroup Sensor_Collision
   *  @{
   *      \name External Parameter
   *      Parameter which are set externally in agentConfiguration file.
   *      @{
   */
  int penetrationTime = 30;  //!< offset time after a collision occured and is set as output
  // externalParameter<int> penetrationTime {0, &parameterMapInt }; //!< offset time after a collision occured and is
  // set as output
  /**
   *      @}
   *  @}
   */

  //! time of first collision
  int timeOfFirstCollision = 0;

  //! Stores th number of collisionData of the previous time step
  size_t previousNumberOfCollisionData = 0;

  //! Specifies the current state of collision the agent is in
  enum class CollisionState
  {
    NOCOLLISION,
    COLLISIONPENETRATING,
    COLLISION
  } collisionState
      = CollisionState::NOCOLLISION;
};

#endif  // SENSOR_COLLISION_IMPLEMENTATION_H
