/********************************************************************************
 * Copyright (c) 2016-2018 ITK Engineering GmbH
 *               2017-2019 in-tech GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
//! @file  spawnPointLibrary.h
//! @brief This file contains the internal representation of the library of a
//!        spawn point.
//-----------------------------------------------------------------------------

#pragma once

#include <boost/dll.hpp>
#include <boost/function.hpp>
#include <memory>
#include <string>
#include <utility>

class CallbackInterface;
class SpawnPointInterface;
struct SpawnPointDependencies;

namespace core
{
class SpawnPoint;

//! This class represents the library of a spawn point
class SpawnPointLibrary
{
public:
  //! Type representing the signature of a dll-function for obtaining the version information of the module
  using SpawnPointInterface_GetVersionType = const std::string&();
  //! Type representing the signature of a dll-function for creating an instance of the module
  using SpawnPointInterface_CreateInstanceType
      = std::unique_ptr<SpawnPointInterface>(const SpawnPointDependencies* dependencies,
                                             const CallbackInterface* callbacks);

  /**
   * @brief SpawnPointLibrary constructor
   *
   * @param[in] libraryPath   Path to the library
   * @param[in] callbacks     Pointer to the callbacks
   */
  SpawnPointLibrary(std::string libraryPath, CallbackInterface* callbacks)
      : libraryPath(std::move(libraryPath)), callbacks(callbacks)
  {
  }
  SpawnPointLibrary(const SpawnPointLibrary&) = delete;
  SpawnPointLibrary(SpawnPointLibrary&&) = delete;
  SpawnPointLibrary& operator=(const SpawnPointLibrary&) = delete;
  SpawnPointLibrary& operator=(SpawnPointLibrary&&) = delete;

  //-----------------------------------------------------------------------------
  //! Destructor, deletes the stored library (unloads it if necessary), if the list
  //! of stored spawn points is empty.
  //-----------------------------------------------------------------------------
  virtual ~SpawnPointLibrary();

  //-----------------------------------------------------------------------------
  //! Creates a boost::dll::shared_library based on the path from the constructor
  //! and stores boost::function for getting the library version, creating and
  //! destroying instances and setting the spawn item
  //! (see typedefs for corresponding signatures).
  //!
  //! @return                 Null pointer
  //-----------------------------------------------------------------------------l
  bool Init();

  //-----------------------------------------------------------------------------
  //! Make sure that the library exists and is loaded, then call the "create instance"
  //! function pointer using the parameters from the spawn point instance to get
  //! a spawn point interface, which is then used to instantiate a spawn point
  //! which is also stored in the list of spawn points.
  //!
  //! @param[in]  dependencies    Spawn point dependencies
  //! @return Spawn point created
  //-----------------------------------------------------------------------------
  std::unique_ptr<SpawnPoint> CreateSpawnPoint(const SpawnPointDependencies& dependencies);

private:
  const std::string DllGetVersionId = "OpenPASS_GetVersion";
  const std::string DllCreateInstanceId = "OpenPASS_CreateInstance";

  std::string libraryPath;
  boost::dll::shared_library library;
  CallbackInterface* callbacks;
  boost::function<SpawnPointInterface_GetVersionType> getVersionFunc;
  boost::function<SpawnPointInterface_CreateInstanceType> createInstanceFunc;
};

}  // namespace core
