/********************************************************************************
 * Copyright (c) 2019-2020 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include "dynamicAgentTypeGenerator.h"

#include <algorithm>
#include <map>
#include <optional>
#include <stdexcept>
#include <unordered_map>
#include <utility>

#include "framework/dynamicParametersSampler.h"
#include "framework/dynamicProfileSampler.h"
#include "include/agentBlueprintInterface.h"
#include "include/agentTypeInterface.h"
#include "include/profilesInterface.h"
#include "include/systemConfigInterface.h"
#include "modelElements/componentType.h"

constexpr const char* DRIVER{"Driver"};

DynamicAgentTypeGenerator::DynamicAgentTypeGenerator(SampledProfiles& sampledProfiles,
                                                     DynamicParameters& dynamicParameters,
                                                     std::shared_ptr<SystemConfigInterface> systemConfigBlueprint,
                                                     const ProfilesInterface* profiles)
    : sampledProfiles(sampledProfiles),
      dynamicParameters(dynamicParameters),
      systemConfigBlueprint(std::move(systemConfigBlueprint)),
      profiles(profiles)
{
}

DynamicAgentTypeGenerator& DynamicAgentTypeGenerator::GatherBasicComponents()
{
  for (const auto& component : defaultComponents.basicComponentNames)
  {
    GatherComponent(component, agentBuildInformation.agentType);
  }

  return *this;
}

DynamicAgentTypeGenerator& DynamicAgentTypeGenerator::GatherDriverComponents()
{
  const auto driverParameters = profiles->GetProfile(DRIVER, sampledProfiles.driverProfileName);
  auto driverModule
      = openpass::parameter::Get<std::string>(driverParameters, "Type");  // Existence checked by ProfilesImporter

  if (auto driverParameterModule = openpass::parameter::Get<std::string>(driverParameters, "ParametersModule"))
  {
    GatherComponentWithParameters(driverParameterModule.value(), agentBuildInformation.agentType, driverParameters);
    GatherComponent(driverModule.value(), agentBuildInformation.agentType);
  }
  else
  {
    GatherComponentWithParameters(driverModule.value(), agentBuildInformation.agentType, driverParameters);
  }

  auto sensorDriverModule = openpass::parameter::Get<std::string>(driverParameters, "SensorDriverModule");
  GatherComponent(sensorDriverModule.value_or("Sensor_Driver"), agentBuildInformation.agentType);

  auto algorithmLateralModule = openpass::parameter::Get<std::string>(driverParameters, "AlgorithmLateralModule");
  GatherComponent(algorithmLateralModule.value_or("AlgorithmLateralDriver"), agentBuildInformation.agentType);

  auto algorithmLongitudinalModule
      = openpass::parameter::Get<std::string>(driverParameters, "AlgorithmLongitudinalModule");
  GatherComponent(algorithmLongitudinalModule.value_or("AlgorithmLongitudinalDriver"), agentBuildInformation.agentType);

  for (const auto& componentName : defaultComponents.driverComponentNames)
  {
    GatherComponent(componentName, agentBuildInformation.agentType);
  }

  return *this;
}

DynamicAgentTypeGenerator& DynamicAgentTypeGenerator::GatherVehicleComponents()
{
  std::shared_ptr<core::AgentType> agentType = agentBuildInformation.agentType;
  VehicleComponentProfileNames vehicleComponentProfileNames = sampledProfiles.vehicleComponentProfileNames;

  if (!vehicleComponentProfileNames.empty())
  {
    for (const std::string& componentName : defaultComponents.vehicleComponentNames)
    {
      GatherComponent(componentName, agentType);
    }
  }

  for (const auto& vehicleComponentProfile : vehicleComponentProfileNames)
  {
    auto parameters = profiles->CloneProfile(vehicleComponentProfile.first, vehicleComponentProfile.second);

    if (profiles->GetSystemProfiles().count(sampledProfiles.systemProfileName) == 0)
    {
      throw std::logic_error("No vehicle profile of type " + sampledProfiles.systemProfileName);
    }
    auto systemProfile = profiles->GetSystemProfiles().at(
        sampledProfiles.systemProfileName);  // Existence checked by DynamicProfileSampler

    auto matchedComponent = std::find_if(systemProfile.vehicleComponents.begin(),
                                         systemProfile.vehicleComponents.end(),
                                         [vehicleComponentProfile](const VehicleComponent& curComponent)
                                         { return curComponent.type == vehicleComponentProfile.first; });

    openpass::parameter::internal::ParameterListLevel2 parameterList;
    for (auto link : matchedComponent->sensorLinks)
    {
      openpass::parameter::internal::ParameterSetLevel3 param;
      param.emplace_back("SensorId", link.sensorId);
      param.emplace_back("InputId", link.inputId);
      parameterList.emplace_back(param);
    }

    parameters.emplace_back("SensorLinks", parameterList);

    GatherComponentWithParameters(vehicleComponentProfile.first, agentType, parameters);
  }

  return *this;
}

DynamicAgentTypeGenerator& DynamicAgentTypeGenerator::GatherSensors()
{
  const std::string sensorAggregationModulName = "SensorAggregation";

  GatherComponent(sensorAggregationModulName, agentBuildInformation.agentType);

  int inputIdSensorAggregation = systemConfigBlueprint->GetSystems()
                                     .at(0)
                                     ->GetComponents()
                                     .at(sensorAggregationModulName)
                                     ->GetInputLinks()
                                     .at(0);
  int sensorNumber = 0;
  const auto systemProfile = profiles->GetSystemProfiles().at(
      sampledProfiles.systemProfileName);  // Existence checked by DynamicProfileSampler

  for (const auto& sensor : systemProfile.sensors)
  {
    auto sensorParam = profiles->CloneProfile(sensor.profile.type, sensor.profile.name);

    // add specific information
    sensorParam.emplace_back("Name", sensor.profile.name);
    sensorParam.emplace_back("Type", sensor.profile.type);
    sensorParam.emplace_back("Id", sensor.id);
    sensorParam.emplace_back("Position", sensor.positionName);
    sensorParam.emplace_back("Latency", dynamicParameters.sensorLatencies.at(sensor.id));

    const std::string uniqueSensorName = "Sensor_" + std::to_string(sensor.id);
    GatherComponentWithParameters(
        uniqueSensorName, agentBuildInformation.agentType, sensorParam, "SensorObjectDetector", sensorNumber);

    if (sensorNumber > 0)
    {
      agentBuildInformation.agentType->GetComponents()
          .at(sensorAggregationModulName)
          ->AddInputLink(sensorNumber, inputIdSensorAggregation + sensorNumber);
    }

    // clone sensor and set specific parameters
    openpass::sensors::Parameter clonedSensor(sensor);
    clonedSensor.profile.parameter = sensorParam;
    agentBuildInformation.sensorParameters.push_back(clonedSensor);

    sensorNumber++;
  }

  return *this;
}

void DynamicAgentTypeGenerator::GatherComponent(
    const std::string componentName,             //NOLINT(performance-unnecessary-value-param)
    std::shared_ptr<core::AgentType> agentType)  //NOLINT(performance-unnecessary-value-param)
{
  const auto& systemComponents = systemConfigBlueprint->GetSystems().at(0)->GetComponents();
  auto findResult = systemComponents.find(componentName);
  if (findResult == systemComponents.end())
  {
    throw std::runtime_error("Could not find component " + componentName + " in systemConfigBlueprint");
  }
  auto componentType = std::make_shared<core::ComponentType>(*(findResult->second));

  for (const auto& componentOutput : componentType->GetOutputLinks())
  {
    agentType->AddChannel(componentOutput.second);
  }

  if (!agentType->AddComponent(componentType))
  {
    throw std::runtime_error("Could not add component to agentType");
  }
}

void DynamicAgentTypeGenerator::GatherComponentWithParameters(const std::string& componentName,
                                                              std::shared_ptr<core::AgentType> agentType,
                                                              const openpass::parameter::ParameterSetLevel1& parameters)
{
  GatherComponentWithParameters(componentName, std::move(agentType), parameters, componentName, 0);
}

void DynamicAgentTypeGenerator::GatherComponentWithParameters(
    const std::string& componentName,
    std::shared_ptr<core::AgentType> agentType,  //NOLINT(performance-unnecessary-value-param)
    const openpass::parameter::ParameterSetLevel1& parameters,
    std::string componentNameInSystemConfigBlueprint,  //NOLINT(performance-unnecessary-value-param)
    int channelOffset)
{
  const auto& systemComponents = systemConfigBlueprint->GetSystems().at(0)->GetComponents();
  auto findResult = systemComponents.find(componentNameInSystemConfigBlueprint);
  if (findResult == systemComponents.end())
  {
    throw std::runtime_error("Could not find component " + componentNameInSystemConfigBlueprint
                             + " in systemConfigBlueprint");
  }
  auto componentTypeFromSystemConfigBlueprint = std::make_shared<core::ComponentType>(*(findResult->second));

  auto componentType = std::make_shared<core::ComponentType>(componentName,
                                                             componentTypeFromSystemConfigBlueprint->GetInit(),
                                                             componentTypeFromSystemConfigBlueprint->GetPriority(),
                                                             componentTypeFromSystemConfigBlueprint->GetOffsetTime(),
                                                             componentTypeFromSystemConfigBlueprint->GetResponseTime(),
                                                             componentTypeFromSystemConfigBlueprint->GetCycleTime(),
                                                             componentTypeFromSystemConfigBlueprint->GetModelLibrary());

  for (const auto& componentOutput : componentTypeFromSystemConfigBlueprint->GetOutputLinks())
  {
    componentType->AddOutputLink(componentOutput.first, componentOutput.second + channelOffset);
    agentType->AddChannel(componentOutput.second + channelOffset);
  }
  for (const auto& componentInput : componentTypeFromSystemConfigBlueprint->GetInputLinks())
  {
    componentType->AddInputLink(componentInput.first, componentInput.second + channelOffset);
  }

  componentType->SetModelParameter(parameters);

  if (!agentType->AddComponent(componentType))
  {
    throw std::runtime_error("Could not add component to agentType");
  }
}

DynamicAgentTypeGenerator AgentBuildInformation::make(SampledProfiles& sampledProfiles,
                                                      DynamicParameters& dynamicParameters,
                                                      std::shared_ptr<SystemConfigInterface> systemConfigBlueprint,
                                                      const ProfilesInterface* profiles)
{
  return {sampledProfiles, dynamicParameters, std::move(systemConfigBlueprint), profiles};
}
