/********************************************************************************
 * Copyright (c) 2019-2020 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#pragma once

#include <memory>
#include <string>
#include <utility>
#include <vector>

#include "common/parameter.h"
#include "common/sensorDefinitions.h"
#include "modelElements/agentType.h"

class ParameterInterface;
class ProfilesInterface;
class SystemConfigInterface;
struct DynamicParameters;
struct SampledProfiles;

/**
 * @brief Defines the default components that all agents need
 */
struct DefaultComponents
{
  /// Basic components that are always needed
  const std::vector<std::string> basicComponentNames{"AgentUpdater",
                                                     "ComponentController",
                                                     "ControllerSwitch",
                                                     "Dynamics_Collision",
                                                     "Dynamics_RegularDriving",
                                                     "Dynamics_Scenario",
                                                     "Parameters_Vehicle",
                                                     "PrioritizerDynamics",
                                                     "PrioritizerLongitudinal",
                                                     "PrioritizerSteering"};

  /// Components that are needed, if at least one driver is present
  const std::vector<std::string> driverComponentNames{
      "Action_LongitudinalDriver", "Action_SecondaryDriverTasks", "PrioritizerTurningIndicator"};

  /// Components that are needed, if at least one vehicle component is present
  const std::vector<std::string> vehicleComponentNames{"Algorithm_LongitudinalVehicleComponents",
                                                       "PrioritizerAccelerationVehicleComponents",
                                                       "PrioritizerSteeringVehicleComponents",
                                                       "LimiterAccelerationVehicleComponents",
                                                       "SensorFusionErrorless"};
};

class DynamicAgentTypeGenerator;

/**
 * @brief The AgentBuildInformation struct contains all information that the AgentBlueprintProvider needs to
 * fill out the AgentBlueprint for a dynamic AgentProfile.
 *
 * An instance of this struct is aquired by calling the static make function and then sequentially calling all
 * building function on the returned DynamicAgentTypeGenerator
 */
struct AgentBuildInformation
{
  /**
   * @brief Builder helper
   * @relates DynamicAgentTypeGenerator
   */
  friend class DynamicAgentTypeGenerator;

  std::shared_ptr<core::AgentType> agentType
      = std::make_shared<core::AgentType>();       ///< AgentType where the results get stored
  openpass::sensors::Parameters sensorParameters;  ///< Parameters which describe the sensor

  /**
   * @brief Creates a new DynamicAgentTypeGenerator which can be used to build an instance of this struct.
   *
   * \param sampledProfiles         Profile names sampled by the DynamicProfilesSampler
   * \param dynamicParameters       Parameters samplied by the DynamicParametersSampler
   * \param systemConfigBlueprint   Imported SystemConfigBlueprint
   * \param profiles                Imported profiles from the ProfilesCatalog
   *
   * @return new instance of DynamicAgentTypeGenerator
   */
  static DynamicAgentTypeGenerator make(SampledProfiles &sampledProfiles,
                                        DynamicParameters &dynamicParameters,
                                        std::shared_ptr<SystemConfigInterface> systemConfigBlueprint,
                                        const ProfilesInterface *profiles);

private:
  /// Private constructor to be called by the DynamicAgentTypeGenerator
  AgentBuildInformation() = default;
};

/**
 * @brief The DynamicAgentTypeGenerator class is used to build the AgentBuildInformation struct in particular the
 * AgentType
 *
 * This class is responsible for generating the AgentType for building an agent in the AgentFactory in the case of a
 * dynamic AgentProfile. This AgentType contains the Driver, all Sensors and VehicleComponents that are definied by the
 * sampled profiles as well as the defaultComponents, that are always needed. The DynamicAgentTypeGenerator also stores
 * the information about the channels between these components in the AgentType.
 */
class DynamicAgentTypeGenerator
{
public:
  /**
   * @brief DynamicAgentTypeGenerator constructor
   *
   * \param sampledProfiles         Profile names sampled by the DynamicProfilesSampler
   * \param dynamicParameters       Parameters samplied by the DynamicParametersSampler
   * \param systemConfigBlueprint   Imported SystemConfigBlueprint
   * \param profiles                Imported profiles from the ProfilesCatalog
   */
  DynamicAgentTypeGenerator(SampledProfiles &sampledProfiles,
                            DynamicParameters &dynamicParameters,
                            std::shared_ptr<SystemConfigInterface> systemConfigBlueprint,
                            const ProfilesInterface *profiles);

  /**
   * @brief Returns the generated AgentBuildInformation;
   */
  operator AgentBuildInformation &&() { return std::move(agentBuildInformation); }

  /**
   * @brief Creates a new ComponentType for all basic components defined in DefaultComponents.basicComponents and stores
   * them in the AgentType. Also collects there output channels
   *
   * @return reference to itself
   */
  DynamicAgentTypeGenerator &GatherBasicComponents();

  /**
   * @brief Creates a new ComponentType for all default driver components defined in DefaultComponents.basicComponents,
   * the ParametersAgents and the driver module itself and stores them in the AgentType. Also collects there output
   * channels
   *
   * @return reference to itself
   */
  DynamicAgentTypeGenerator &GatherDriverComponents();

  /**
   * @brief Creates a new ComponentType for all default vehicle components defined in DefaultComponents.basicComponents
   * and the sampled vehicle components module and stores them in the AgentType. Also collects there output channels
   *
   * @return reference to itself
   */
  DynamicAgentTypeGenerator &GatherVehicleComponents();

  /**
   * @brief Creates a new ComponentType for all sensor defined by the VehicleModel and the SensorFusion module if needed
   * and stores them in the AgentType. Also collects there output channels
   *
   * @return reference to itself
   */
  DynamicAgentTypeGenerator &GatherSensors();

  /**
   * @brief Gathers a component and adds it to the AgentType
   *
   * @details  Retrieves the information of a component from the app config and adds the ComponentType to the AgentType.
   * The parameter set is read from the app config specified by the parameter set name
   *
   * @param[in]    componentName               Name of the component
   * @param[out]   agentType                   AgentType where the results get stored
   */
  void GatherComponent(const std::string componentName, std::shared_ptr<core::AgentType> agentType);

  /**
   * @brief Gathers a component with specified parameter set and adds it to the AgentType
   *
   * @details  Retrieves the information of a component from the app config and adds the ComponentType to the AgentType.
   * The parameter set is explictly specified.
   *
   * @param[in]    componentName               Name of the component
   * @param[out]   agentType                   AgentType where the results get stored
   * @param[in]    parameters                  Parameters set.
   */
  void GatherComponentWithParameters(const std::string &componentName,
                                     std::shared_ptr<core::AgentType> agentType,
                                     const openpass::parameter::ParameterSetLevel1 &parameters);

  /**
   * @brief Gathers a component and adds it to the AgentType
   *
   * @details  Retrieves the information of a component from the app config and adds the ComponentType to the AgentType.
   * The parameter set is explictly specified.
   *
   * @param[in]    componentName                           Name of the component
   * @param[out]   agentType                               AgentType where the results get stored
   * @param[in]    parameters                              Parameters set.
   * @param[in]    componentNameInSystemConfigBlueprint    Name of the component in the app config, different if
   * different from componentName (currently used for sensors)
   * @param[in]    channelOffset                           Increase in the ids of the output channels (currently used
   * for sensors)
   */
  void GatherComponentWithParameters(const std::string &componentName,
                                     std::shared_ptr<core::AgentType> agentType,
                                     const openpass::parameter::ParameterSetLevel1 &parameters,
                                     std::string componentNameInSystemConfigBlueprint,
                                     int channelOffset);

  /**
   * @brief Samples the latency of a sensor
   *
   * @details  Samples the latency of a sensor based on the parameters provided.
   *
   * @param[in]    curParameters               Parameter interface containing the mean value, standard deviation,
   * minimum and maximum
   * @return       latency in s
   */
  double SampleSensorLatency(std::shared_ptr<ParameterInterface> curParameters);

private:
  AgentBuildInformation agentBuildInformation;

  SampledProfiles &sampledProfiles;
  DynamicParameters &dynamicParameters;
  std::shared_ptr<SystemConfigInterface> systemConfigBlueprint;
  const ProfilesInterface *profiles;
  DefaultComponents defaultComponents;
};
