/********************************************************************************
 * Copyright (c) 2019 in-tech GmbH
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#pragma once
#include <string>
#include <unordered_map>
#include <utility>

class DynamicProfileSampler;
class ProfilesInterface;
class StochasticsInterface;

/*!
 * \brief The SampledProfiles struct constains the name of all profiles sampled by the DynamicProfileSampler.
 * With this the DynamicAgentTypeGenerator can generate an AgentType.
 *
 * An instance of this struct is aquired by calling the static make function and then sequentially calling
 * all sample function on the returned DynamicProfileSampler.
 */
struct SampledProfiles
{
  /**
   * @brief Sampler helper
   * @relates DynamicProfileSampler
   */
  friend class DynamicProfileSampler;

  std::string driverProfileName;                                              //!< Name of the driver profile
  std::string systemProfileName;                                              //!< Name of the system profile
  std::unordered_map<std::string, std::string> vehicleComponentProfileNames;  //!< Vehicle component profile names

  /*!
   * \brief Creates a new DynamicProfileSampler which can be used to build an instance of this struct.
   * \param agentProfileName  name of the AgentProfile samplied by the AgentProfileSampler
   * \param stochastics       Reference to the stochasticInterface
   * \param profiles          imported profiles from the ProfilesCatalog
   * \return new instance of DynamicProfileSampler
   */
  static DynamicProfileSampler make(std::string agentProfileName,
                                    StochasticsInterface& stochastics,
                                    const ProfilesInterface* profiles);

private:
  SampledProfiles() = default;
};

/*!
 * \brief The DynamicProfileSampler class is responsible for sampling all profile names (except AgentProfile which is sampled by the AgentProfilesSampler)
 */
class DynamicProfileSampler
{
public:
  //! DynamicProfileSampler constructor
  //!
  //! @param[in] agentProfileName     Name of the AgentProfile samplied by the AgentProfileSampler
  //! @param[in] stochastics          Reference to the stochasticInterface
  //! @param[in] profiles             Imported profiles from the ProfilesCatalog
  DynamicProfileSampler(std::string agentProfileName,
                        StochasticsInterface& stochastics,
                        const ProfilesInterface* profiles)
      : agentProfileName(agentProfileName), stochastics(stochastics), profiles(profiles)
  {
  }

  /*!
   * \brief Samples the driver profile name based on the agent profile name.
   * \return reference to itself
   */
  DynamicProfileSampler& SampleDriverProfile();

  /*!
   * \brief Samples the vehicle profile name based on the agent profile name.
   * \return reference to itself
   */
  DynamicProfileSampler& SampleSystemProfile();

  /*!
   * \brief Samples all vehicle component profile names based on the vehicle profile name.
   * Therefore SampleSystemProfile has to be called before this method.
   *
   * \return reference to itself
   */
  DynamicProfileSampler& SampleVehicleComponentProfiles();

  /*!
   * \brief Returns the generated SampledProfiles struct.
   */
  operator SampledProfiles&&() { return std::move(sampledProfiles); }

private:
  SampledProfiles sampledProfiles;
  std::string agentProfileName;
  StochasticsInterface& stochastics;
  const ProfilesInterface* profiles;
};
