/********************************************************************************
 * Copyright (c) 2017-2019 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include "agentParser.h"

#include <cstddef>
#include <functional>
#include <map>
#include <utility>

#include "framework/scheduler/tasks.h"
#include "include/componentInterface.h"
#include "modelElements/agent.h"
#include "modelElements/channel.h"

namespace core::scheduling
{

using namespace core;

AgentParser::AgentParser(const int &currentTime) : currentTime{currentTime} {}

void AgentParser::Parse(const Agent &agent)
{
  for (auto [_, component] : agent.GetComponents())
  {
    std::vector<TaskItem> taskItems;

    auto priority = component->GetPriority();
    auto cycleTime = component->GetCycleTime();
    auto triggerDelay = component->GetOffsetTime();
    auto updateDelay = component->GetResponseTime();
    auto agentId = agent.GetId();

    std::function<bool()> triggerFunc = std::bind(  // NOLINT(modernize-avoid-bind)
        &ComponentInterface::TriggerCycle,
        component,
        std::ref(currentTime));
    taskItems.push_back({TriggerTaskItem(agentId, priority, cycleTime, triggerDelay, triggerFunc)});
    // std::cout << "Trigger Task: AgentId: " << agentId << " prio: " << priority <<
    //               " cycleTime: " << cycleTime <<  " delay: " << triggerDelay << std::endl;

    for (const auto &[outputLinkId, outputLink] : component->GetOutputLinks())
    {
      std::function<bool()> updateFunc
          = std::bind(&ComponentInterface::AcquireOutputData,  // NOLINT(modernize-avoid-bind)
                      component,
                      outputLinkId,
                      std::ref(currentTime));
      taskItems.push_back({UpdateTaskItem(agentId, priority, cycleTime, updateDelay, updateFunc)});
      // std::cout << "UpdateOUT Task: AgentId: " << agentId << " prio: " << priority <<
      //               " cycleTime: " << cycleTime <<  " delay: " << updateDelay << std::endl;

      for (const auto &target : outputLink->GetTargets())
      {
        int targetLinkId = std::get<static_cast<size_t>(Channel::TargetLinkType::LinkId)>(target);
        ComponentInterface *targetComponent = std::get<static_cast<size_t>(Channel::TargetLinkType::Component)>(target);

        std::function<bool()> updateInFunc
            = std::bind(&ComponentInterface::UpdateInputData,  // NOLINT(modernize-avoid-bind)
                        targetComponent,
                        targetLinkId,
                        std::ref(currentTime));
        taskItems.push_back({UpdateTaskItem(agentId, priority, cycleTime, updateDelay, updateInFunc)});
        // std::cout << "UpdateIN Task: AgentId: " << agentId << " prio: " << priority <<
        //               " cycleTime: " << cycleTime <<  " delay: " << updateDelay << std::endl;
      }
    }

    if (component->GetInit())
    {
      nonRecurringTasks.insert(nonRecurringTasks.end(),
                               std::make_move_iterator(taskItems.begin()),
                               std::make_move_iterator(taskItems.end()));
    }
    else
    {
      recurringTasks.insert(
          recurringTasks.end(), std::make_move_iterator(taskItems.begin()), std::make_move_iterator(taskItems.end()));
    }
  }
}

std::vector<TaskItem> AgentParser::GetNonRecurringTasks()
{
  return nonRecurringTasks;
}

std::vector<TaskItem> AgentParser::GetRecurringTasks()
{
  return recurringTasks;
}

}  //namespace core::scheduling
