/********************************************************************************
 * Copyright (c) 2017-2021 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
/*!
 * \file  observation_entityRepositoryImplementation.h
 * \brief Adds the RunStatistic information to the simulation output.
 * \details  Writes the RunStatistic information into the simulation output.
 *           Also manages the stop reasons of the simulation.
 */
//-----------------------------------------------------------------------------

#pragma once

#include <filesystem>
#include <stdint.h>
#include <string>
#include <unordered_map>
#include <utility>
#include <vector>

#include "include/dataBufferInterface.h"
#include "include/observationInterface.h"

class CallbackInterface;
class ParameterInterface;
class RunResultInterface;
class StochasticsInterface;
class WorldInterface;

/// Container for Metainfo provided by an entity
using Metainfo = std::unordered_map<std::string, std::string>;

static constexpr const char *NON_PERSISTENT_ENTITIES{"Entities/NonPersistent"};
static constexpr const char *PERSISTENT_ENTITIES{"Entities/Persistent"};

/**
 * @brief Definition of the headers, which shall be written to the CSV
 * @note  "id" and "source" are fixed (\see CsvRow), the other columns are for describing metainfo (\see CSV_METAINFO)
 */
static const std::vector<std::string> CSV_HEADERS{"id", "source", "version", "name", "secondary id", "type", "subtype"};

/**
 * @brief Definition of metainfo-mapping, where the first entry maps to the first column after the fixed header columns
 *
 * Example:
 * If an entity defines "MetaInfo/id", the value will be published to the column "secondary id", otherwise the value
 * will be ""
 *
 * The structure allows to define alternative mappings by extending the mapping e.g. for ids to {"id", "ID",
 * "Identifier", "__id__"}. If an entity now defines any of the given entries (e.g. "MetaInfo/__id__"), it will be
 * mapped to the same column This helps to group arbitary entities to the same column
 */
static const std::vector<std::vector<std::string>> CSV_METAINFO{{"version"}, {"name"}, {"id"}, {"type"}, {"subtype"}};

/// The delimiter for the CSV file
static constexpr const char *CSV_DELIMITER{","};

/**
 * @brief Single entry holding information for printing entity information into CSV
 */
struct CsvRow
{
  /**
   * @brief CsvRow constructor
   *
   * @param[in]   id          Entity id
   * @param[in]   source      The source attribute from an entity
   * @param[in]   metainfo    Metainfo provided by an entity
   */
  CsvRow(uint64_t id, std::string source, Metainfo metainfo)
      : id{id}, source{std::move(source)}, metainfo{std::move(metainfo)}
  {
  }

  /// Entity id
  const uint64_t id;
  /// The source attribute from an entity
  const std::string source;
  /// Metainfo provided by an entity
  const Metainfo metainfo;
};

/**
 * \brief   This class adds the RunStatistic information to the simulation output.
 *   \details This class inherits the ObservationLogGeneric which creates the basic simulation output
 *            and adds the RunStatistic information to the output.
 *            This class also manages the stop reasons of the simulation.
 *
 *   \ingroup ObservationLog
 */
class ObservationEntityRepositoryImplementation : ObservationInterface
{
public:
  /// Name of the current component
  const std::string COMPONENTNAME = "Observation_EntityRepository";

  /**
   * @brief ObservationEntityRepositoryImplementation constructor
   *
   * @param[in]   stochastics     Pointer to the stochastics class loaded by the framework
   * @param[in]   world           Pointer to the world
   * @param[in]   parameters      Pointer to the parameters of the module
   * @param[in]   callbacks       Pointer to the callbacks
   * @param[in]   dataBuffer      Pointer to the data buffer that provides read-only access to the data
   */
  ObservationEntityRepositoryImplementation(StochasticsInterface *stochastics,
                                            WorldInterface *world,
                                            const ParameterInterface *parameters,
                                            const CallbackInterface *callbacks,
                                            DataBufferReadInterface *dataBuffer);
  ObservationEntityRepositoryImplementation(const ObservationEntityRepositoryImplementation &) = delete;
  ObservationEntityRepositoryImplementation(ObservationEntityRepositoryImplementation &&) = delete;
  ObservationEntityRepositoryImplementation &operator=(const ObservationEntityRepositoryImplementation &) = delete;
  ObservationEntityRepositoryImplementation &operator=(ObservationEntityRepositoryImplementation &&) = delete;
  ~ObservationEntityRepositoryImplementation() override = default;

  void OpSimulationPreHook() override;
  void OpSimulationPreRunHook() override;
  void OpSimulationPostRunHook(const RunResultInterface &runResult) override;
  void OpSimulationUpdateHook(int, RunResultInterface &) override {}
  void OpSimulationPostHook() override;

private:
  /**
   * @brief Get the Entities from the databuffer as CSV row container
   *
   * @param entityBaseKey \see NON_PERSISTENT_ENTITIES or \see PERSISTENT_ENTITIES
   * @return All retrieved entities as individual entries
   */
  std::vector<CsvRow> GetEntities(const std::string &entityBaseKey);

  /**
   * @brief Helper method to retrieve the source attribute from an entity
   *
   * @param source_key   Attribute key within databuffer, e.g. 'Entities/Persistent/0/Source'
   * @return The contained value, or "<error>" if not available
   */
  std::string GetEntitySource(const std::string &source_key);

  /**
   * @brief Helper method to retreive the Metainfo provided by an entity
   *
   * @param metainfo_key Attribute key within databuffer, e.g. 'Entities/Persistent/0/Metainfo'
   * @return key/value string pairs containing the parsed metainfo (\see to_string)
   */
  Metainfo GetEntityMetainfo(const std::string &metainfo_key);

  /**
   * @brief Converts any static dataBuffer value into its string representation
   *
   * @param values value retrieved from the dataBuffer
   * @return converted string
   */
  const std::string to_string(const Values &values);

  /**
   * @brief Takes a CSV row container and writes it to a stream
   *
   * @param entities the entities as CSV rows (\see GetEntities)
   * @param stream   the open stream, where the data should be written to
   */
  void WriteEntities(const std::vector<CsvRow> &entities, std::stringstream &stream);

  DataBufferReadInterface *dataBuffer;
  int runNumber = 0;
  std::filesystem::path directory;
  std::string filenamePrefix{"Repository"};
  bool writePersitent{true};
  bool persistentInSeparateFile{false};
  std::vector<CsvRow> persistentEntities{};
};
