/********************************************************************************
 * Copyright (c) 2017-2022 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

//-----------------------------------------------------------------------------
/*!
 * \file  observation_logImplementation.h
 * \brief Adds the RunStatistic information to the simulation output.
 * \details  Writes the RunStatistic information into the simulation output.
 *           Also manages the stop reasons of the simulation.
 */
//-----------------------------------------------------------------------------

#pragma once

#include <map>
#include <string>
#include <utility>
#include <vector>

#include "include/observationInterface.h"
#include "observationCyclics.h"
#include "observationFileHandler.h"

class CallbackInterface;
class DataBufferReadInterface;
class ParameterInterface;
class RunResultInterface;
class StochasticsInterface;
class WorldInterface;
namespace openpass
{
namespace common
{
struct RuntimeInformation;
}  // namespace common
}  // namespace openpass

/*!
 * \brief Reads values of type double from databuffer for each agent
 *
 * DataStore is queried for key. Results are stored in the result map for each agent id where the key matches.
 *
 * \throws std::bad_variant_access if a key of other type than double is queried.
 *
 * \param[in]    key          Key to read at
 * \param[in]    dataBuffer   DataBuffer to read from
 * \param[out]   result       Storage for result
 */
extern void GetAgentValues(const std::string& key,
                           const DataBufferReadInterface* dataBuffer,
                           std::map<std::string, double>& result);

//-----------------------------------------------------------------------------
/** \brief This class adds the RunStatistic information to the simulation output.
 *   \details This class inherits the ObservationLogGeneric which creates the basic simulation output
 *            and adds the RunStatistic information to the output.
 *            This class also manages the stop reasons of the simulation.
 *
 *   \ingroup ObservationLog
 */
//-----------------------------------------------------------------------------
class ObservationLogImplementation : ObservationInterface
{
public:
  /// Name of the current component
  const std::string COMPONENTNAME = "ObservationLog";

  /**
   * @brief ObservationLogImplementation constructor
   *
   * @param[in]   stochastics     Pointer to the stochastics class loaded by the framework
   * @param[in]   world           Pointer to the world
   * @param[in]   parameters      Pointer to the parameters of the module
   * @param[in]   callbacks       Pointer to the callbacks
   * @param[in]   dataBuffer      Pointer to the data buffer that provides read-only access to the data
   */
  ObservationLogImplementation(StochasticsInterface* stochastics,
                               WorldInterface* world,
                               const ParameterInterface* parameters,
                               const CallbackInterface* callbacks,
                               DataBufferReadInterface* dataBuffer);
  ObservationLogImplementation(const ObservationLogImplementation&) = delete;
  ObservationLogImplementation(ObservationLogImplementation&&) = delete;
  ObservationLogImplementation& operator=(const ObservationLogImplementation&) = delete;
  ObservationLogImplementation& operator=(ObservationLogImplementation&&) = delete;
  ~ObservationLogImplementation() override = default;

  void OpSimulationPreHook() override;
  void OpSimulationPreRunHook() override;
  //-----------------------------------------------------------------------------
  //! Called by framework in simulation at each time step.
  //! Observation module can indicate end of simulation run here.
  //!
  //! @param[in]     time        Current scheduling time
  //! @param[in,out] runResult   Reference to run result
  //-----------------------------------------------------------------------------
  void OpSimulationUpdateHook(int time, [[maybe_unused]] RunResultInterface& runResult) override;
  void OpSimulationPostRunHook(const RunResultInterface& runResult) override;
  void OpSimulationPostHook() override;

private:
  const openpass::common::RuntimeInformation& runtimeInformation;
  DataBufferReadInterface* dataBuffer;
  ObservationFileHandler fileHandler;
  ObservationCyclics cyclics;
  ObservationCyclics cyclicsActiveComponent;
  Events events;
  std::vector<std::string> selectedColumns;
  std::vector<std::pair<std::string, std::string>> selectedRegexColumns;
};
