/********************************************************************************
 * Copyright (c) 2020 ITK Engineering GmbH
 *               2019 in-tech GmbH
 *			         2024 Volkswagen AG
 *               2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#pragma once

#include <cstdint>
#include <map>
#include <set>
#include <string>
#include <vector>

/**
 *  @brief The ObservationCyclics stores the samples which are logged by various modules
 *         and written into the cyclics tag of the simulationOutput.xml
 */
class ObservationCyclics
{
public:
  ObservationCyclics() {}

  ObservationCyclics(const ObservationCyclics&) = delete;
  ObservationCyclics(ObservationCyclics&&) = delete;
  ObservationCyclics& operator=(const ObservationCyclics&) = delete;
  ObservationCyclics& operator=(ObservationCyclics&&) = delete;

  ~ObservationCyclics() = default;

  /**
   * Inserts a parameter into the samples.
   * If the key does not exist yet, it inserts it into the list of keys
   * and all previous timesteps are filled with an empty string.
   * If the key exists, but some timesteps are missing for this key, then
   * missing timesteps are filled with an empty string.
   *
   * @param[in]     time       Timestep in milliseconds.
   * @param[in]     key        Name of parameter as string.
   * @param[in]     value      Value of parameter as string.
   */
  void Insert(int time, const std::string& key, const std::string& value);

  /**
   * @brief Returns the string header that is written into the simulationOuput.xml
   *
   * @return string header that is written into the simulationOuput.xml
   */
  std::string GetHeader();

  /**
   * @brief Returns the string header that is written into the Cyclics_*.csv including agentId
   *
   * @return string header that is written into the Cyclics_*.csv including agentId
   */
  std::string GetAgentHeader();

  /**
   * @brief Returns a single line of samples for the given timestep for writing into the simulationOutput
   *
   * @param[in] timeStepNumber    Number of timesteps
   * @return single line of samples
   */
  std::string GetSamplesLine(std::uint32_t timeStepNumber);

  /**
   * @brief Returns for each agent a single line of samples for the given timestep for writing into the cyclics_*.csv
   *
   * @param[in] timeStepNumber    Number of timesteps
   * @return single line of samples for each agent
   */
  std::vector<std::string> GetAgentSamplesLine(std::uint32_t timeStepNumber);

  /**
   * @brief Returns all timesteps for which samples exist
   *
   * @return timesteps
   */
  std::set<int>& GetTimeSteps() { return timeSteps; }

  /**
   * @brief Clears all samples for a new run
   */
  void Clear();

private:
  std::set<int> timeSteps;
  std::map<std::string, std::vector<std::string>> samples;
};
