/********************************************************************************
 * Copyright (c) 2018 AMFD GmbH
 *               2016-2018 ITK Engineering GmbH
 *               2017-2020 in-tech GmbH
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

/** \addtogroup Stochastics
* @{
* \brief implementation of StochasticsInterface
*
* \details The Stochastics module implements a StochasticsInterface which is used
* by all agents to create a random behavior.
*
* \section Stochastics_Inputs Inputs
* name | meaning
* -----|---------
* callbacks | CallbackInterface to send log information to the framework
*
* \section Stochastics_Outputs Outputs
 name | meaning
* -----|---------
* *StochasticsInterface | Provides a reference to an implementation of a StochasticsInterface to the framework.
*
* @see StochasticsInterface
*
* @} */

#pragma once

#include <cstdint>
#include <random>
#include <string>
#include <vector>

#include "include/callbackInterface.h"
#include "include/stochasticsInterface.h"

using BinomialDist = std::binomial_distribution<>;

/*!
 * \brief implementation of StochasticsInterface
 *
 * The Stochastics module implements a StochasticsInterface which is used
 * by all agents to create a random behavior.
 *
 * \ingroup Stochastics
 */
class StochasticsImplementation : public StochasticsInterface
{
public:
  /**
   * @brief StochasticsImplementation constructor
   *
   * @param[in]   callbacks       Pointer to the callbacks
   */
  StochasticsImplementation(const CallbackInterface *callbacks);
  StochasticsImplementation(const StochasticsImplementation &) = delete;
  StochasticsImplementation(StochasticsImplementation &&) = delete;
  StochasticsImplementation &operator=(const StochasticsImplementation &) = delete;
  StochasticsImplementation &operator=(StochasticsImplementation &&) = delete;
  ~StochasticsImplementation() override = default;

  double GetUniformDistributed(double a, double b) override;

  double GetNormalDistributed(double mean, double stdDeviation) override;

  int GetBinomialDistributed(int upperRangeNum, double probSuccess) override;

  double GetExponentialDistributed(double lambda) override;

  double GetGammaDistributed(double mean, double stdDeviation) override;

  double GetGammaDistributedShapeScale(double shape, double scale) override;

  double GetLogNormalDistributed(double mean, double stdDeviation) override;

  double GetLogNormalDistributedMuSigma(double mu, double sigma) override;

  double GetSpecialDistributed(std::string distributionName, std::vector<double> args) override;

  double GetRandomCdfLogNormalDistributed(double mean, double stdDeviation) override;

  double GetPercentileLogNormalDistributed(double mean, double stdDeviation, double probability) override;

  std::uint32_t GetRandomSeed() const override;

  void ReInit() override;

  void InitGenerator(std::uint32_t seed) override;

protected:
  /*! Provides callback to LOG() macro
   *
   * @param[in]     logLevel    Importance of log
   * @param[in]     file        Name of file where log is called
   * @param[in]     line        Line within file where log is called
   * @param[in]     message     Message to log
   */
  void Log(CbkLogLevel logLevel, const char *file, int line, const std::string &message)
  {
    if (callbacks)
    {
      callbacks->Log(logLevel, file, line, message);
    }
  }

private:
  std::uint32_t randomSeed = 0;

  std::mt19937 baseGenerator;
  std::uniform_real_distribution<double> uniformDistribution;
  BinomialDist binomialDistribution;
  std::normal_distribution<double> normalDistribution;
  std::exponential_distribution<double> exponentialDistribution;

  const CallbackInterface *callbacks;
};
