/********************************************************************************
 * Copyright (c) 2020 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include "JointsBuilder.h"

#include <MantleAPI/Common/floating_point_helper.h>
#include <cassert>
#include <cmath>
#include <cstddef>
#include <memory>
#include <type_traits>
#include <utility>

#include "common/vector2d.h"
#include "commonHelper.h"
#include "include/roadInterface/roadLaneInterface.h"

units::area::square_meter_t BorderPoints::GetSquaredError(const BorderPoints& begin, const BorderPoints& end) const
{
  assert(points.size() == begin.points.size());
  assert(points.size() == end.points.size());

  units::area::square_meter_t maxSquaredError = 0.0_sq_m;

  for (size_t i = 0; i < points.size(); i++)
  {
    const auto& currentPoint = points[i].point;
    const auto& startPoint = begin.points[i].point;
    const auto vector = end.points[i].point - startPoint;
    const auto projection = startPoint + vector * (vector.Dot(currentPoint - startPoint) / vector.Dot(vector));
    const auto difference = currentPoint - projection;
    const auto squaredError = difference.Dot(difference);
    maxSquaredError = std::max(maxSquaredError, squaredError);
  }
  return maxSquaredError;
}

JointsBuilder& JointsBuilder::CalculatePoints()
{
  for (const auto& [s, heading, currentBorderPoints] : sampledGeometry.borderPoints)
  {
    Joint joint;
    joint.s = s;

    for (size_t i = 0; i < currentBorderPoints.size(); i++)
    {
      const auto laneId = currentBorderPoints[i].lane->GetId();

      LaneJoint laneJoint;
      laneJoint.lane = currentBorderPoints[i].lane;
      if (laneId < 0)
      {
        laneJoint.left = currentBorderPoints[i - 1].point;
        laneJoint.right = currentBorderPoints[i].point;
        laneJoint.t_left = currentBorderPoints[i - 1].t;
        laneJoint.t_right = currentBorderPoints[i].t;
      }
      else if (laneId > 0)
      {
        laneJoint.left = currentBorderPoints[i].point;
        laneJoint.right = currentBorderPoints[i + 1].point;
        laneJoint.t_left = currentBorderPoints[i].t;
        laneJoint.t_right = currentBorderPoints[i + 1].t;
      }
      else
      {
        laneJoint.left = currentBorderPoints[i].point;
        laneJoint.right = currentBorderPoints[i].point;
        laneJoint.t_left = 0_m;
        laneJoint.t_right = 0_m;
      }
      laneJoint.center = (laneJoint.left + laneJoint.right) * 0.5;
      laneJoint.rawHeading = heading;
      joint.laneJoints.insert(std::make_pair(laneId, laneJoint));
    }
    joints.push_back(joint);
  }

  return *this;
}

JointsBuilder& JointsBuilder::CalculateHeadings()
{
  for (auto joint = joints.begin(); joint < joints.end(); joint++)
  {
    for (auto& [laneId, laneJoint] : joint->laneJoints)
    {
      if (joint == joints.end() - 1)
      {
        laneJoint.heading = (joints.end() - 2)->laneJoints.at(laneId).heading;
        continue;
      }
      auto& currentCenter = laneJoint.center;
      auto& nextCenter = (joint + 1)->laneJoints.at(laneId).center;
      auto vector = nextCenter - currentCenter;
      if (vector.y > 0_m)
      {
        laneJoint.heading = units::angle::radian_t(M_PI_2) - units::math::atan(vector.x / vector.y);
      }
      if (vector.y < 0_m)
      {
        laneJoint.heading = units::angle::radian_t(-M_PI_2) - units::math::atan(vector.x / vector.y);
      }
    }
  }

  return *this;
}

JointsBuilder& JointsBuilder::CalculateCurvatures()
{
  for (auto joint = joints.begin(); joint < joints.end(); joint++)
  {
    for (auto& [laneId, laneJoint] : joint->laneJoints)
    {
      units::angle::radian_t previousHeading;
      if (joint == joints.begin())
      {
        previousHeading = sampledGeometry.startHeading;
      }
      else
      {
        previousHeading = (joint - 1)->laneJoints.at(laneId).heading;
      }
      units::angle::radian_t nextHeading;
      if (joint == joints.end() - 1)
      {
        nextHeading = sampledGeometry.endHeading;
      }
      else
      {
        nextHeading = laneJoint.heading;
      }
      auto deltaHeading = CommonHelper::SetAngleToValidRange(nextHeading - previousHeading);
      units::length::meter_t deltaS;
      if (joint == joints.begin())
      {
        auto currentCenter = laneJoint.center;
        auto nextElementCenter = ((joint + 1)->laneJoints.at(laneId).center + laneJoint.center) * 0.5;
        deltaS = (nextElementCenter - currentCenter).Length();
      }
      else if (joint == joints.end() - 1)
      {
        auto previousElementCenter = ((joint - 1)->laneJoints.at(laneId).center + laneJoint.center) * 0.5;
        auto currentCenter = laneJoint.center;
        deltaS = (currentCenter - previousElementCenter).Length();
      }
      else
      {
        auto previousElementCenter = ((joint - 1)->laneJoints.at(laneId).center + laneJoint.center) * 0.5;
        auto nextElementCenter = ((joint + 1)->laneJoints.at(laneId).center + laneJoint.center) * 0.5;
        deltaS = (nextElementCenter - previousElementCenter).Length();
      }
      laneJoint.curvature = units::unit_cast<double>(deltaHeading) / deltaS;
    }
  }

  return *this;
}

Joints& JointsBuilder::GetJoints()
{
  return joints;
}

void SampledGeometry::Combine(SampledGeometry& other)
{
  if (other.borderPoints.size() < 2)
  {
    return;
  }

  assert(mantle_api::AlmostEqual(borderPoints.back().s, other.borderPoints.front().s));
  borderPoints.insert(borderPoints.end(), other.borderPoints.begin() + 1, other.borderPoints.end());
  endHeading = other.endHeading;
}
