/********************************************************************************
 * Copyright (c) 2021-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include "RoadStream.h"

#include <algorithm>
#include <cmath>
#include <stdexcept>
#include <string>
#include <vector>

#include "LaneStream.h"
#include "OWL/DataTypes.h"
#include "common/commonHelper.h"

const OWL::Interfaces::Section& GetSectionAtDistance(const OWL::Interfaces::Road& road, units::length::meter_t distance)
{
  for (const auto* section : road.GetSections())
  {
    if (section->Covers(distance))
    {
      return *section;
    }
  }
  throw std::runtime_error("RoadStream distance out of range");
}

units::length::meter_t GetLaneWidth(const OWL::Interfaces::Section& section,
                                    OWL::OdId laneId,
                                    units::length::meter_t distance)
{
  for (const auto* lane : section.GetLanes())
  {
    if (lane->GetOdId() == laneId)
    {
      return lane->GetWidth(distance);
    }
  }
  return 0.0_m;
}

StreamPosition RoadStream::GetStreamPosition(const GlobalRoadPosition& roadPosition) const
{
  StreamPosition streamPosition{-1_m, 0_m};

  for (const auto& element : elements)
  {
    if (element.road->GetId() == roadPosition.roadId)
    {
      streamPosition.s = element.GetStreamPosition(roadPosition.roadPosition.s);
      streamPosition.t = roadPosition.roadPosition.t;
      if (roadPosition.laneId < 0)
      {
        const auto& section = GetSectionAtDistance(*element.road, roadPosition.roadPosition.s);
        for (int laneId = -1; laneId > roadPosition.laneId; --laneId)
        {
          streamPosition.t -= GetLaneWidth(section, laneId, roadPosition.roadPosition.s);
        }
        streamPosition.t -= 0.5 * GetLaneWidth(section, roadPosition.laneId, roadPosition.roadPosition.s);
      }
      if (roadPosition.laneId > 0)
      {
        const auto& section = GetSectionAtDistance(*element.road, roadPosition.roadPosition.s);
        for (int laneId = 1; laneId < roadPosition.laneId; ++laneId)
        {
          streamPosition.t += GetLaneWidth(section, laneId, roadPosition.roadPosition.s);
        }
        streamPosition.t += 0.5 * GetLaneWidth(section, roadPosition.laneId, roadPosition.roadPosition.s);
      }
      streamPosition.t *= element.inStreamDirection ? 1 : -1;
      streamPosition.hdg = CommonHelper::SetAngleToValidRange(
          roadPosition.roadPosition.hdg + (element.inStreamDirection ? 0_rad : units::angle::radian_t(M_PI)));
      return streamPosition;
    }
  }
  return streamPosition;
}

GlobalRoadPosition RoadStream::GetRoadPosition(const StreamPosition& streamPosition) const
{
  GlobalRoadPosition globalRoadPosition;

  for (const auto& element : elements)
  {
    if (element.StartS() <= streamPosition.s && element.EndS() >= streamPosition.s)
    {
      globalRoadPosition.roadId = element.road->GetId();
      globalRoadPosition.roadPosition.s = element.GetElementPosition(streamPosition.s);
      const auto& section = GetSectionAtDistance(*element.road, globalRoadPosition.roadPosition.s);
      auto remainingOffset = streamPosition.t * (element.inStreamDirection ? 1 : -1);
      if (element.inStreamDirection xor (streamPosition.t > 0_m))
      {
        int laneId = -1;
        auto laneWidth = GetLaneWidth(section, laneId, globalRoadPosition.roadPosition.s);
        while (-remainingOffset > laneWidth)
        {
          remainingOffset += laneWidth;
          --laneId;
          laneWidth = GetLaneWidth(section, laneId, globalRoadPosition.roadPosition.s);
        }
        globalRoadPosition.laneId = laneId;
        globalRoadPosition.roadPosition.t = remainingOffset + 0.5 * laneWidth;
      }
      else
      {
        int laneId = 1;
        auto laneWidth = GetLaneWidth(section, laneId, globalRoadPosition.roadPosition.s);
        while (remainingOffset > laneWidth)
        {
          remainingOffset -= laneWidth;
          ++laneId;
          laneWidth = GetLaneWidth(section, laneId, globalRoadPosition.roadPosition.s);
        }
        globalRoadPosition.laneId = laneId;
        globalRoadPosition.roadPosition.t = remainingOffset - 0.5 * laneWidth;
      }
      globalRoadPosition.roadPosition.hdg = CommonHelper::SetAngleToValidRange(
          streamPosition.hdg + (element.inStreamDirection ? 0_rad : units::angle::radian_t(M_PI)));
      return globalRoadPosition;
    }
  }

  return globalRoadPosition;
}

bool LaneIsInDirection(const OWL::Interfaces::Lane* lane, bool odDirection)
{
  return odDirection ? (lane->GetOdId() < 0) : (lane->GetOdId() > 0);
}

std::vector<std::unique_ptr<LaneStreamInterface>> RoadStream::GetAllLaneStreams() const
{
  std::vector<std::unique_ptr<LaneStreamInterface>> laneStreams;
  OWL::Interfaces::Lanes const* lastLanes{nullptr};
  for (auto element : elements)
  {
    OWL::Interfaces::Sections sections = element.road->GetSections();
    if (!element.inStreamDirection)
    {
      std::reverse(sections.begin(), sections.end());
    }
    for (const auto* section : sections)
    {
      for (const auto* lane : section->GetLanes())
      {
        if (!LaneIsInDirection(lane, element.inStreamDirection))
        {
          continue;
        }
        bool hasPredecessorInStream = false;
        if (lastLanes)
        {
          auto predecessors = element.inStreamDirection ? lane->GetPrevious() : lane->GetNext();

          for (const auto& lastLane : *lastLanes)
          {
            if (std::find(predecessors.cbegin(), predecessors.cend(), lastLane->GetId()) != predecessors.cend())
            {
              hasPredecessorInStream = true;
              break;
            }
          }
        }
        if (hasPredecessorInStream)
        {
          continue;
        }
        StreamPosition start{element.GetStreamPosition(lane->GetDistance(OWL::MeasurementPoint::RoadStart)), 0.0_m};
        laneStreams.push_back(GetLaneStream(start, static_cast<int>(lane->GetOdId())));
      }
      lastLanes = &section->GetLanes();
    }
  }
  return laneStreams;
}

std::unique_ptr<LaneStreamInterface> RoadStream::GetLaneStream(const StreamPosition& startPosition,
                                                               const int laneId) const
{
  return std::make_unique<LaneStream>(CreateLaneStream(startPosition, laneId));
}

std::unique_ptr<LaneStreamInterface> RoadStream::GetLaneStream(const GlobalRoadPosition& startPosition) const
{
  return std::make_unique<LaneStream>(CreateLaneStream(startPosition));
}

std::vector<LaneStreamElement> CreateLaneStreamFromStartLane(const std::vector<RoadStreamElement>& elements,
                                                             std::vector<RoadStreamElement>::const_iterator& element,
                                                             OWL::Interfaces::Sections& sections,
                                                             OWL::Interfaces::Sections::iterator& section,
                                                             const OWL::Interfaces::Lane* const lane)
{
  std::vector<LaneStreamElement> laneStreamElements;
  const auto* currentLane = lane;
  while (currentLane)
  {
    LaneStreamElement newElement{};
    newElement.lane = currentLane;
    newElement.inStreamDirection = element->inStreamDirection;
    newElement.sOffset = element->GetStreamPosition(currentLane->GetDistance(OWL::MeasurementPoint::RoadStart));
    laneStreamElements.push_back(newElement);

    const auto successors = element->inStreamDirection ? currentLane->GetNext() : currentLane->GetPrevious();
    ++section;
    if (section == sections.cend())
    {
      ++element;
      if (element == elements.cend())
      {
        break;
      }
      sections = element->road->GetSections();
      if (!element->inStreamDirection)
      {
        std::reverse(sections.begin(), sections.end());
      }
      section = sections.begin();
    }
    for (const auto& nextLane : (*section)->GetLanes())
    {
      if (std::find(successors.cbegin(), successors.cend(), nextLane->GetId()) != successors.cend())
      {
        currentLane = nextLane;
        break;
      }
      currentLane = nullptr;
    }
  }
  return laneStreamElements;
}

std::vector<LaneStreamElement> RoadStream::CreateLaneStream(const StreamPosition& startPosition, const int laneId) const
{
  std::vector<LaneStreamElement> laneStreamElements;
  for (auto element = elements.begin(); element != elements.cend(); ++element)
  {
    if (element->StartS() > startPosition.s || element->EndS() < startPosition.s)
    {
      continue;
    }
    OWL::Interfaces::Sections sections = element->road->GetSections();
    if (!element->inStreamDirection)
    {
      std::reverse(sections.begin(), sections.end());
    }
    for (auto section = sections.begin(); section != sections.cend(); ++section)
    {
      if (!(*section)->Covers(element->GetElementPosition(startPosition.s)))
      {
        continue;
      }
      for (const auto& lane : (*section)->GetLanes())
      {
        if (lane->GetOdId() != laneId)
        {
          continue;
        }
        return CreateLaneStreamFromStartLane(elements, element, sections, section, lane);
      }
    }
  }
  return laneStreamElements;
}

std::vector<LaneStreamElement> RoadStream::CreateLaneStream(const GlobalRoadPosition& startPosition) const
{
  std::vector<LaneStreamElement> laneStreamElements;
  for (auto element = elements.begin(); element != elements.cend(); ++element)
  {
    if (element->road->GetId() != startPosition.roadId)
    {
      continue;
    }
    OWL::Interfaces::Sections sections = element->road->GetSections();
    if (!element->inStreamDirection)
    {
      std::reverse(sections.begin(), sections.end());
    }
    for (auto section = sections.begin(); section != sections.cend(); ++section)
    {
      if (!(*section)->Covers(startPosition.roadPosition.s))
      {
        continue;
      }
      for (const auto& lane : (*section)->GetLanes())
      {
        if (lane->GetOdId() != startPosition.laneId)
        {
          continue;
        }
        return CreateLaneStreamFromStartLane(elements, element, sections, section, lane);
      }
    }
  }
  return laneStreamElements;
}

units::length::meter_t RoadStream::GetLength() const
{
  return elements.back().EndS();
}
