/********************************************************************************
 * Copyright (c) 2019-2023 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/
#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <cfloat>
#include <utility>
#include <variant>
#include <vector>

#include "dontCare.h"
#include "fakeStochastics.h"
#include "sampler.h"
#include "stochasticDefinitions.h"

using ::testing::_;
using ::testing::DontCare;
using ::testing::Eq;
using ::testing::NiceMock;
using ::testing::Return;

TEST(Sampler_UnitTests, RollForReturnsTrue)
{
  NiceMock<FakeStochastics> fakeStochastic;
  ON_CALL(fakeStochastic, GetUniformDistributed(_, _)).WillByDefault(Return(0.5));

  ASSERT_THAT(Sampler::RollFor(0.5, &fakeStochastic), Eq(true));
}

TEST(Sampler_UnitTests, RollForReturnsFalse)
{
  NiceMock<FakeStochastics> fakeStochastic;
  ON_CALL(fakeStochastic, GetUniformDistributed(_, _)).WillByDefault(Return(0.6));

  ASSERT_THAT(Sampler::RollFor(0.5, &fakeStochastic), Eq(false));
}

TEST(Sampler_UnitTests, RollForWithChanceZeroReturnsFalse)
{
  NiceMock<FakeStochastics> fakeStochastic;
  ASSERT_THAT(Sampler::RollFor(0.0, &fakeStochastic), Eq(false));
}

TEST(Sampler_UnitTests, SampleIntProbabilityReturnThirdElement)
{
  std::vector<std::pair<int, double>> fakeProbabilities = {{0, 0.0}, {1, 0.3}, {2, 0.7}};

  NiceMock<FakeStochastics> fakeStochastic;
  ON_CALL(fakeStochastic, GetUniformDistributed(_, _)).WillByDefault(Return(0.4));

  int result = Sampler::Sample(fakeProbabilities, &fakeStochastic);

  ASSERT_EQ(result, 2);
}

TEST(Sampler_UnitTests, SampleNormalDistributionProbability_AppropriatelyHandlesZeroProbability)
{
  openpass::parameter::NormalDistribution impossibleParameter = DontCare<openpass::parameter::NormalDistribution>();
  openpass::parameter::NormalDistribution expectedParameter{
      DontCare<double>(), DontCare<double>(), DontCare<double>(), DontCare<double>()};
  std::vector<std::pair<openpass::parameter::StochasticDistribution, double>> fakeProbabilities{
      {impossibleParameter, DBL_EPSILON}, {expectedParameter, 1.0}};
  NiceMock<FakeStochastics> fakeStochastics;
  ON_CALL(fakeStochastics, GetUniformDistributed(_, _)).WillByDefault(Return(0.5));

  const auto sampledNormalDistributionParameter = Sampler::Sample(fakeProbabilities, &fakeStochastics);

  EXPECT_THAT(sampledNormalDistributionParameter, expectedParameter);
}
