/********************************************************************************
 * Copyright (c) 2019 AMFD GmbH
 *               2019-2020 in-tech GmbH
 *               2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <map>
#include <memory>
#include <optional>
#include <string>
#include <type_traits>
#include <units.h>
#include <utility>
#include <vector>

#include "Signals/sensorDriverSignal.h"
#include "Signals/sensor_driverDefinitions.h"
#include "common/globalDefinitions.h"
#include "common/vector2d.h"
#include "common/worldDefinitions.h"
#include "fakeAgent.h"
#include "fakeEgoAgent.h"
#include "fakePublisher.h"
#include "fakeStochastics.h"
#include "fakeWorld.h"
#include "fakeWorldObject.h"
#include "sensor_driverCalculations.h"
#include "sensor_driverImpl.h"

class SignalInterface;
class WorldObjectInterface;

using ::testing::_;
using ::testing::Eq;
using ::testing::NiceMock;
using ::testing::Return;
using ::testing::ReturnRef;
using ::testing::VariantWith;

TEST(SensorDriver_UnitTests, CorrectInformationInSignal)
{
  NiceMock<FakeAgent> fakeAgent;
  NiceMock<FakeEgoAgent> fakeEgoAgent;
  NiceMock<FakeWorld> fakeWorld;
  NiceMock<FakeStochastics> fakeStochastics;
  NiceMock<FakePublisher> fakePublisher;

  ON_CALL(fakeAgent, GetEgoAgent()).WillByDefault(ReturnRef(fakeEgoAgent));
  ON_CALL(fakeEgoAgent, GetAgent()).WillByDefault(Return(&fakeAgent));
  ON_CALL(fakeEgoAgent, HasValidRoute()).WillByDefault(Return(true));
  const std::string roadId = "SomeRoad";
  RoadGraph roadGraph;
  RoadGraphVertex start = add_vertex(RouteElement{roadId, true}, roadGraph);
  RoadGraphVertex target = add_vertex(roadGraph);
  RoadGraphEdge edge = add_edge(start, target, roadGraph).first;

  GlobalRoadPositions frontPosition{{roadId, GlobalRoadPosition{roadId, -2, 50_m, 4.0_m, 0.5_rad}}};
  ON_CALL(fakeAgent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillByDefault(ReturnRef(frontPosition));
  ON_CALL(fakeEgoAgent, GetReferencePointPosition()).WillByDefault(Return(frontPosition.at(roadId)));
  ON_CALL(fakeAgent, GetVelocity(_)).WillByDefault(Return(Common::Vector2d{2.0_mps, 0.0_mps}));
  ON_CALL(fakeAgent, GetAcceleration(_)).WillByDefault(Return(Common::Vector2d{3.0_mps_sq, 0.0_mps_sq}));
  ON_CALL(fakeEgoAgent, GetPositionLateral()).WillByDefault(Return(4.0_m));
  ON_CALL(fakeEgoAgent, GetRelativeYaw()).WillByDefault(Return(5.0_rad));
  ON_CALL(fakeEgoAgent, GetLaneRemainder(Side::Left)).WillByDefault(Return(6.0_m));
  ON_CALL(fakeEgoAgent, GetLaneRemainder(Side::Right)).WillByDefault(Return(7.0_m));
  std::vector<std::pair<ObjectTypeOSI, int>> collisionPartners = {{ObjectTypeOSI::Vehicle, 1}};
  ON_CALL(fakeAgent, GetCollisionPartners()).WillByDefault(Return(collisionPartners));
  ON_CALL(fakeAgent, GetRoads(_)).WillByDefault(Return(std::vector<std::string>{roadId}));
  ON_CALL(fakeEgoAgent, GetLaneCurvature(1)).WillByDefault(Return(0.5_i_m));
  ON_CALL(fakeEgoAgent, GetLaneCurvature(0)).WillByDefault(Return(0.6_i_m));
  ON_CALL(fakeEgoAgent, GetLaneCurvature(-1)).WillByDefault(Return(0.7_i_m));
  ON_CALL(fakeEgoAgent, GetLaneWidth(1)).WillByDefault(Return(5.0_m));
  ON_CALL(fakeEgoAgent, GetLaneWidth(0)).WillByDefault(Return(6.0_m));
  ON_CALL(fakeEgoAgent, GetLaneWidth(-1)).WillByDefault(Return(7.0_m));
  ON_CALL(fakeEgoAgent, GetDistanceToEndOfLane(_, 1)).WillByDefault(Return(100.0_m));
  ON_CALL(fakeEgoAgent, GetDistanceToEndOfLane(_, 0)).WillByDefault(Return(200.0_m));
  ON_CALL(fakeEgoAgent, GetDistanceToEndOfLane(_, -1)).WillByDefault(Return(300.0_m));
  CommonTrafficSign::Entity trafficSign;
  std::vector<CommonTrafficSign::Entity> trafficSigns{{trafficSign}};
  ON_CALL(fakeEgoAgent, GetTrafficSignsInRange(_, 1)).WillByDefault(Return(trafficSigns));
  ON_CALL(fakeEgoAgent, GetTrafficSignsInRange(_, 0)).WillByDefault(Return(trafficSigns));
  ON_CALL(fakeEgoAgent, GetTrafficSignsInRange(_, -1)).WillByDefault(Return(trafficSigns));
  LaneMarking::Entity laneMarking;
  std::vector<LaneMarking::Entity> laneMarkings{{laneMarking}};
  ON_CALL(fakeEgoAgent, GetLaneMarkingsInRange(_, _, _)).WillByDefault(Return(laneMarkings));
  RelativeWorldView::Lanes relativeLanes{{0.0_m,
                                          0.0_m,
                                          {{-1, true, LaneType::Driving, std::nullopt, std::nullopt},
                                           {0, true, LaneType::Driving, std::nullopt, std::nullopt},
                                           {1, true, LaneType::Driving, std::nullopt, std::nullopt}}}};
  ON_CALL(fakeEgoAgent, GetRelativeLanes(_, _, _)).WillByDefault(Return(relativeLanes));

  ON_CALL(fakeWorld, GetRoadGraph(RouteElement{roadId, true}, _, true))
      .WillByDefault(Return(std::make_pair(roadGraph, start)));
  ON_CALL(fakeWorld, IsDirectionalRoadExisting(roadId, true)).WillByDefault(Return(true));
  ON_CALL(fakeWorld, GetVisibilityDistance()).WillByDefault(Return(123.4_m));
  std::map<RoadGraphEdge, double> edgeWeights{{edge, 1.0}};
  ON_CALL(fakeWorld, GetEdgeWeights(_))
      .WillByDefault(
          [&edgeWeights](const RoadGraph &graph)
          {
            auto [firstEdge, edgeEnd] = edges(graph);
            return std::map<RoadGraphEdge, double>{{*firstEdge, 1.0}};
          });

  std::vector<const WorldObjectInterface *> noObjects{};
  EXPECT_CALL(fakeEgoAgent, GetObjectsInRange(_, _, _)).WillRepeatedly(Return(noObjects));

  NiceMock<FakeAgent> otherAgent;
  std::vector<const WorldObjectInterface *> objectsInFront{{&otherAgent}};
  EXPECT_CALL(fakeEgoAgent, GetObjectsInRange(0.0_m, _, 0)).WillRepeatedly(Return(objectsInFront));
  ON_CALL(otherAgent, GetId()).WillByDefault(Return(2));
  ON_CALL(fakeEgoAgent, GetNetDistance(&otherAgent)).WillByDefault(Return(50.0_m));
  ON_CALL(otherAgent, GetYaw()).WillByDefault(Return(0.1_rad));
  ON_CALL(otherAgent, GetLength()).WillByDefault(Return(1.0_m));
  ON_CALL(otherAgent, GetWidth()).WillByDefault(Return(1.1_m));
  ON_CALL(otherAgent, GetHeight()).WillByDefault(Return(1.2_m));
  ON_CALL(otherAgent, GetVelocity(_)).WillByDefault(Return(Common::Vector2d{10.0_mps, 0.0_mps}));
  ON_CALL(otherAgent, GetAcceleration(_)).WillByDefault(Return(Common::Vector2d{11.0_mps_sq, 0.0_mps_sq}));

  NiceMock<FakeWorldObject> trafficObject;
  std::vector<const WorldObjectInterface *> objectsBehind{{&trafficObject}};
  EXPECT_CALL(fakeEgoAgent, GetObjectsInRange(_, 0.0_m, 1)).WillRepeatedly(Return(objectsBehind));
  ON_CALL(trafficObject, GetId()).WillByDefault(Return(3));
  ON_CALL(fakeEgoAgent, GetNetDistance(&trafficObject)).WillByDefault(Return(60.0_m));
  ON_CALL(trafficObject, GetYaw()).WillByDefault(Return(0.2_rad));
  ON_CALL(trafficObject, GetLength()).WillByDefault(Return(2.0_m));
  ON_CALL(trafficObject, GetWidth()).WillByDefault(Return(2.1_m));
  ON_CALL(trafficObject, GetHeight()).WillByDefault(Return(2.2_m));

  SensorDriverImplementation sensorDriver(
      "SensorDriver", false, 0, 0, 0, 100, &fakeStochastics, &fakeWorld, nullptr, &fakePublisher, nullptr, &fakeAgent);

  sensorDriver.Trigger(0);
  std::shared_ptr<const SignalInterface> data;
  sensorDriver.UpdateOutput(0, data, 0);

  std::shared_ptr<const SensorDriverSignal> sensorDriverSignal
      = std::dynamic_pointer_cast<const SensorDriverSignal>(data);

  auto ownVehicleInformation = sensorDriverSignal->GetOwnVehicleInformation();
  EXPECT_THAT(ownVehicleInformation.absoluteVelocity.value(), Eq(2.0));
  EXPECT_THAT(ownVehicleInformation.acceleration.value(), Eq(3.0));
  EXPECT_THAT(ownVehicleInformation.lateralPosition.value(), Eq(4.0));
  EXPECT_THAT(ownVehicleInformation.heading.value(), Eq(5.0));
  EXPECT_THAT(ownVehicleInformation.distanceToLaneBoundaryLeft.value(), Eq(6.0));
  EXPECT_THAT(ownVehicleInformation.distanceToLaneBoundaryRight.value(), Eq(7.0));
  EXPECT_THAT(ownVehicleInformation.collision, Eq(true));

  auto geometryInformation = sensorDriverSignal->GetGeometryInformation();
  EXPECT_THAT(geometryInformation.visibilityDistance.value(), Eq(123.4));
  EXPECT_THAT(geometryInformation.laneEgo.exists, Eq(true));
  EXPECT_THAT(geometryInformation.laneEgo.curvature.value(), Eq(0.6));
  EXPECT_THAT(geometryInformation.laneEgo.width.value(), Eq(6.0));
  EXPECT_THAT(geometryInformation.laneEgo.distanceToEndOfLane.value(), Eq(200.0));
  EXPECT_THAT(geometryInformation.laneLeft.exists, Eq(true));
  EXPECT_THAT(geometryInformation.laneLeft.curvature.value(), Eq(0.5));
  EXPECT_THAT(geometryInformation.laneLeft.width.value(), Eq(5.0));
  EXPECT_THAT(geometryInformation.laneLeft.distanceToEndOfLane.value(), Eq(100.0));
  EXPECT_THAT(geometryInformation.laneRight.exists, Eq(true));
  EXPECT_THAT(geometryInformation.laneRight.curvature.value(), Eq(0.7));
  EXPECT_THAT(geometryInformation.laneRight.width.value(), Eq(7.0));
  EXPECT_THAT(geometryInformation.laneRight.distanceToEndOfLane.value(), Eq(300.0));

  auto trafficRuleInformation = sensorDriverSignal->GetTrafficRuleInformation();
  EXPECT_THAT(trafficRuleInformation.laneEgo.trafficSigns.size(), Eq(1));
  EXPECT_THAT(trafficRuleInformation.laneLeft.trafficSigns.size(), Eq(1));
  EXPECT_THAT(trafficRuleInformation.laneRight.trafficSigns.size(), Eq(1));

  auto surroundingObjects = sensorDriverSignal->GetSurroundingObjects();
  EXPECT_THAT(surroundingObjects.objectFront.exist, Eq(true));
  EXPECT_THAT(surroundingObjects.objectFront.isStatic, Eq(false));
  EXPECT_THAT(surroundingObjects.objectFront.id, Eq(2));
  EXPECT_THAT(surroundingObjects.objectFront.absoluteVelocity.value(), Eq(10.0));
  EXPECT_THAT(surroundingObjects.objectFront.acceleration.value(), Eq(11.0));
  EXPECT_THAT(surroundingObjects.objectFront.heading.value(), Eq(0.1));
  EXPECT_THAT(surroundingObjects.objectFront.length.value(), Eq(1.0));
  EXPECT_THAT(surroundingObjects.objectFront.width.value(), Eq(1.1));
  EXPECT_THAT(surroundingObjects.objectFront.height.value(), Eq(1.2));
  EXPECT_THAT(surroundingObjects.objectFront.relativeLongitudinalDistance.value(), Eq(50.0));
  // EXPECT_THAT(surroundingObjects.objectFront.relativeLateralDistance, DoubleEq(-2.0));
  EXPECT_THAT(surroundingObjects.objectRearLeft.exist, Eq(true));
  EXPECT_THAT(surroundingObjects.objectRearLeft.isStatic, Eq(true));
  EXPECT_THAT(surroundingObjects.objectRearLeft.id, Eq(3));
  EXPECT_THAT(surroundingObjects.objectRearLeft.heading.value(), Eq(0.2));
  EXPECT_THAT(surroundingObjects.objectRearLeft.length.value(), Eq(2.0));
  EXPECT_THAT(surroundingObjects.objectRearLeft.width.value(), Eq(2.1));
  EXPECT_THAT(surroundingObjects.objectRearLeft.height.value(), Eq(2.2));
  EXPECT_THAT(surroundingObjects.objectRearLeft.relativeLongitudinalDistance.value(), Eq(60.0));
  EXPECT_THAT(surroundingObjects.objectRear.exist, Eq(false));
  EXPECT_THAT(surroundingObjects.objectFrontLeft.exist, Eq(false));
  EXPECT_THAT(surroundingObjects.objectFrontRight.exist, Eq(false));
  EXPECT_THAT(surroundingObjects.objectRearRight.exist, Eq(false));
}

TEST(SensorDriverCalculations_UnitTests, GetLateralDistanceToObjectSameLane)
{
  NiceMock<FakeAgent> agent;
  FakeEgoAgent egoAgent;
  ON_CALL(egoAgent, GetLaneIdFromRelative(-1)).WillByDefault(Return(-4));
  ON_CALL(egoAgent, GetLaneIdFromRelative(0)).WillByDefault(Return(-3));
  ON_CALL(egoAgent, GetLaneIdFromRelative(1)).WillByDefault(Return(-2));
  ON_CALL(egoAgent, GetPositionLateral()).WillByDefault(Return(0.5_m));

  NiceMock<FakeAgent> otherAgent;
  GlobalRoadPositions otherAgentPosition{{"MyRoad", GlobalRoadPosition{"MyRoad", -3, 50_m, -1.0_m, 0_rad}}};
  ON_CALL(otherAgent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::Reference)))
      .WillByDefault(ReturnRef(otherAgentPosition));

  SensorDriverCalculations sensorDriverCalculations(egoAgent);
  auto lateralDistance = sensorDriverCalculations.GetLateralDistanceToObject("MyRoad", &otherAgent);
  ASSERT_THAT(lateralDistance.value(), Eq(-1.5));
}

TEST(SensorDriverCalculations_UnitTests, GetLateralDistanceToObjectLeftLane)
{
  NiceMock<FakeAgent> agent;
  FakeEgoAgent egoAgent;
  ON_CALL(egoAgent, GetLaneIdFromRelative(-1)).WillByDefault(Return(-4));
  ON_CALL(egoAgent, GetLaneIdFromRelative(0)).WillByDefault(Return(-3));
  ON_CALL(egoAgent, GetLaneIdFromRelative(1)).WillByDefault(Return(-2));
  GlobalRoadPositions position{{"MyRoad", GlobalRoadPosition{"", -3, 0_m, 0.5_m, 0_rad}}};
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillByDefault(ReturnRef(position));
  ON_CALL(egoAgent, GetPositionLateral()).WillByDefault(Return(0.5_m));
  ON_CALL(egoAgent, GetLaneWidth(0)).WillByDefault(Return(4.0_m));
  ON_CALL(egoAgent, GetLaneWidth(1)).WillByDefault(Return(5.0_m));

  NiceMock<FakeAgent> otherAgent;
  GlobalRoadPositions otherAgentPosition{{"MyRoad", GlobalRoadPosition{"MyRoad", -2, 50_m, -1.0_m, 0_rad}}};
  ON_CALL(otherAgent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::Reference)))
      .WillByDefault(ReturnRef(otherAgentPosition));

  SensorDriverCalculations sensorDriverCalculations(egoAgent);
  auto lateralDistance = sensorDriverCalculations.GetLateralDistanceToObject("MyRoad", &otherAgent);
  ASSERT_THAT(lateralDistance.value(), Eq(3.0));
}

TEST(SensorDriverCalculations_UnitTests, GetLateralDistanceToObjectRightLane)
{
  NiceMock<FakeAgent> agent;
  FakeEgoAgent egoAgent;
  ON_CALL(egoAgent, GetLaneIdFromRelative(-1)).WillByDefault(Return(-4));
  ON_CALL(egoAgent, GetLaneIdFromRelative(0)).WillByDefault(Return(-3));
  ON_CALL(egoAgent, GetLaneIdFromRelative(1)).WillByDefault(Return(-2));
  GlobalRoadPositions position{{"MyRoad", GlobalRoadPosition{"", -3, 0_m, 0.5_m, 0_rad}}};
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillByDefault(ReturnRef(position));
  ON_CALL(egoAgent, GetPositionLateral()).WillByDefault(Return(0.5_m));
  ON_CALL(egoAgent, GetLaneWidth(0)).WillByDefault(Return(4.0_m));
  ON_CALL(egoAgent, GetLaneWidth(-1)).WillByDefault(Return(5.0_m));

  NiceMock<FakeAgent> otherAgent;
  GlobalRoadPositions otherAgentPosition{{"MyRoad", GlobalRoadPosition{"MyRoad", -4, 50_m, -1.0_m, 0_rad}}};
  ON_CALL(otherAgent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::Reference)))
      .WillByDefault(ReturnRef(otherAgentPosition));

  SensorDriverCalculations sensorDriverCalculations(egoAgent);
  auto lateralDistance = sensorDriverCalculations.GetLateralDistanceToObject("MyRoad", &otherAgent);
  ASSERT_THAT(lateralDistance.value(), Eq(-6.0));
}
