/********************************************************************************
 * Copyright (c) 2019-2020 in-tech GmbH
 *               2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <algorithm>
#include <cstdint>
#include <functional>
#include <memory>
#include <string>
#include <units.h>
#include <utility>
#include <vector>

#include "common/SpawnerDefinitions.h"
#include "common/WorldAnalyzer.h"
#include "common/globalDefinitions.h"
#include "common/vector2d.h"
#include "common/worldDefinitions.h"
#include "dontCare.h"
#include "fakeAgent.h"
#include "fakeStream.h"
#include "fakeWorld.h"
#include "include/streamInterface.h"
#include "include/worldInterface.h"

class WorldObjectInterface;

using ::testing::_;
using ::testing::DontCare;
using ::testing::DoubleEq;
using ::testing::Return;
using ::testing::UnorderedElementsAreArray;
using ::testing::VariantWith;

namespace
{

static const auto NO_AGENTS_IN_RANGE = AgentInterfaces{};
static const auto NO_OBJECTS_IN_RANGE = std::vector<const WorldObjectInterface *>{};

static inline std::pair<RoadGraph, RoadGraphVertex> GetSingleVertexRoadGraph(const RouteElement &routeElement)
{
  RoadGraph roadGraph;
  auto vertex = add_vertex(routeElement, roadGraph);
  return {roadGraph, vertex};
}
}  // namespace

bool operator==(const RoadGraph &lhs, const RoadGraph &rhs)
{
  return get(RouteElement(), lhs, 0) == get(RouteElement(), rhs, 0);
}

TEST(WorldAnalyzer, GetValidLaneSpawningRanges_NoScenarioAgentsAllDrivingLanes_ReturnsFullRange)
{
  const RoadId roadId{"ROADID"};
  const auto [roadGraph, vertex] = GetSingleVertexRoadGraph(RouteElement{roadId, true});
  const LaneId laneId{-1};
  const SPosition sStart{0_m};
  const SPosition sEnd{100_m};
  const LaneSection fakeLaneSection{sStart, sEnd, {laneId}};
  const LaneSections fakeLaneSections{fakeLaneSection};
  const Ranges expectedValidRanges{{sStart, sEnd}};
  const RouteQueryResult<units::length::meter_t> endOfLaneResult{{vertex, sEnd}};

  const LaneTypes validLaneTypes = {LaneType::Driving};

  std::unique_ptr<FakeLaneStream> laneStream = std::make_unique<FakeLaneStream>();
  ON_CALL(*laneStream, GetLength()).WillByDefault(Return(10000_m));
  EXPECT_CALL(*laneStream, GetAgentsInRange(StreamPosition{sStart, 0_m}, StreamPosition{sEnd, 0_m}))
      .WillOnce(Return(NO_AGENTS_IN_RANGE));
  EXPECT_CALL(*laneStream, GetLaneTypes())
      .WillOnce(
          Return(std::vector<std::pair<units::length::meter_t, LaneType>>{std::make_pair(0_m, LaneType::Driving)}));

  FakeWorld fakeWorld;
  WorldAnalyzer worldAnalyzer(&fakeWorld, [](const auto &) {});

  const auto &actualResult
      = worldAnalyzer.GetValidLaneSpawningRanges(std::move(laneStream), sStart, sEnd, validLaneTypes);
  EXPECT_THAT(actualResult, expectedValidRanges);
}

TEST(WorldAnalyzer, GetValidLaneSpawningRanges_NoScenarioAgentsMixedLaneTypes_ReturnsDrivingLaneRanges)
{
  const RoadId roadId{"ROADID"};
  const auto [roadGraph, vertex] = GetSingleVertexRoadGraph(RouteElement{roadId, true});
  const LaneId laneId{-1};
  const SPosition sStart{10_m};
  const SPosition sEnd{100_m};
  const SPosition sLaneType1{0_m};
  const SPosition sLaneType2{50_m};
  const SPosition sLaneType3{90_m};
  const LaneSection fakeLaneSection{sStart, sEnd, {laneId}};
  const LaneSections fakeLaneSections{fakeLaneSection};
  const Ranges expectedValidRanges{{sStart, sLaneType2}, {sLaneType3, sEnd}};
  const RouteQueryResult<units::length::meter_t> endOfLaneResult{{vertex, sEnd}};

  const LaneTypes validLaneTypes = {LaneType::Driving};

  std::unique_ptr<FakeLaneStream> laneStream = std::make_unique<FakeLaneStream>();
  ON_CALL(*laneStream, GetLength()).WillByDefault(Return(10000_m));
  ON_CALL(*laneStream, GetAgentsInRange(_, _)).WillByDefault(Return(NO_AGENTS_IN_RANGE));
  std::vector<std::pair<units::length::meter_t, LaneType>> laneTypes{
      {sLaneType1, LaneType::Driving}, {sLaneType2, LaneType::None}, {sLaneType3, LaneType::Driving}};
  EXPECT_CALL(*laneStream, GetLaneTypes()).WillOnce(Return(laneTypes));

  FakeWorld fakeWorld;
  WorldAnalyzer worldAnalyzer(&fakeWorld, [](const auto &) {});

  const auto &actualResult
      = worldAnalyzer.GetValidLaneSpawningRanges(std::move(laneStream), sStart, sEnd, validLaneTypes);
  EXPECT_THAT(actualResult, expectedValidRanges);
}

struct GetValidLaneSpawningRanges_OneAgent_Data
{
  const SPosition sStart;
  const SPosition sEnd;

  const AgentCategory agentCategory;
  const Range scenarioAgentBounds;

  const std::vector<Range> expectedValidRanges;
};

class GetValidLaneSpawningRanges_OneAgent : public ::testing::TestWithParam<GetValidLaneSpawningRanges_OneAgent_Data>
{
public:
  GetValidLaneSpawningRanges_OneAgent() = default;

  FakeAgent fakeAgent;
  FakeWorld fakeWorld;
  std::unique_ptr<FakeLaneStream> laneStream = std::make_unique<FakeLaneStream>();
};

TEST_P(GetValidLaneSpawningRanges_OneAgent, GetValidLaneSpawningRanges)
{
  const auto data = GetParam();
  const LaneTypes validLaneTypes = {LaneType::Driving};

  ON_CALL(*laneStream, GetLength()).WillByDefault(Return(10000_m));
  EXPECT_CALL(*laneStream, GetAgentsInRange(StreamPosition{data.sStart, 0_m}, StreamPosition{data.sEnd, 0_m}))
      .WillOnce(Return(AgentInterfaces{&fakeAgent}));

  EXPECT_CALL(fakeAgent, GetAgentCategory()).WillRepeatedly(Return(data.agentCategory));

  EXPECT_CALL(*laneStream,
              GetStreamPosition(&fakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillRepeatedly(Return(StreamPosition{data.scenarioAgentBounds.first, 0._m}));

  EXPECT_CALL(*laneStream,
              GetStreamPosition(&fakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillRepeatedly(Return(StreamPosition{data.scenarioAgentBounds.second, 0._m}));

  EXPECT_CALL(*laneStream, GetLaneTypes())
      .WillOnce(
          Return(std::vector<std::pair<units::length::meter_t, LaneType>>{std::make_pair(0_m, LaneType::Driving)}));

  WorldAnalyzer worldAnalyzer{&fakeWorld, [](const auto &) {}};

  const auto &actualResult
      = worldAnalyzer.GetValidLaneSpawningRanges(std::move(laneStream), data.sStart, data.sEnd, validLaneTypes);

  Ranges expectedValidLaneRanges{};

  for (const auto &range : data.expectedValidRanges)
  {
    expectedValidLaneRanges.emplace_back(range.first, range.second);
  }

  EXPECT_THAT(actualResult, UnorderedElementsAreArray(expectedValidLaneRanges));
}

INSTANTIATE_TEST_CASE_P(
    WorldAnalyzer_AltersValidSpawnRangeCorrectly,
    GetValidLaneSpawningRanges_OneAgent,
    ::testing::Values(
        // sStart | sEnd |      agentCategory     |      bounds      |      expectedValidRanges |
        // single common agent does not impact valid ranges
        GetValidLaneSpawningRanges_OneAgent_Data{0_m, 100_m, AgentCategory::Common, {5_m, 10_m}, {{0_m, 100_m}}},
        // single ego agent renders agent bounds invalid as spawning points
        GetValidLaneSpawningRanges_OneAgent_Data{
            0_m, 100_m, AgentCategory::Ego, {5_m, 10_m}, {{0_m, 4.999_m}, {10.001_m, 100_m}}},
        // single scenario agent renders agent bounds invalid as spawning points
        GetValidLaneSpawningRanges_OneAgent_Data{
            0_m, 100_m, AgentCategory::Scenario, {5_m, 10_m}, {{0_m, 4.999_m}, {10.001_m, 100_m}}},
        // a range entirely encapsulated by a single agent is invalid
        GetValidLaneSpawningRanges_OneAgent_Data{25_m, 30_m, AgentCategory::Scenario, {20_m, 35_m}, {}},
        // a single ego agent outside the range leave full range valid
        GetValidLaneSpawningRanges_OneAgent_Data{0_m, 100_m, AgentCategory::Ego, {105_m, 110_m}, {{0_m, 100_m}}},
        // a single scenario agent outside the range leave full range valid
        GetValidLaneSpawningRanges_OneAgent_Data{0_m, 100_m, AgentCategory::Scenario, {105_m, 110_m}, {{0_m, 100_m}}}));

struct GetValidLaneSpawningRanges_TwoAgents_Data
{
  const SPosition sStart;
  const SPosition sEnd;

  const AgentCategory firstAgentCategory;
  const Range firstScenarioAgentBounds;

  const AgentCategory secondAgentCategory;
  const Range secondScenarioAgentBounds;

  const std::vector<Range> expectedValidRanges;
};

class GetValidLaneSpawningRanges_TwoAgents : public ::testing::TestWithParam<GetValidLaneSpawningRanges_TwoAgents_Data>
{
public:
  GetValidLaneSpawningRanges_TwoAgents() = default;

  FakeAgent firstFakeAgent;
  FakeAgent secondFakeAgent;
  FakeWorld fakeWorld;
  std::unique_ptr<FakeLaneStream> laneStream = std::make_unique<FakeLaneStream>();
};

TEST_P(GetValidLaneSpawningRanges_TwoAgents, GetValidLaneSpawningRanges)
{
  const auto data = GetParam();

  ON_CALL(*laneStream, GetLength()).WillByDefault(Return(10000_m));
  EXPECT_CALL(*laneStream, GetAgentsInRange(StreamPosition{data.sStart, 0_m}, StreamPosition{data.sEnd, 0_m}))
      .WillOnce(Return(AgentInterfaces{&firstFakeAgent, &secondFakeAgent}));

  EXPECT_CALL(firstFakeAgent, GetAgentCategory()).WillRepeatedly(Return(data.firstAgentCategory));

  EXPECT_CALL(*laneStream,
              GetStreamPosition(&firstFakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillRepeatedly(Return(StreamPosition{data.firstScenarioAgentBounds.first, 0._m}));

  EXPECT_CALL(
      *laneStream,
      GetStreamPosition(&firstFakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillRepeatedly(Return(StreamPosition{data.firstScenarioAgentBounds.second, 0._m}));

  EXPECT_CALL(secondFakeAgent, GetAgentCategory()).WillRepeatedly(Return(data.secondAgentCategory));

  EXPECT_CALL(
      *laneStream,
      GetStreamPosition(&secondFakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillRepeatedly(Return(StreamPosition{data.secondScenarioAgentBounds.first, 0._m}));

  EXPECT_CALL(
      *laneStream,
      GetStreamPosition(&secondFakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillRepeatedly(Return(StreamPosition{data.secondScenarioAgentBounds.second, 0._m}));

  EXPECT_CALL(*laneStream, GetLaneTypes())
      .WillOnce(
          Return(std::vector<std::pair<units::length::meter_t, LaneType>>{std::make_pair(0_m, LaneType::Driving)}));

  WorldAnalyzer worldAnalyzer{&fakeWorld, [](const auto &) {}};

  const LaneTypes validLaneTypes = {LaneType::Driving};

  const auto &actualResult
      = worldAnalyzer.GetValidLaneSpawningRanges(std::move(laneStream), data.sStart, data.sEnd, validLaneTypes);

  Ranges expectedValidLaneRanges{};

  for (const auto &range : data.expectedValidRanges)
  {
    expectedValidLaneRanges.emplace_back(range.first, range.second);
  }

  EXPECT_THAT(actualResult, UnorderedElementsAreArray(expectedValidLaneRanges));
}

INSTANTIATE_TEST_CASE_P(
    WorldAnalyzer_AltersValidSpawnRangeCorrectly,
    GetValidLaneSpawningRanges_TwoAgents,
    ::testing::Values(
        //sStart | sEnd |    firstAgentCategory  | firstScenarioAgentBounds | secondAgentCategory    | secondAgentBounds
        //| shouldHaveValidRanges | expectedValidRanges
        // only common agents -- full range should be valid
        GetValidLaneSpawningRanges_TwoAgents_Data{0_m,
                                                  100_m,
                                                  AgentCategory::Common,
                                                  DontCare<Range>(),
                                                  AgentCategory::Common,
                                                  DontCare<Range>(),
                                                  {{0_m, 100_m}}},
        // only a scenario agent and a common agent (the reverse will act as a one agent situation with the agent having
        // category common) only range outside of agent bounds should be valid
        GetValidLaneSpawningRanges_TwoAgents_Data{
            0_m,
            100_m,
            AgentCategory::Ego,
            {5_m, 10_m},
            AgentCategory::Common,
            {20_m, 25_m},
            {{0_m, 4.999_m},
             {10.001_m,
              100_m}}},  // the first value of further ranges are padded with .01 to avoid re-detecting the same agent
        GetValidLaneSpawningRanges_TwoAgents_Data{0_m,
                                                  100_m,
                                                  AgentCategory::Scenario,
                                                  {5_m, 10_m},
                                                  AgentCategory::Common,
                                                  {20_m, 25_m},
                                                  {{0_m, 4.999_m}, {10.001_m, 100_m}}},
        // two internal scenario agents
        // spawn range before rear of nearest agent and after front of furthest agent is valid
        GetValidLaneSpawningRanges_TwoAgents_Data{0_m,
                                                  100_m,
                                                  AgentCategory::Ego,
                                                  {5_m, 10_m},
                                                  AgentCategory::Scenario,
                                                  {25_m, 50_m},
                                                  {{0_m, 4.999_m}, {50.001_m, 100_m}}},
        GetValidLaneSpawningRanges_TwoAgents_Data{0_m,
                                                  100_m,
                                                  AgentCategory::Scenario,
                                                  {5_m, 10_m},
                                                  AgentCategory::Scenario,
                                                  {25_m, 50_m},
                                                  {{0_m, 4.999_m}, {50.001_m, 100_m}}},
        // two external scenario agents encapsualting spawn range -- no valid spawn range
        GetValidLaneSpawningRanges_TwoAgents_Data{
            10_m, 100_m, AgentCategory::Ego, {0_m, 5_m}, AgentCategory::Scenario, {100_m, 105_m}, {}},
        GetValidLaneSpawningRanges_TwoAgents_Data{
            10_m, 100_m, AgentCategory::Scenario, {0_m, 5_m}, AgentCategory::Scenario, {100_m, 105_m}, {}},
        // one internal scenario agent and one external (beyond range) -- partial valid spawn range
        GetValidLaneSpawningRanges_TwoAgents_Data{10_m,
                                                  100_m,
                                                  AgentCategory::Scenario,
                                                  {0_m, 5_m},
                                                  AgentCategory::Scenario,
                                                  {50_m, 55_m},
                                                  {{55.001_m, 100_m}}},
        // one internal scenario agent and one external (before range) -- partial valid spawn range
        GetValidLaneSpawningRanges_TwoAgents_Data{0_m,
                                                  75_m,
                                                  AgentCategory::Scenario,
                                                  {50_m, 55_m},
                                                  AgentCategory::Scenario,
                                                  {100_m, 105_m},
                                                  {{0_m, 49.999_m}}},
        // two external agents (outside of range - beyond) -- full valid range
        GetValidLaneSpawningRanges_TwoAgents_Data{
            0_m, 100_m, AgentCategory::Ego, {105_m, 110_m}, AgentCategory::Scenario, {125_m, 130_m}, {{0_m, 100_m}}},
        // two external agents (outside of range - before) -- full valid range
        GetValidLaneSpawningRanges_TwoAgents_Data{100_m,
                                                  200_m,
                                                  AgentCategory::Scenario,
                                                  {0_m, 5_m},
                                                  AgentCategory::Scenario,
                                                  {55_m, 60_m},
                                                  {{100_m, 200_m}}}));

struct CalculateSpawnVelocityToPreventCrashing_Data
{
  const units::length::meter_t opponentRearDistanceFromStartOfRoad;
  const units::velocity::meters_per_second_t opponentVelocity;

  const units::length::meter_t intendedSpawnPosition;
  const units::length::meter_t agentFrontLength;
  const units::length::meter_t agentRearLength;
  const units::velocity::meters_per_second_t intendedVelocity;
  const units::velocity::meters_per_second_t expectedAdjustedVelocity;
};

class CalculateSpawnVelocityToPreventCrashingTests
    : public ::testing::TestWithParam<CalculateSpawnVelocityToPreventCrashing_Data>
{
public:
  CalculateSpawnVelocityToPreventCrashingTests() = default;

  FakeWorld fakeWorld;
  FakeAgent fakeAgent;
};

TEST_P(CalculateSpawnVelocityToPreventCrashingTests, AdjustsVelocityToPreventCrashing)
{
  const auto data = GetParam();

  FakeAgent fakeAgent;
  FakeWorld fakeWorld;

  std::unique_ptr<FakeLaneStream> laneStream = std::make_unique<FakeLaneStream>();

  EXPECT_CALL(*laneStream, GetObjectsInRange(_, _))
      .WillOnce(Return(std::vector<const WorldObjectInterface *>{&fakeAgent}));

  EXPECT_CALL(*laneStream,
              GetStreamPosition(&fakeAgent, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillOnce(Return(StreamPosition{data.opponentRearDistanceFromStartOfRoad, 0._m}));
  EXPECT_CALL(fakeAgent, GetVelocity(_))
      .WillOnce(Return(Common::Vector2d{data.opponentVelocity, 0.0_mps}))
      .WillOnce(Return(Common::Vector2d{data.opponentVelocity, 0.0_mps}));

  const auto emptyLoggingCallback = [](const std::string &) {};
  WorldAnalyzer worldAnalyzer(&fakeWorld, emptyLoggingCallback);
  const auto actualAdjustedVelocity = worldAnalyzer.CalculateSpawnVelocityToPreventCrashing(std::move(laneStream),
                                                                                            data.intendedSpawnPosition,
                                                                                            data.agentFrontLength,
                                                                                            data.agentRearLength,
                                                                                            data.intendedVelocity);

  EXPECT_THAT(actualAdjustedVelocity.value(), DoubleEq(data.expectedAdjustedVelocity.value()));
}

INSTANTIATE_TEST_CASE_P(WorldAnalyzer_CalculateSpawnVelocityToPreventCrashing,
                        CalculateSpawnVelocityToPreventCrashingTests,
                        ::testing::Values(
                            // opponentRearDistanceFromStartOfRoad |  opponentVelocity | intendedSpawnPosition |
                            // agentFrontLength | agentRearLength | intendedVelocity | expectedAdjustedVelocity |
                            CalculateSpawnVelocityToPreventCrashing_Data{
                                60_m, 20_mps, 49.5_m, 0.5_m, 0.5_m, 30_mps, 25.0_mps}));

struct SpawnWillCauseCrash_Data
{
  const SPosition spawnPosition;
  const units::velocity::meters_per_second_t spawnVelocity;
  const Direction direction;
  const bool objectExistsInSearchDirection;
  const SPosition objectPosition;
  const units::velocity::meters_per_second_t objectVelocity;
  const bool expectedSpawnWillCauseCrashResult;
};

class SpawnWillCauseCrashTests : public ::testing::TestWithParam<SpawnWillCauseCrash_Data>
{
public:
  SpawnWillCauseCrashTests() = default;

  FakeWorld fakeWorld;
  FakeAgent fakeObjectInLane;
  std::unique_ptr<FakeLaneStream> laneStream = std::make_unique<FakeLaneStream>();
};

TEST_P(SpawnWillCauseCrashTests, PredictsCrashesAccurately)
{
  const auto data = GetParam();

  const SPosition sPosition{data.spawnPosition};
  const units::length::meter_t agentFrontLength{.5};
  const units::length::meter_t agentRearLength{.5};
  const units::velocity::meters_per_second_t velocity{data.spawnVelocity};

  if (!data.objectExistsInSearchDirection)
  {
    EXPECT_CALL(*laneStream, GetObjectsInRange(_, _)).WillOnce(Return(NO_OBJECTS_IN_RANGE));
  }
  else
  {
    std::vector<const WorldObjectInterface *> objectsInRange{&fakeObjectInLane};
    EXPECT_CALL(*laneStream, GetObjectsInRange(_, _)).WillOnce(Return(objectsInRange));

    EXPECT_CALL(fakeObjectInLane, GetVelocity(_)).WillOnce(Return(Common::Vector2d{data.objectVelocity, 0.0_mps}));

    if (data.direction == Direction::FORWARD)
    {
      EXPECT_CALL(
          *laneStream,
          GetStreamPosition(&fakeObjectInLane, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
          .WillOnce(Return(StreamPosition{data.objectPosition, 0._m}));
    }
    else
    {
      EXPECT_CALL(
          *laneStream,
          GetStreamPosition(&fakeObjectInLane, VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
          .WillOnce(Return(StreamPosition{data.objectPosition, 0._m}));
    }
  }

  const auto emptyLoggingCallback = [](const std::string &) {};
  WorldAnalyzer worldAnalyzer(&fakeWorld, emptyLoggingCallback);
  const auto actualSpawnWillCauseCrashResult = worldAnalyzer.SpawnWillCauseCrash(
      std::move(laneStream), sPosition, agentFrontLength, agentRearLength, velocity, data.direction);

  EXPECT_THAT(actualSpawnWillCauseCrashResult, data.expectedSpawnWillCauseCrashResult);
}

INSTANTIATE_TEST_CASE_P(
    WorldAnalyzer_SpawnWillCauseCrash,
    SpawnWillCauseCrashTests,
    ::testing::Values(
        // spawnPosition     |   spawnVelocity   |      direction     | objectExistsInSearchDirection | objectPosition
        // |   objectVelocity  | expectedSpawnWillCauseCrashResult
        SpawnWillCauseCrash_Data{DontCare<SPosition>(),
                                 DontCare<units::velocity::meters_per_second_t>(),
                                 Direction::FORWARD,
                                 false,
                                 DontCare<SPosition>(),
                                 DontCare<units::velocity::meters_per_second_t>(),
                                 false},
        SpawnWillCauseCrash_Data{0_m, 30.0_mps, Direction::FORWARD, true, 75.0_m, 0.0_mps, true},
        SpawnWillCauseCrash_Data{75.0_m, 0.0_mps, Direction::BACKWARD, true, 0.0_m, 30.0_mps, true}));
