/********************************************************************************
 * Copyright (c) 2021-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <cmath>
#include <memory>
#include <optional>
#include <string>
#include <units.h>
#include <utility>
#include <vector>

#include "DataTypes.h"
#include "LaneStream.h"
#include "RoadStream.h"
#include "common/worldDefinitions.h"
#include "commonHelper.h"
#include "fakeAgent.h"
#include "fakeLane.h"
#include "fakeMovingObject.h"
#include "fakeRoad.h"
#include "fakeSection.h"
#include "include/streamInterface.h"

using testing::_;
using testing::DoubleEq;
using testing::DoubleNear;
using testing::ElementsAre;
using testing::Eq;
using testing::IsEmpty;
using testing::Return;
using testing::ReturnRef;
using testing::SizeIs;
using testing::VariantWith;

TEST(RoadStream_Tests, GetStreamPosition_FirstRoadInStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad1{&section1};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sectionsRoad1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetStreamPosition(GlobalRoadPosition{"Road1", -2, 30._m, 1._m, 0._rad});

  ASSERT_THAT(result.s.value(), DoubleEq(30.));
  ASSERT_THAT(result.t.value(), DoubleEq(-5.5));
}

TEST(RoadStream_Tests, GetStreamPosition_FirstRoadAgainstStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad1{&section1};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sectionsRoad1));
  RoadStreamElement element1{&road1, lengthRoad1, false};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetStreamPosition(GlobalRoadPosition{"Road1", -2, 30._m, 1._m, 0._rad});

  ASSERT_THAT(result.s.value(), DoubleEq(80.));
  ASSERT_THAT(result.t.value(), DoubleEq(5.5));
}

TEST(RoadStream_Tests, GetStreamPosition_SecondRoadInStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad2{&section1};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sectionsRoad2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetStreamPosition(GlobalRoadPosition{"Road2", 2, 30._m, -1._m, 0._rad});

  ASSERT_THAT(result.s.value(), DoubleEq(140.));
  ASSERT_THAT(result.t.value(), DoubleEq(5.5));
}

TEST(RoadStream_Tests, GetStreamPosition_SecondRoadAgainstStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  RoadStreamElement element1{&road1, 0._m, true};
  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad2{&section1};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sectionsRoad2));
  RoadStreamElement element2{&road2, lengthRoad1 + lengthRoad2, false};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetStreamPosition(GlobalRoadPosition{"Road2", 2, 30._m, -1._m, 0._rad});

  ASSERT_THAT(result.s.value(), DoubleEq(290.));
  ASSERT_THAT(result.t.value(), DoubleEq(-5.5));
}

TEST(RoadStream_Tests, GetRoadPosition_FirstRoadInStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad1{&section1};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sectionsRoad1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetRoadPosition(StreamPosition{30._m, -5.5_m});

  ASSERT_THAT(result.roadId, Eq(idRoad1));
  ASSERT_THAT(result.laneId, Eq(-2));
  ASSERT_THAT(result.roadPosition.s.value(), DoubleEq(30.));
  ASSERT_THAT(result.roadPosition.t.value(), DoubleEq(1.));
}

TEST(RoadStream_Tests, GetRoadPosition_FirstRoadAgainstStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad1{&section1};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sectionsRoad1));
  RoadStreamElement element1{&road1, lengthRoad1, false};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetRoadPosition(StreamPosition{30._m, 5.5_m});

  ASSERT_THAT(result.roadId, Eq(idRoad1));
  ASSERT_THAT(result.laneId, Eq(-2));
  ASSERT_THAT(result.roadPosition.s.value(), DoubleEq(80.));
  ASSERT_THAT(result.roadPosition.t.value(), DoubleEq(1.));
}

TEST(RoadStream_Tests, GetRoadPosition_SecondRoadInStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad2{&section1};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sectionsRoad2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetRoadPosition(StreamPosition{130._m, 5.5_m});

  ASSERT_THAT(result.roadId, Eq(idRoad2));
  ASSERT_THAT(result.laneId, Eq(2));
  ASSERT_THAT(result.roadPosition.s.value(), DoubleEq(20.));
  ASSERT_THAT(result.roadPosition.t.value(), DoubleEq(-1.));
}

TEST(RoadStream_Tests, GetRoadPosition_SecondRoadAgainstStreamDirection)
{
  constexpr units::length::meter_t lengthRoad1 = 110_m;
  constexpr units::length::meter_t lengthRoad2 = 210_m;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault(Return(true));
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(1));
  ON_CALL(lane1, GetWidth(_)).WillByDefault(Return(4._m));
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(2));
  ON_CALL(lane2, GetWidth(_)).WillByDefault(Return(5._m));
  OWL::Interfaces::Lanes lanesSection1{&lane1, &lane2};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));
  OWL::Interfaces::Sections sectionsRoad2{&section1};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sectionsRoad2));
  RoadStreamElement element2{&road2, lengthRoad1 + lengthRoad2, false};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.GetRoadPosition(StreamPosition{130._m, -5.5_m});

  ASSERT_THAT(result.roadId, Eq(idRoad2));
  ASSERT_THAT(result.laneId, Eq(2));
  ASSERT_THAT(result.roadPosition.s.value(), DoubleEq(190.));
  ASSERT_THAT(result.roadPosition.t.value(), DoubleEq(-1.));
}

TEST(RoadStream_Tests, CreateLaneStream_WithStreamPosition_InStreamDirection)
{
  constexpr units::length::meter_t lengthSection1A = 40_m;
  constexpr units::length::meter_t lengthSection1B = 70_m;
  constexpr units::length::meter_t lengthRoad1 = lengthSection1A + lengthSection1B;
  constexpr units::length::meter_t lengthSection2A = 80_m;
  constexpr units::length::meter_t lengthSection2B = 130_m;
  constexpr units::length::meter_t lengthRoad2 = lengthSection2A + lengthSection2B;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";
  constexpr OWL::Id idLane2AMinus1 = 1;
  constexpr OWL::Id idLane2AMinus2 = 2;
  constexpr OWL::Id idLane2BMinus1 = 3;
  constexpr OWL::Id idOtherLane = 4;

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1A;
  ON_CALL(section1A, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthSection1A; });
  OWL::Fakes::Section section1B;
  ON_CALL(section1B, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= lengthSection1A && s <= lengthRoad1; });
  OWL::Fakes::Lane lane1BMinus1;
  ON_CALL(lane1BMinus1, GetOdId()).WillByDefault(Return(-1));
  OWL::Fakes::Lane lane1BMinus2;
  ON_CALL(lane1BMinus2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane1BMinus2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection1A));
  std::vector<OWL::Id> successorsLane1BMinus2{idOtherLane, idLane2AMinus2};
  ON_CALL(lane1BMinus2, GetNext()).WillByDefault(ReturnRef(successorsLane1BMinus2));
  OWL::Interfaces::Lanes lanesSection1B{&lane1BMinus1, &lane1BMinus2};
  ON_CALL(section1B, GetLanes()).WillByDefault(ReturnRef(lanesSection1B));
  std::vector<const OWL::Interfaces::Section*> sections1{&section1A, &section1B};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sections1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section2A;
  OWL::Fakes::Section section2B;
  OWL::Fakes::Lane lane2AMinus1;
  ON_CALL(lane2AMinus1, GetId()).WillByDefault(Return(idLane2AMinus1));
  ON_CALL(lane2AMinus1, GetOdId()).WillByDefault(Return(-1));
  OWL::Fakes::Lane lane2AMinus2;
  ON_CALL(lane2AMinus2, GetId()).WillByDefault(Return(idLane2AMinus2));
  ON_CALL(lane2AMinus2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2AMinus2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0_m));
  std::vector<OWL::Id> successorsLane2AMinus2{idLane2BMinus1};
  ON_CALL(lane2AMinus2, GetNext()).WillByDefault(ReturnRef(successorsLane2AMinus2));
  OWL::Interfaces::Lanes lanesSection2A{&lane2AMinus1, &lane2AMinus2};
  ON_CALL(section2A, GetLanes()).WillByDefault(ReturnRef(lanesSection2A));
  OWL::Fakes::Lane lane2BMinus1;
  ON_CALL(lane2BMinus1, GetId()).WillByDefault(Return(idLane2BMinus1));
  ON_CALL(lane2BMinus1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane2BMinus1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection2A));
  std::vector<OWL::Id> successorsLane2BMinus1{};
  ON_CALL(lane2BMinus1, GetNext()).WillByDefault(ReturnRef(successorsLane2BMinus1));
  OWL::Interfaces::Lanes lanesSection2B{&lane2BMinus1};
  ON_CALL(section2B, GetLanes()).WillByDefault(ReturnRef(lanesSection2B));
  std::vector<const OWL::Interfaces::Section*> sections2{&section2A, &section2B};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sections2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.CreateLaneStream(StreamPosition{50._m, 0._m}, -2);
  ASSERT_THAT(result, SizeIs(3));
  EXPECT_THAT(result.at(0).lane, Eq(&lane1BMinus2));
  EXPECT_THAT(result.at(0).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(0).sOffset, Eq(lengthSection1A));
  EXPECT_THAT(result.at(1).lane, Eq(&lane2AMinus2));
  EXPECT_THAT(result.at(1).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(1).sOffset, Eq(lengthRoad1));
  EXPECT_THAT(result.at(2).lane, Eq(&lane2BMinus1));
  EXPECT_THAT(result.at(2).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(2).sOffset, Eq(lengthRoad1 + lengthSection2A));
}

TEST(RoadStream_Tests, CreateLaneStream_WithStreamPosition_AgainstStreamDirection)
{
  constexpr units::length::meter_t lengthSection1A = 40_m;
  constexpr units::length::meter_t lengthSection1B = 70_m;
  constexpr units::length::meter_t lengthRoad1 = lengthSection1A + lengthSection1B;
  constexpr units::length::meter_t lengthSection2A = 80_m;
  constexpr units::length::meter_t lengthSection2B = 130_m;
  constexpr units::length::meter_t lengthRoad2 = lengthSection2A + lengthSection2B;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";
  constexpr OWL::Id idLane2AMinus1 = 1;
  constexpr OWL::Id idLane2AMinus2 = 2;
  constexpr OWL::Id idLane2BMinus1 = 3;
  constexpr OWL::Id idOtherLane = 4;

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1A;
  ON_CALL(section1A, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthSection1A; });
  OWL::Fakes::Section section1B;
  ON_CALL(section1B, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= lengthSection1A && s <= lengthRoad1; });
  OWL::Fakes::Lane lane1AMinus1;
  ON_CALL(lane1AMinus1, GetOdId()).WillByDefault(Return(-1));
  OWL::Fakes::Lane lane1AMinus2;
  ON_CALL(lane1AMinus2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane1AMinus2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  std::vector<OWL::Id> predecessorsLane1AMinus2{idOtherLane, idLane2BMinus1};
  ON_CALL(lane1AMinus2, GetPrevious()).WillByDefault(ReturnRef(predecessorsLane1AMinus2));
  OWL::Interfaces::Lanes lanesSection1A{&lane1AMinus1, &lane1AMinus2};
  ON_CALL(section1A, GetLanes()).WillByDefault(ReturnRef(lanesSection1A));
  std::vector<const OWL::Interfaces::Section*> sections1{&section1A, &section1B};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sections1));
  RoadStreamElement element1{&road1, lengthRoad1, false};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section2A;
  OWL::Fakes::Section section2B;
  OWL::Fakes::Lane lane2AMinus1;
  ON_CALL(lane2AMinus1, GetId()).WillByDefault(Return(idLane2AMinus1));
  ON_CALL(lane2AMinus1, GetOdId()).WillByDefault(Return(-1));
  OWL::Fakes::Lane lane2AMinus2;
  ON_CALL(lane2AMinus2, GetId()).WillByDefault(Return(idLane2AMinus2));
  ON_CALL(lane2AMinus2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2AMinus2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0_m));
  std::vector<OWL::Id> predecessorsLane2AMinus2{};
  ON_CALL(lane2AMinus2, GetPrevious()).WillByDefault(ReturnRef(predecessorsLane2AMinus2));
  OWL::Interfaces::Lanes lanesSection2A{&lane2AMinus1, &lane2AMinus2};
  ON_CALL(section2A, GetLanes()).WillByDefault(ReturnRef(lanesSection2A));
  OWL::Fakes::Lane lane2BMinus1;
  ON_CALL(lane2BMinus1, GetId()).WillByDefault(Return(idLane2BMinus1));
  ON_CALL(lane2BMinus1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane2BMinus1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection2A));
  std::vector<OWL::Id> predecessorsLane2BMinus1{idLane2AMinus2};
  ON_CALL(lane2BMinus1, GetPrevious()).WillByDefault(ReturnRef(predecessorsLane2BMinus1));
  OWL::Interfaces::Lanes lanesSection2B{&lane2BMinus1};
  ON_CALL(section2B, GetLanes()).WillByDefault(ReturnRef(lanesSection2B));
  std::vector<const OWL::Interfaces::Section*> sections2{&section2A, &section2B};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sections2));
  RoadStreamElement element2{&road2, lengthRoad1 + lengthRoad2, false};

  RoadStream roadStream{{element1, element2}};

  const auto result = roadStream.CreateLaneStream(StreamPosition{90._m, 0._m}, -2);
  ASSERT_THAT(result, SizeIs(3));
  EXPECT_THAT(result.at(0).lane, Eq(&lane1AMinus2));
  EXPECT_THAT(result.at(0).inStreamDirection, Eq(false));
  EXPECT_THAT(result.at(0).sOffset, Eq(lengthRoad1));
  EXPECT_THAT(result.at(1).lane, Eq(&lane2BMinus1));
  EXPECT_THAT(result.at(1).inStreamDirection, Eq(false));
  EXPECT_THAT(result.at(1).sOffset, Eq(lengthRoad1 + lengthSection2B));
  EXPECT_THAT(result.at(2).lane, Eq(&lane2AMinus2));
  EXPECT_THAT(result.at(2).inStreamDirection, Eq(false));
  EXPECT_THAT(result.at(2).sOffset, Eq(lengthRoad1 + lengthRoad2));
}

TEST(RoadStream_Tests, CreateLaneStream_WithGlobalRoadPosition_OnFirstRoad)
{
  constexpr units::length::meter_t lengthSection1A = 40_m;
  constexpr units::length::meter_t lengthSection1B = 70_m;
  constexpr units::length::meter_t lengthRoad1 = lengthSection1A + lengthSection1B;
  constexpr units::length::meter_t lengthSection2A = 80_m;
  constexpr units::length::meter_t lengthSection2B = 130_m;
  constexpr units::length::meter_t lengthRoad2 = lengthSection2A + lengthSection2B;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";
  constexpr OWL::Id idLane2AMinus1 = 1;
  constexpr OWL::Id idLane2APlus1 = 2;
  constexpr OWL::Id idLane2APlus2 = 3;
  constexpr OWL::Id idLane2BPlus1 = 4;
  constexpr OWL::Id idOtherLane = 5;

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1A;
  ON_CALL(section1A, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthSection1A; });
  OWL::Fakes::Section section1B;
  ON_CALL(section1B, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= lengthSection1A && s <= lengthRoad1; });
  OWL::Fakes::Lane lane1BMinus1;
  ON_CALL(lane1BMinus1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane1BMinus1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection1A));
  std::vector<OWL::Id> successorsLane1BMinus1{idOtherLane, idLane2APlus1};
  ON_CALL(lane1BMinus1, GetNext()).WillByDefault(ReturnRef(successorsLane1BMinus1));
  OWL::Fakes::Lane lane1BMinus2;
  ON_CALL(lane1BMinus2, GetOdId()).WillByDefault(Return(-2));
  OWL::Interfaces::Lanes lanesSection1B{&lane1BMinus1, &lane1BMinus2};
  ON_CALL(section1B, GetLanes()).WillByDefault(ReturnRef(lanesSection1B));
  std::vector<const OWL::Interfaces::Section*> sections1{&section1A, &section1B};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sections1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section2A;
  OWL::Fakes::Section section2B;
  OWL::Fakes::Lane lane2AMinus1;
  ON_CALL(lane2AMinus1, GetId()).WillByDefault(Return(idLane2AMinus1));
  ON_CALL(lane2AMinus1, GetOdId()).WillByDefault(Return(-1));
  OWL::Fakes::Lane lane2APlus1;
  ON_CALL(lane2APlus1, GetId()).WillByDefault(Return(idLane2APlus1));
  ON_CALL(lane2APlus1, GetOdId()).WillByDefault(Return(1));
  ON_CALL(lane2APlus1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0_m));
  std::vector<OWL::Id> successorsLane2APlus1{idLane2BPlus1};
  ON_CALL(lane2APlus1, GetNext()).WillByDefault(ReturnRef(successorsLane2APlus1));
  OWL::Fakes::Lane lane2APlus2;
  ON_CALL(lane2APlus2, GetId()).WillByDefault(Return(idLane2APlus2));
  ON_CALL(lane2APlus2, GetOdId()).WillByDefault(Return(2));
  OWL::Interfaces::Lanes lanesSection2A{&lane2AMinus1, &lane2APlus1, &lane2APlus2};
  ON_CALL(section2A, GetLanes()).WillByDefault(ReturnRef(lanesSection2A));
  OWL::Fakes::Lane lane2BPlus1;
  ON_CALL(lane2BPlus1, GetId()).WillByDefault(Return(idLane2BPlus1));
  ON_CALL(lane2BPlus1, GetOdId()).WillByDefault(Return(1));
  ON_CALL(lane2BPlus1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection2A));
  std::vector<OWL::Id> successorsLane2BPlus1{};
  ON_CALL(lane2BPlus1, GetNext()).WillByDefault(ReturnRef(successorsLane2BPlus1));
  OWL::Interfaces::Lanes lanesSection2B{&lane2BPlus1};
  ON_CALL(section2B, GetLanes()).WillByDefault(ReturnRef(lanesSection2B));
  std::vector<const OWL::Interfaces::Section*> sections2{&section2A, &section2B};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sections2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  GlobalRoadPosition roadPosition{idRoad1, -1, lengthRoad1, 0_m, 0_rad};

  const auto result = roadStream.CreateLaneStream(roadPosition);
  ASSERT_THAT(result, SizeIs(3));
  EXPECT_THAT(result.at(0).lane, Eq(&lane1BMinus1));
  EXPECT_THAT(result.at(0).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(0).sOffset, Eq(lengthSection1A));
  EXPECT_THAT(result.at(1).lane, Eq(&lane2APlus1));
  EXPECT_THAT(result.at(1).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(1).sOffset, Eq(lengthRoad1));
  EXPECT_THAT(result.at(2).lane, Eq(&lane2BPlus1));
  EXPECT_THAT(result.at(2).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(2).sOffset, Eq(lengthRoad1 + lengthSection2A));
}

TEST(RoadStream_Tests, CreateLaneStream_WithGlobalRoadPosition_OnSecondRoad)
{
  constexpr units::length::meter_t lengthSection1A = 40_m;
  constexpr units::length::meter_t lengthSection1B = 70_m;
  constexpr units::length::meter_t lengthRoad1 = lengthSection1A + lengthSection1B;
  constexpr units::length::meter_t lengthSection2A = 80_m;
  constexpr units::length::meter_t lengthSection2B = 130_m;
  constexpr units::length::meter_t lengthRoad2 = lengthSection2A + lengthSection2B;
  const std::string idRoad1 = "Road1";
  const std::string idRoad2 = "Road2";
  constexpr OWL::Id idLane2AMinus1 = 1;
  constexpr OWL::Id idLane2APlus1 = 2;
  constexpr OWL::Id idLane2APlus2 = 3;
  constexpr OWL::Id idLane2BPlus1 = 4;
  constexpr OWL::Id idOtherLane = 5;

  OWL::Fakes::Road road1;
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  ON_CALL(road1, GetLength()).WillByDefault(Return(lengthRoad1));
  OWL::Fakes::Section section1A;
  OWL::Fakes::Section section1B;
  OWL::Fakes::Lane lane1BMinus1;
  ON_CALL(lane1BMinus1, GetOdId()).WillByDefault(Return(-1));
  OWL::Fakes::Lane lane1BMinus2;
  ON_CALL(lane1BMinus2, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane1BMinus2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection1A));
  std::vector<OWL::Id> successorsLane1BMinus2{idOtherLane, idLane2APlus2};
  ON_CALL(lane1BMinus2, GetNext()).WillByDefault(ReturnRef(successorsLane1BMinus2));
  OWL::Interfaces::Lanes lanesSection1B{&lane1BMinus1, &lane1BMinus2};
  ON_CALL(section1B, GetLanes()).WillByDefault(ReturnRef(lanesSection1B));
  std::vector<const OWL::Interfaces::Section*> sections1{&section1A, &section1B};
  ON_CALL(road1, GetSections()).WillByDefault(ReturnRef(sections1));
  RoadStreamElement element1{&road1, 0._m, true};

  OWL::Fakes::Road road2;
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  ON_CALL(road2, GetLength()).WillByDefault(Return(lengthRoad2));
  OWL::Fakes::Section section2A;
  ON_CALL(section2A, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthSection2A; });
  OWL::Fakes::Section section2B;
  ON_CALL(section2B, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= lengthSection2A && s <= lengthRoad2; });
  OWL::Fakes::Lane lane2AMinus1;
  ON_CALL(lane2AMinus1, GetId()).WillByDefault(Return(idLane2AMinus1));
  ON_CALL(lane2AMinus1, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane2AMinus1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0_m));
  std::vector<OWL::Id> successorsLane2AMinus1{};
  ON_CALL(lane2AMinus1, GetNext()).WillByDefault(ReturnRef(successorsLane2AMinus1));
  OWL::Fakes::Lane lane2APlus1;
  ON_CALL(lane2APlus1, GetId()).WillByDefault(Return(idLane2APlus1));
  ON_CALL(lane2APlus1, GetOdId()).WillByDefault(Return(1));
  OWL::Fakes::Lane lane2APlus2;
  ON_CALL(lane2APlus2, GetId()).WillByDefault(Return(idLane2APlus2));
  ON_CALL(lane2APlus2, GetOdId()).WillByDefault(Return(2));
  OWL::Interfaces::Lanes lanesSection2A{&lane2AMinus1, &lane2APlus1, &lane2APlus2};
  ON_CALL(section2A, GetLanes()).WillByDefault(ReturnRef(lanesSection2A));
  OWL::Fakes::Lane lane2BPlus1;
  ON_CALL(lane2BPlus1, GetId()).WillByDefault(Return(idLane2BPlus1));
  ON_CALL(lane2BPlus1, GetOdId()).WillByDefault(Return(1));
  OWL::Interfaces::Lanes lanesSection2B{&lane2BPlus1};
  ON_CALL(section2B, GetLanes()).WillByDefault(ReturnRef(lanesSection2B));
  std::vector<const OWL::Interfaces::Section*> sections2{&section2A, &section2B};
  ON_CALL(road2, GetSections()).WillByDefault(ReturnRef(sections2));
  RoadStreamElement element2{&road2, lengthRoad1, true};

  RoadStream roadStream{{element1, element2}};

  GlobalRoadPosition roadPosition{idRoad2, -1, 0_m, 0_m, 0_rad};

  const auto result = roadStream.CreateLaneStream(roadPosition);
  ASSERT_THAT(result, SizeIs(1));
  EXPECT_THAT(result.at(0).lane, Eq(&lane2AMinus1));
  EXPECT_THAT(result.at(0).inStreamDirection, Eq(true));
  EXPECT_THAT(result.at(0).sOffset, Eq(lengthRoad1));
}

TEST(RoadStream_Tests, GetAllLaneStreams_SingleRoadWithTwoSections)
{
  constexpr units::length::meter_t lengthSection1 = 40_m;
  constexpr units::length::meter_t lengthSection2 = 70_m;
  constexpr OWL::Id idLane1A = 10;
  constexpr OWL::Id idLane1B = 11;
  constexpr OWL::Id idLane2A = 12;
  constexpr OWL::Id idLane2B = 13;

  std::vector<OWL::Id> emptyIds{};
  OWL::Fakes::Road road;

  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthSection1; });
  OWL::Fakes::Lane lane1A;
  ON_CALL(lane1A, GetId()).WillByDefault(Return(idLane1A));
  ON_CALL(lane1A, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane1A, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane1A, GetNext()).WillByDefault(ReturnRef(emptyIds));
  ON_CALL(lane1A, GetPrevious()).WillByDefault(ReturnRef(emptyIds));
  OWL::Fakes::Lane lane1B;
  ON_CALL(lane1B, GetId()).WillByDefault(Return(idLane1B));
  ON_CALL(lane1B, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane1B, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  std::vector<OWL::Id> successorsLane1B{idLane2A};
  ON_CALL(lane1B, GetNext()).WillByDefault(ReturnRef(successorsLane1B));
  ON_CALL(lane1B, GetPrevious()).WillByDefault(ReturnRef(emptyIds));
  OWL::Interfaces::Lanes lanesSection1{&lane1A, &lane1B};
  ON_CALL(section1, GetLanes()).WillByDefault(ReturnRef(lanesSection1));

  OWL::Fakes::Section section2;
  ON_CALL(section2, Covers(_))
      .WillByDefault([&](units::length::meter_t s)
                     { return s >= lengthSection1 && s <= lengthSection1 + lengthSection2; });
  OWL::Fakes::Lane lane2A;
  ON_CALL(lane2A, GetId()).WillByDefault(Return(idLane2A));
  ON_CALL(lane2A, GetOdId()).WillByDefault(Return(-1));
  ON_CALL(lane2A, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection1));
  ON_CALL(lane2A, GetNext()).WillByDefault(ReturnRef(emptyIds));
  std::vector<OWL::Id> predecessorsLane2A{idLane1B};
  ON_CALL(lane2A, GetPrevious()).WillByDefault(ReturnRef(predecessorsLane2A));
  OWL::Fakes::Lane lane2B;
  ON_CALL(lane2B, GetId()).WillByDefault(Return(idLane2B));
  ON_CALL(lane2B, GetOdId()).WillByDefault(Return(-2));
  ON_CALL(lane2B, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthSection1));
  ON_CALL(lane2B, GetNext()).WillByDefault(ReturnRef(emptyIds));
  ON_CALL(lane2B, GetPrevious()).WillByDefault(ReturnRef(emptyIds));
  OWL::Interfaces::Lanes lanesSection2{&lane2A, &lane2B};
  ON_CALL(section2, GetLanes()).WillByDefault(ReturnRef(lanesSection2));

  OWL::Interfaces::Sections sections{&section1, &section2};
  ON_CALL(road, GetSections()).WillByDefault(ReturnRef(sections));
  RoadStreamElement element{&road, 0._m, true};

  RoadStream roadStream{{element}};

  const auto result = roadStream.GetAllLaneStreams();

  ASSERT_THAT(result, SizeIs(3));
}

TEST(LaneStream_Tests, GetAgentsInRange_EmptyLanes)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  OWL::Interfaces::LaneAssignments emptyObjects{};
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetLength()).WillByDefault(Return(lengthLane1));
  ON_CALL(lane1, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element1{&lane1, 0_m, true};
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetLength()).WillByDefault(Return(lengthLane2));
  ON_CALL(lane2, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element2{&lane2, lengthLane1, true};

  LaneStream laneStream{{element1, element2}};

  auto result = laneStream.GetAgentsInRange({0_m, 0_m}, {210_m, 0_m});
  ASSERT_THAT(result, IsEmpty());
}

TEST(LaneStream_Tests, GetAgentsInRange_InStreamDirection)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  OWL::Fakes::MovingObject moving1;
  OWL::Fakes::MovingObject moving2;
  OWL::Fakes::MovingObject moving3;
  OWL::Fakes::MovingObject moving4;
  OWL::Fakes::StationaryObject stationary;
  FakeAgent agent1;
  FakeAgent agent2;
  FakeAgent agent3;
  FakeAgent agent4;
  moving1.SetLinkedObjectForTesting(&agent1);
  moving2.SetLinkedObjectForTesting(&agent2);
  moving3.SetLinkedObjectForTesting(&agent3);
  moving4.SetLinkedObjectForTesting(&agent4);
  OWL::LaneOverlap overlap1{GlobalRoadPosition{"", 0, 10_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 15_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 12_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 12_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap2{GlobalRoadPosition{"", 0, 40_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 45_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 42_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 42_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap3{GlobalRoadPosition{"", 0, 60_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 65_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 62_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 62_m, 0_m, 0_rad}};
  OWL::Interfaces::LaneAssignments objectsLane1{{overlap1, &moving1}, {overlap2, &moving2}, {overlap3, &stationary}};
  OWL::LaneOverlap overlap4{GlobalRoadPosition{"", 0, 100_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 105_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 102_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 102_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap5{GlobalRoadPosition{"", 0, 130_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 135_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 132_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 132_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap6{GlobalRoadPosition{"", 0, 150_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 155_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 152_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 152_m, 0_m, 0_rad}};
  OWL::Interfaces::LaneAssignments objectsLane2{{overlap4, &moving2}, {overlap5, &moving3}, {overlap6, &moving4}};
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetLength()).WillByDefault(Return(lengthLane1));
  ON_CALL(lane1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane1, GetWorldObjects(true)).WillByDefault(ReturnRef(objectsLane1));
  LaneStreamElement element1{&lane1, 0_m, true};
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetLength()).WillByDefault(Return(lengthLane2));
  ON_CALL(lane2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthLane1));
  ON_CALL(lane2, GetWorldObjects(true)).WillByDefault(ReturnRef(objectsLane2));
  LaneStreamElement element2{&lane2, lengthLane1, true};

  LaneStream laneStream{{element1, element2}};

  auto result = laneStream.GetAgentsInRange({20_m, 0_m}, {140_m, 0_m});
  ASSERT_THAT(result, ElementsAre(&agent2, &agent3));
}

TEST(LaneStream_Tests, GetAgentsInRange_AgainstStreamDirection)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  OWL::Fakes::MovingObject moving1;
  OWL::Fakes::MovingObject moving2;
  OWL::Fakes::MovingObject moving3;
  OWL::Fakes::MovingObject moving4;
  OWL::Fakes::StationaryObject stationary;
  FakeAgent agent1;
  FakeAgent agent2;
  FakeAgent agent3;
  FakeAgent agent4;
  moving1.SetLinkedObjectForTesting(&agent1);
  moving2.SetLinkedObjectForTesting(&agent2);
  moving3.SetLinkedObjectForTesting(&agent3);
  moving4.SetLinkedObjectForTesting(&agent4);
  OWL::LaneOverlap overlap1{GlobalRoadPosition{"", 0, 80_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 85_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 82_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 82_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap2{GlobalRoadPosition{"", 0, 60_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 65_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 62_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 62_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap3{GlobalRoadPosition{"", 0, 10_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 15_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 12_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 12_m, 0_m, 0_rad}};
  OWL::Interfaces::LaneAssignments objectsLane1{{overlap1, &moving1}, {overlap2, &stationary}, {overlap3, &moving2}};
  OWL::LaneOverlap overlap4{GlobalRoadPosition{"", 0, 170_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 175_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 172_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 172_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap5{GlobalRoadPosition{"", 0, 130_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 135_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 132_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 132_m, 0_m, 0_rad}};
  OWL::LaneOverlap overlap6{GlobalRoadPosition{"", 0, 100_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 105_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 102_m, 0_m, 0_rad},
                            GlobalRoadPosition{"", 0, 102_m, 0_m, 0_rad}};
  OWL::Interfaces::LaneAssignments objectsLane2{{overlap4, &moving2}, {overlap5, &moving3}, {overlap6, &moving4}};
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetLength()).WillByDefault(Return(lengthLane1));
  ON_CALL(lane1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane1, GetWorldObjects(false)).WillByDefault(ReturnRef(objectsLane1));
  LaneStreamElement element1{&lane1, lengthLane1, false};
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetLength()).WillByDefault(Return(lengthLane2));
  ON_CALL(lane2, GetWorldObjects(false)).WillByDefault(ReturnRef(objectsLane2));
  ON_CALL(lane2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthLane1));
  LaneStreamElement element2{&lane2, lengthLane1 + lengthLane2, false};

  LaneStream laneStream{{element1, element2}};

  auto result = laneStream.GetAgentsInRange({20_m, 0_m}, {190_m, 0_m});
  ASSERT_THAT(result, ElementsAre(&agent2, &agent3));
}

TEST(LaneStream_Tests, GetStreamDistanceWithAgent_SingleRoadWithThreeSections)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  constexpr units::length::meter_t lengthLane3 = 210_m;

  OWL::Fakes::Road road;
  const std::string idRoad = "RoadB";
  ON_CALL(road, GetId()).WillByDefault(ReturnRef(idRoad));

  OWL::Fakes::Lane lane1;
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthLane1; });
  ON_CALL(lane1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane1, GetSection()).WillByDefault(ReturnRef(section1));
  ON_CALL(lane1, GetRoad()).WillByDefault(ReturnRef(road));
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element1{&lane1, 0_m, true};

  OWL::Fakes::Lane lane2;
  OWL::Fakes::Section section2;
  ON_CALL(section2, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= lengthLane1 && s <= lengthLane1 + lengthLane2; });
  ON_CALL(lane2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthLane1));
  ON_CALL(lane2, GetSection()).WillByDefault(ReturnRef(section2));
  ON_CALL(lane2, GetRoad()).WillByDefault(ReturnRef(road));
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element2{&lane2, lengthLane1, true};

  OWL::Fakes::Lane lane3;
  OWL::Fakes::Section section3;
  ON_CALL(section3, Covers(_))
      .WillByDefault([&](units::length::meter_t s)
                     { return s >= lengthLane1 + lengthLane2 && s <= lengthLane1 + lengthLane2 + lengthLane3; });
  ON_CALL(lane3, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthLane1 + lengthLane2));
  ON_CALL(lane3, GetSection()).WillByDefault(ReturnRef(section3));
  ON_CALL(lane3, GetRoad()).WillByDefault(ReturnRef(road));
  ON_CALL(lane3, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element3{&lane3, lengthLane1 + lengthLane2, true};

  FakeAgent agent;
  GlobalRoadPositions referencePoint
      = {{"RoadA", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -2, 200._m, 1._m, 0.1_rad}}};
  GlobalRoadPositions mainLocatePoint
      = {{"RoadA", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -2, 220._m, -1._m, 0.2_rad}}};
  GlobalRoadPositions rearPoint = {{"RoadA", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -2, 190._m, -1._m, 0.3_rad}}};
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::Reference)))
      .WillByDefault(ReturnRef(referencePoint));
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillByDefault(ReturnRef(mainLocatePoint));
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillByDefault(ReturnRef(rearPoint));
  LaneStream laneStream{{element1, element2, element3}};

  auto referencePointOnStream = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::Reference);
  auto front = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::FrontCenter);
  auto rear = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::RearCenter);

  EXPECT_THAT(referencePointOnStream->s.value(), DoubleEq(200.));
  EXPECT_THAT(referencePointOnStream->t.value(), DoubleEq(1.));
  EXPECT_THAT(referencePointOnStream->hdg.value(), DoubleNear(0.1, CommonHelper::EPSILON));
  EXPECT_THAT(front->s.value(), DoubleEq(220.));
  EXPECT_THAT(front->t.value(), DoubleEq(-1.));
  EXPECT_THAT(front->hdg.value(), DoubleNear(0.2, CommonHelper::EPSILON));
  EXPECT_THAT(rear->s.value(), DoubleEq(190.));
  EXPECT_THAT(rear->t.value(), DoubleEq(-1.));
  EXPECT_THAT(rear->hdg.value(), DoubleNear(0.3, CommonHelper::EPSILON));
}

TEST(LaneStream_Tests, GetStreamDistanceWithAgent_ThreeRoads)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  constexpr units::length::meter_t lengthLane3 = 210_m;

  OWL::Fakes::Road road1;
  const std::string idRoad1 = "RoadA";
  ON_CALL(road1, GetId()).WillByDefault(ReturnRef(idRoad1));
  OWL::Fakes::Road road2;
  const std::string idRoad2 = "RoadB";
  ON_CALL(road2, GetId()).WillByDefault(ReturnRef(idRoad2));
  OWL::Fakes::Road road3;
  const std::string idRoad3 = "RoadC";
  ON_CALL(road3, GetId()).WillByDefault(ReturnRef(idRoad3));

  OWL::Fakes::Lane lane1;
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthLane1; });
  ON_CALL(lane1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane1, GetSection()).WillByDefault(ReturnRef(section1));
  ON_CALL(lane1, GetRoad()).WillByDefault(ReturnRef(road1));
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element1{&lane1, 0_m, true};

  OWL::Fakes::Lane lane2;
  OWL::Fakes::Section section2;
  ON_CALL(section2, Covers(_)).WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthLane2; });
  ON_CALL(lane2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane2, GetSection()).WillByDefault(ReturnRef(section2));
  ON_CALL(lane2, GetRoad()).WillByDefault(ReturnRef(road2));
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element2{&lane2, lengthLane1 + lengthLane2, false};

  OWL::Fakes::Lane lane3;
  OWL::Fakes::Section section3;
  ON_CALL(section3, Covers(_)).WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthLane3; });
  ON_CALL(lane3, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane3, GetSection()).WillByDefault(ReturnRef(section3));
  ON_CALL(lane3, GetRoad()).WillByDefault(ReturnRef(road3));
  ON_CALL(lane3, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element3{&lane3, lengthLane1 + lengthLane2, true};

  FakeAgent agent;
  GlobalRoadPositions referencePoint
      = {{"RoadX", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -2, 10._m, 1._m, 0.1_rad}}};
  GlobalRoadPositions mainLocatePoint
      = {{"RoadX", {}}, {"RoadC", GlobalRoadPosition{"RoadC", -2, 5._m, -1._m, 0.2_rad}}};
  GlobalRoadPositions rearPoint = {{"RoadX", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -2, 0._m, -1._m, 0.3_rad}}};
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::Reference)))
      .WillByDefault(ReturnRef(referencePoint));
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillByDefault(ReturnRef(mainLocatePoint));
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillByDefault(ReturnRef(rearPoint));
  LaneStream laneStream{{element1, element2, element3}};

  auto referencePointOnStream = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::Reference);
  auto front = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::FrontCenter);
  auto rear = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::RearCenter);

  EXPECT_THAT(referencePointOnStream->s.value(), DoubleEq(200.));
  EXPECT_THAT(referencePointOnStream->t.value(), DoubleEq(-1.));
  EXPECT_THAT(referencePointOnStream->hdg.value(), DoubleNear(0.1 - M_PI, CommonHelper::EPSILON));
  EXPECT_THAT(front->s.value(), DoubleEq(215.));
  EXPECT_THAT(front->t.value(), DoubleEq(-1.));
  EXPECT_THAT(front->hdg.value(), DoubleNear(0.2, CommonHelper::EPSILON));
  EXPECT_THAT(rear->s.value(), DoubleEq(210.));
  EXPECT_THAT(rear->t.value(), DoubleEq(1.));
  EXPECT_THAT(rear->hdg.value(), DoubleNear(0.3 - M_PI, CommonHelper::EPSILON));
}

TEST(LaneStream_Tests, GetStreamDistanceWithAgent_AgentOffLane)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  constexpr units::length::meter_t lengthLane3 = 210_m;

  OWL::Fakes::Road road;
  const std::string idRoad = "RoadB";
  ON_CALL(road, GetId()).WillByDefault(ReturnRef(idRoad));

  OWL::Fakes::Lane lane1;
  OWL::Fakes::Section section1;
  ON_CALL(section1, Covers(_)).WillByDefault([&](units::length::meter_t s) { return s >= 0_m && s <= lengthLane1; });
  ON_CALL(lane1, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(0._m));
  ON_CALL(lane1, GetSection()).WillByDefault(ReturnRef(section1));
  ON_CALL(lane1, GetRoad()).WillByDefault(ReturnRef(road));
  ON_CALL(lane1, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element1{&lane1, 0_m, true};

  OWL::Fakes::Lane lane2;
  OWL::Fakes::Section section2;
  ON_CALL(section2, Covers(_))
      .WillByDefault([&](units::length::meter_t s) { return s >= lengthLane1 && s <= lengthLane1 + lengthLane2; });
  ON_CALL(lane2, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthLane1));
  ON_CALL(lane2, GetSection()).WillByDefault(ReturnRef(section2));
  ON_CALL(lane2, GetRoad()).WillByDefault(ReturnRef(road));
  ON_CALL(lane2, GetOdId()).WillByDefault(Return(-2));
  LaneStreamElement element2{&lane2, lengthLane1, true};

  OWL::Fakes::Lane lane3;
  OWL::Fakes::Section section3;
  ON_CALL(section3, Covers(_))
      .WillByDefault([&](units::length::meter_t s)
                     { return s >= lengthLane1 + lengthLane2 && s <= lengthLane1 + lengthLane2 + lengthLane3; });
  ON_CALL(lane3, GetDistance(OWL::MeasurementPoint::RoadStart)).WillByDefault(Return(lengthLane1 + lengthLane2));
  ON_CALL(lane3, GetSection()).WillByDefault(ReturnRef(section3));
  ON_CALL(lane3, GetRoad()).WillByDefault(ReturnRef(road));
  ON_CALL(lane3, GetOdId()).WillByDefault(Return(-3));
  LaneStreamElement element3{&lane3, lengthLane1 + lengthLane2, true};

  FakeAgent agent;
  GlobalRoadPositions referencePoint
      = {{"RoadA", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -3, 200._m, 1._m, 0._rad}}};
  GlobalRoadPositions mainLocatePoint
      = {{"RoadA", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -2, 220._m, -1._m, 0._rad}}};
  GlobalRoadPositions rearPoint = {{"RoadA", {}}, {"RoadB", GlobalRoadPosition{"RoadB", -3, 190._m, -1._m, 0._rad}}};
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::Reference)))
      .WillByDefault(ReturnRef(referencePoint));
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::FrontCenter)))
      .WillByDefault(ReturnRef(mainLocatePoint));
  ON_CALL(agent, GetRoadPosition(VariantWith<ObjectPointPredefined>(ObjectPointPredefined::RearCenter)))
      .WillByDefault(ReturnRef(rearPoint));
  LaneStream laneStream{{element1, element2, element3}};

  auto referencePointOnStream = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::Reference);
  auto front = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::FrontCenter);
  auto rear = laneStream.GetStreamPosition(&agent, ObjectPointPredefined::RearCenter);

  EXPECT_THAT(referencePointOnStream, Eq(std::nullopt));
  EXPECT_THAT(front, Eq(std::nullopt));
  EXPECT_THAT(rear, Eq(std::nullopt));
}

TEST(LaneStream_Tests, GetLaneTypes_AllOfSameType)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  constexpr units::length::meter_t lengthLane3 = 150_m;
  OWL::Interfaces::LaneAssignments emptyObjects{};
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetLength()).WillByDefault(Return(lengthLane1));
  ON_CALL(lane1, GetLaneType()).WillByDefault(Return(LaneType::Driving));
  ON_CALL(lane1, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element1{&lane1, 0_m, true};
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetLength()).WillByDefault(Return(lengthLane2));
  ON_CALL(lane2, GetLaneType()).WillByDefault(Return(LaneType::Driving));
  ON_CALL(lane2, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element2{&lane2, lengthLane1, true};
  OWL::Fakes::Lane lane3;
  ON_CALL(lane3, GetLength()).WillByDefault(Return(lengthLane3));
  ON_CALL(lane3, GetLaneType()).WillByDefault(Return(LaneType::Driving));
  ON_CALL(lane3, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element3{&lane3, lengthLane1 + lengthLane2, true};

  LaneStream laneStream{{element1, element2, element3}};

  auto result = laneStream.GetLaneTypes();

  ASSERT_THAT(result, ElementsAre(std::make_pair(0_m, LaneType::Driving)));
}

TEST(LaneStream_Tests, GetLaneTypes_MixedTypes)
{
  constexpr units::length::meter_t lengthLane1 = 90_m;
  constexpr units::length::meter_t lengthLane2 = 120_m;
  constexpr units::length::meter_t lengthLane3 = 150_m;
  OWL::Interfaces::LaneAssignments emptyObjects{};
  OWL::Fakes::Lane lane1;
  ON_CALL(lane1, GetLength()).WillByDefault(Return(lengthLane1));
  ON_CALL(lane1, GetLaneType()).WillByDefault(Return(LaneType::Driving));
  ON_CALL(lane1, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element1{&lane1, 0_m, true};
  OWL::Fakes::Lane lane2;
  ON_CALL(lane2, GetLength()).WillByDefault(Return(lengthLane2));
  ON_CALL(lane2, GetLaneType()).WillByDefault(Return(LaneType::Driving));
  ON_CALL(lane2, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element2{&lane2, lengthLane1, true};
  OWL::Fakes::Lane lane3;
  ON_CALL(lane3, GetLength()).WillByDefault(Return(lengthLane3));
  ON_CALL(lane3, GetLaneType()).WillByDefault(Return(LaneType::None));
  ON_CALL(lane3, GetWorldObjects(true)).WillByDefault(ReturnRef(emptyObjects));
  LaneStreamElement element3{&lane3, lengthLane1 + lengthLane2, true};

  LaneStream laneStream{{element1, element2, element3}};

  auto result = laneStream.GetLaneTypes();

  ASSERT_THAT(
      result,
      ElementsAre(std::make_pair(0_m, LaneType::Driving), std::make_pair(lengthLane1 + lengthLane2, LaneType::None)));
}
