/********************************************************************************
 * Copyright (c) 2019 in-tech GmbH
 *               2023-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <cstddef>
#include <libxml/tree.h>
#include <stdexcept>
#include <string>
#include <vector>

#include "common/helper/importerHelper.h"
#include "common/sensorDefinitions.h"
#include "include/profilesInterface.h"
#include "profiles.h"
#include "profilesImporter.h"

using namespace Importer;

TEST(ProfilesImporter_ImportAllVehicleComponentsOfSystemProfile, GivenValidXml_ImportsValues)
{
  xmlNodePtr validXml = documentRootFromString(
      "<root>"
      "<Components>"
      "<Component Type=\"ComponentA\">"
      "<Profiles>"
      "<Profile Name=\"No warning\" Probability=\"0.5\"/>"
      "<Profile Name=\"Standard\" Probability=\"0.5\"/>"
      "</Profiles>"
      "<SensorLinks>"
      "<SensorLink SensorId=\"0\" InputId=\"Camera\"/>"
      "<SensorLink SensorId=\"1\" InputId=\"OtherSensor\"/>"
      "<SensorLink SensorId=\"2\" InputId=\"Camera\"/>"
      "</SensorLinks>"
      "</Component>"
      "</Components>"
      "</root>");

  SystemProfile profiles;

  EXPECT_NO_THROW(ProfilesImporter::ImportAllVehicleComponentsOfSystemProfile(validXml, profiles));

  ASSERT_EQ(profiles.vehicleComponents.size(), (size_t)1);

  VehicleComponent resultVehicleComponent = profiles.vehicleComponents.front();

  ASSERT_EQ(resultVehicleComponent.componentProfiles.size(), (size_t)2);
  ASSERT_EQ(resultVehicleComponent.sensorLinks.size(), (size_t)3);

  xmlFreeDoc(validXml->doc);
}

TEST(ProfilesImporter_ImportAllVehicleComponentsOfSystemProfile, WithMissingComponentsTag_Throws)
{
  xmlNodePtr missingComponentsTag = documentRootFromString(
      "<root>"
      "<Component Type=\"ComponentA\">"
      "<Profiles>"
      "<Profile Name=\"No warning\" Probability=\"0.5\"/>"
      "<Profile Name=\"Standard\" Probability=\"0.3\"/>"
      "</Profiles>"
      "<SensorLinks>"
      "<SensorLink SensorId=\"0\" InputId=\"Camera\"/>"
      "<SensorLink SensorId=\"1\" InputId=\"OtherSensor\"/>"
      "<SensorLink SensorId=\"2\" InputId=\"Camera\"/>"
      "</SensorLinks>"
      "</Component>"
      "</root>");

  SystemProfile profiles;
  EXPECT_THROW(ProfilesImporter::ImportAllVehicleComponentsOfSystemProfile(missingComponentsTag, profiles),
               std::runtime_error);

  xmlFreeDoc(missingComponentsTag->doc);
}

TEST(ProfilesImporter_ImportAllVehicleComponentsOfSystemProfile, SumOfProbabilityGreatherOne_Throws)
{
  xmlNodePtr sumOfProbabilityGreatherOne = documentRootFromString(
      "<root>"
      "<Components>"
      "<Component Type=\"ComponentA\">"
      "<Profiles>"
      "<Profile Name=\"No warning\" Probability=\"0.5\"/>"
      "<Profile Name=\"Standard\" Probability=\"0.8\"/>"
      "</Profiles>"
      "<SensorLinks>"
      "<SensorLink SensorId=\"0\" InputId=\"Camera\"/>"
      "<SensorLink SensorId=\"1\" InputId=\"OtherSensor\"/>"
      "<SensorLink SensorId=\"2\" InputId=\"Camera\"/>"
      "</SensorLinks>"
      "</Component>"
      "</Components>"
      "</root>");

  SystemProfile profiles;
  EXPECT_THROW(ProfilesImporter::ImportAllVehicleComponentsOfSystemProfile(sumOfProbabilityGreatherOne, profiles),
               std::runtime_error);

  xmlFreeDoc(sumOfProbabilityGreatherOne->doc);
}

TEST(ProfilesImporter_ImportAllSensorsOfSystemProfile, GivenValidXml_ImportsValues)
{
  xmlNodePtr validXml = documentRootFromString(
      "<root>"
      "<Sensors>"
      "<Sensor Id=\"0\" Position=\"PositionA\">"
      "<Profile Type=\"Geometric2D\" Name=\"Camera\"/>"
      "</Sensor>"
      "<Sensor Id=\"1\" Position=\"PositionB\">"
      "<Profile Type=\"OtherSensor\" Name=\"Camera\"/>"
      "</Sensor>"
      "</Sensors>"
      "</root>");

  SystemProfile profiles;
  EXPECT_NO_THROW(ProfilesImporter::ImportAllSensorsOfSystemProfile(validXml, profiles));

  openpass::sensors::Parameters resultSensors = profiles.sensors;

  ASSERT_EQ(resultSensors.size(), (size_t)2);

  openpass::sensors::Parameter resultSensorParameter = resultSensors.front();
  openpass::sensors::Profile resultSensorProfile = resultSensorParameter.profile;

  ASSERT_EQ(resultSensorParameter.positionName, "PositionA");
  ASSERT_EQ(resultSensorProfile.type, "Geometric2D");
  ASSERT_EQ(resultSensorProfile.name, "Camera");

  xmlFreeDoc(validXml->doc);
}

TEST(ProfilesImporter_ImportAllSensorsOfSystemProfile, SensorsTagMissing_Throws)
{
  xmlNodePtr sensorsTagMissing = documentRootFromString(
      "<root>"
      "<Sensor Id=\"0\" Position=\"PositionA\">"
      "<Profile Type=\"Geometric2D\" Name=\"Camera\"/>"
      "</Sensor>"
      "<Sensor Id=\"1\" Position=\"PositionB\">"
      "<Profile Type=\"OtherSensor\" Name=\"Camera\"/>"
      "</Sensor>"
      "</root>");
  SystemProfile profiles;
  EXPECT_THROW(ProfilesImporter::ImportAllSensorsOfSystemProfile(sensorsTagMissing, profiles), std::runtime_error);

  xmlFreeDoc(sensorsTagMissing->doc);
}

TEST(ProfilesImporter_ImportAllSensorsOfSystemProfile, PositionElementMissing_Throws)
{
  xmlNodePtr positionTagMissing = documentRootFromString(
      "<root>"
      "<Sensors>"
      "<Sensor Id=\"0\">"
      "<Profile Type=\"Geometric2D\" Name=\"Camera\"/>"
      "</Sensor>"
      "</Sensors>"
      "</root>");
  SystemProfile profiles;
  EXPECT_THROW(ProfilesImporter::ImportAllSensorsOfSystemProfile(positionTagMissing, profiles), std::runtime_error);

  xmlFreeDoc(positionTagMissing->doc);
}

TEST(ProfilesImporter_ImportAllSensorsOfSystemProfile, ProfileTagMissing_Throws)
{
  xmlNodePtr profileTagMissing = documentRootFromString(
      "<root>"
      "<Sensors>"
      "<Sensor Id=\"0\" Position=\"PositionA\">"
      "</Sensor>"
      "</Sensors>"
      "</root>");
  SystemProfile profiles;
  EXPECT_THROW(ProfilesImporter::ImportAllSensorsOfSystemProfile(profileTagMissing, profiles), std::runtime_error);

  xmlFreeDoc(profileTagMissing->doc);
}

TEST(ProfilesImporter_ImportAgentProfiles, MissingAtLeastOneElement_Throws)
{
  xmlNodePtr missingAtLeastOneElement = documentRootFromString(
      "<root>"
      "<AgentProfile Name = \"Superhero\">"
      "<DriverProfiles>"
      "<DriverProfile Name = \"Regular\" Probability = \"0.5\"/>"
      "<DriverProfile Name = \"Trajectory\" Probability = \"0.5\"/>"
      "</DriverProfiles>"
      "<ADASProfiles>"
      "<ADASProfile Name = \"ComponentA\" Probability = \"0.5\"/>"
      "</ADASProfiles>"
      "<VehicleModels>"
      "</VehicleModels>"
      "</AgentProfile>"
      "</root>");
  Profiles profiles;
  EXPECT_THROW(ProfilesImporter::ImportAgentProfiles(missingAtLeastOneElement, profiles), std::runtime_error);

  xmlFreeDoc(missingAtLeastOneElement->doc);
}

TEST(ProfilesImporter_ImportAgentProfiles, WrongProbabilities_Throws)
{
  xmlNodePtr wrongProbabilities = documentRootFromString(
      "<root>"
      "<AgentProfile Name = \"Superhero\">"
      "<DriverProfiles>"
      "<DriverProfile Name = \"Regular\" Probability = \"0.5\"/>"
      "<DriverProfile Name = \"Trajectory\" Probability = \"0.5\"/>"
      "</DriverProfiles>"
      "<ADASProfiles>"
      "<ADASProfile Name = \"ComponentA\" Probability = \"0.5\"/>"
      "<ADASProfile Name = \"AEB\" Probability = \"1.0\"/>"
      "</ADASProfiles>"
      "<VehicleModels>"
      "<VehicleModel Name = \"Batmobile\" Probability = \"1.5\"/>"
      "</VehicleModels>"
      "</AgentProfile>"
      "</root>");
  Profiles profiles;
  EXPECT_THROW(ProfilesImporter::ImportAgentProfiles(wrongProbabilities, profiles), std::runtime_error);

  xmlFreeDoc(wrongProbabilities->doc);
}

TEST(ProfilesImporter_ImportAgentProfiles, MissingTag_Throws)
{
  xmlNodePtr missingTag = documentRootFromString(
      "<root>"
      "<AgentProfile Name = \"Superhero\">"
      "<ADASProfiles>"
      "<ADASProfile Name = \"ComponentA\" Probability = \"0.5\"/>"
      "<ADASProfile Name = \"AEB\" Probability = \"1.0\"/>"
      "</ADASProfiles>"
      "<VehicleModels>"
      "<VehicleModel Name = \"Batmobile\" Probability = \"1.0\"/>"
      "</VehicleModels>"
      "</AgentProfile>"
      "</root>");
  Profiles profiles;
  EXPECT_THROW(ProfilesImporter::ImportAgentProfiles(missingTag, profiles), std::runtime_error);

  xmlFreeDoc(missingTag->doc);
}
