/********************************************************************************
 * Copyright (c) 2017-2021 in-tech GmbH
 *               2022-2024 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 ********************************************************************************/

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <optional>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <units.h>
#include <vector>

#include "common/helper/importerHelper.h"
#include "fakeConnection.h"
#include "fakeJunction.h"
#include "fakeOdRoad.h"
#include "include/roadInterface/roadElementTypes.h"
#include "sceneryImporter.h"

using ::testing::_;
using ::testing::Invoke;
using ::testing::NiceMock;
using ::testing::Return;
using ::testing::SizeIs;

using namespace Importer;

TEST(SceneryImporter_UnitTests, ParseRoadTypes)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "    <type s=\"100.1\" type=\"motorway\" />"
      "    <type s=\"100.2\" type=\"town\" />"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  EXPECT_CALL(mockRoad, AddRoadType(_)).Times(2);
  ASSERT_NO_THROW(SceneryImporter::ParseRoadTypes(documentRoot, &mockRoad));

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseJunctionConnections)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<junction>"
      "     <connection id=\"0\" incomingRoad=\"44\" connectingRoad=\"48\" contactPoint=\"start\">"
      "            <laneLink from=\"-5\" to=\"-1\"/>"
      "            <laneLink from=\"-7\" to=\"-2\"/>"
      "            <laneLink from=\"-8\" to=\"-3\"/>"
      "      </connection>"
      "</junction>");

  NiceMock<FakeJunction> mockJunction;
  NiceMock<FakeConnection> mockConnection;

  EXPECT_CALL(mockJunction, AddConnection("0", "44", "48", ContactPointType::Start))
      .Times(1)
      .WillOnce(Return(&mockConnection));
  EXPECT_CALL(mockConnection, AddLink(-5, -1)).Times(1);
  EXPECT_CALL(mockConnection, AddLink(-7, -2)).Times(1);
  EXPECT_CALL(mockConnection, AddLink(-8, -3)).Times(1);

  ASSERT_NO_THROW(SceneryImporter::ParseJunctionConnections(documentRoot, &mockJunction));

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseSignalsWithThreeValidSignals_DetectsThreeSignals)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <signals>"
      "    <signal s=\"100.1\" t=\"1.1\" id=\"id0\" name=\"n0\" dynamic=\"no\"  orientation=\"+\"    zOffset=\"0.1\" "
      "country=\"c1\" type=\"-1\" subtype=\"none\" value=\"0.123\"               height=\"1.01\" width=\"0.11\" "
      "text=\"t1\" hOffset=\"1e-3\" pitch=\"1e-2\"/>"
      "    <signal s=\"200.2\" t=\"2.2\" id=\"id1\" name=\"n1\" dynamic=\"no\"  orientation=\"-\"    zOffset=\"0.2\" "
      "country=\"c2\" type=\"none\" subtype=\"-1\"   value=\"0.456\" unit=\"mph\"  height=\"2.02\" width=\"0.22\" "
      "text=\"t2\" hOffset=\"2e-3\" pitch=\"2e-2\"/>"
      "    <signal s=\"300.3\" t=\"3.3\" id=\"id2\" name=\"n2\" dynamic=\"yes\" orientation=\"none\" zOffset=\"0.3\" "
      "country=\"c3\" type=\"274\" subtype=\"100\" value=\"0.789\" unit=\"km/h\" height=\"3.03\" width=\"0.33\" "
      "text=\"t3\" hOffset=\"3e-3\" pitch=\"3e-2\"/>"
      "  </signals>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  //EXPECT_CALL(mockRoad, AddRoadSignal(_)).Times(3);
  ASSERT_NO_THROW(SceneryImporter::ParseSignals(documentRoot, &mockRoad));

  xmlFreeDoc(documentRoot->doc);
}

struct SignalInterceptor
{
  static RoadSignalSpecification signal;  // NOLINT(cppcoreguidelines-avoid-non-const-global-variables)

  static void intercept(const RoadSignalSpecification& signal) { SignalInterceptor::signal = signal; }
};

RoadSignalSpecification SignalInterceptor::signal;  // NOLINT(cppcoreguidelines-avoid-non-const-global-variables)

TEST(SceneryImporter_UnitTests, ParseSignalsWithValidSignal_ReturnsCorrectValues)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <signals>"
      "    <signal s=\"100.1\" t=\"1.1\" id=\"id0\" name=\"n0\" dynamic=\"no\" orientation=\"-\" zOffset=\"0.1\" "
      "country=\"c1\" type=\"none\" subtype=\"none\" value=\"0.123\"  height=\"1.01\" width=\"0.11\" text=\"t1\" "
      "hOffset=\"1e-3\" pitch=\"1e-2\">"
      "      <validity fromLane=\"-3\" toLane=\"-2\" />"
      "     </signal>"
      "  </signals>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  ON_CALL(mockRoad, AddRoadSignal(_)).WillByDefault(Invoke(SignalInterceptor::intercept));

  ASSERT_NO_THROW(SceneryImporter::ParseSignals(documentRoot, &mockRoad));

  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.s.value(), 100.1);
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.t.value(), 1.1);
  EXPECT_EQ(SignalInterceptor::signal.id, "id0");
  EXPECT_EQ(SignalInterceptor::signal.name, "n0");
  EXPECT_EQ(SignalInterceptor::signal.orientation, "-");
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.zOffset.value(), 0.1);
  EXPECT_EQ(SignalInterceptor::signal.country, "c1");
  EXPECT_EQ(SignalInterceptor::signal.type, "none");
  EXPECT_EQ(SignalInterceptor::signal.subtype,
            "-1");  //accoount for OpenDRIVE standard 1.6, subtype of "none" and "" are parsed to "-1"
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.value.value(), 0.123);
  EXPECT_EQ(SignalInterceptor::signal.unit, RoadSignalUnit::Undefined);
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.height.value(), 1.01);
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.width.value(), 0.11);
  EXPECT_EQ(SignalInterceptor::signal.text, "t1");
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.hOffset.value(), 1e-3);
  EXPECT_DOUBLE_EQ(SignalInterceptor::signal.pitch.value(), 1e-2);
  EXPECT_EQ(SignalInterceptor::signal.validity.lanes[0], -3);
  EXPECT_EQ(SignalInterceptor::signal.validity.lanes[1], -2);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseSignalsWithMissingField_Fails)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <signals>"
      "    <signal s=\"100.1\" t=\"1.1\" id=\"id0\" name=\"n0\" dynamic=\"no\" zOffset=\"0.1\" country=\"c1\" "
      "type=\"none\" subtype=\"none\" value=\"0.123\" height=\"1.01\" width=\"0.11\" text=\"t1\" hOffset=\"1e-3\" "
      "pitch=\"1e-2\"/>"
      "  </signals>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  ASSERT_THROW(SceneryImporter::ParseSignals(documentRoot, &mockRoad), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseSignalsWithInvalidField_Fails)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <signals>"
      "    <signal s=\"-100.1\" t=\"x1\" id=\"id0\" name=\"n0\" dynamic=\"no\" zOffset=\"0.1\" country=\"c1\" "
      "type=\"\" subtype=\"\" value=\"0.123\" height=\"1.01\" width=\"0.11\" text=\"t1\" hOffset=\"1e-3\" "
      "pitch=\"1e-2\"/>"
      "  </signals>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  ASSERT_THROW(SceneryImporter::ParseSignals(documentRoot, &mockRoad), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ValidityWithThreeIncreasingLaneNumbers_DeliversThreeValidityEntries)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <validity fromLane=\"0\" toLane=\"2\" />"
      "</root>");

  RoadElementValidity validitiy;

  ASSERT_NO_THROW(SceneryImporter::ParseElementValidity(documentRoot, validitiy));
  ASSERT_THAT(validitiy.lanes, SizeIs(3));
  EXPECT_FALSE(validitiy.all);
  EXPECT_EQ(validitiy.lanes[0], 0);
  EXPECT_EQ(validitiy.lanes[1], 1);
  EXPECT_EQ(validitiy.lanes[2], 2);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ValidityWithThreeDecreasingLaneNumbers_DeliversThreeValidityEntries)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <validity fromLane=\"0\" toLane=\"-2\" />"
      "</root>");

  RoadElementValidity validitiy;

  ASSERT_NO_THROW(SceneryImporter::ParseElementValidity(documentRoot, validitiy));
  ASSERT_THAT(validitiy.lanes, SizeIs(3));
  EXPECT_FALSE(validitiy.all);
  EXPECT_EQ(validitiy.lanes[0], -2);
  EXPECT_EQ(validitiy.lanes[1], -1);
  EXPECT_EQ(validitiy.lanes[2], 0);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ValidityWithMissingFromLane_Fails)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <validity toLane=\"0\" />"
      "</root>");

  RoadElementValidity validitiy;

  ASSERT_THROW(SceneryImporter::ParseElementValidity(documentRoot, validitiy), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ValidityWithMissingToLane_Fails)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <validity fromLane=\"0\" />"
      "</root>");

  RoadElementValidity validitiy;

  ASSERT_THROW(SceneryImporter::ParseElementValidity(documentRoot, validitiy), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, NoValidity_SetsAllValidityFlag)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "</root>");

  RoadElementValidity validitiy;

  ASSERT_NO_THROW(SceneryImporter::ParseElementValidity(documentRoot, validitiy));
  EXPECT_TRUE(validitiy.all);
  EXPECT_THAT(validitiy.lanes, SizeIs(0));

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObject_ThreeBarriersAreDetected)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b01\" s=\"100\" t=\"0\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"10\" width=\"2\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b02\" s=\"110\" t=\"1\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"11\" width=\"3\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b03\" s=\"120\" t=\"2\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"12\" width=\"4\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  //EXPECT_CALL(mockRoad, AddRoadObject(_)).Times(3);
  ASSERT_NO_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad));

  xmlFreeDoc(documentRoot->doc);
}

struct ObjectInterceptor
{
  static RoadObjectSpecification object;  // NOLINT(cppcoreguidelines-avoid-non-const-global-variables)

  static void intercept(const RoadObjectSpecification& object) { ObjectInterceptor::object = object; }
};

RoadObjectSpecification ObjectInterceptor::object;  // NOLINT(cppcoreguidelines-avoid-non-const-global-variables)

TEST(SceneryImporter_UnitTests, ParseObjectsWithValidObject_ReturnsCorrectValues)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b03\" s=\"120\" t=\"2\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"+\" length=\"12\" width=\"4\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  ON_CALL(mockRoad, AddRoadObject(_)).WillByDefault(Invoke(ObjectInterceptor::intercept));
  ASSERT_NO_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad));

  EXPECT_EQ(ObjectInterceptor::object.type, RoadObjectType::barrier);
  EXPECT_EQ(ObjectInterceptor::object.name, "obstacle");
  EXPECT_EQ(ObjectInterceptor::object.id, "b03");
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.s.value(), 120);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.t.value(), 2);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.zOffset.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.validLength.value(), 0);
  EXPECT_EQ(ObjectInterceptor::object.orientation, RoadElementOrientation::positive);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.width.value(), 4);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.length.value(), 12);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.radius.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.height.value(), 2);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.hdg.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.pitch.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.roll.value(), 0);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObjectsWithMissingField_ThrowsException)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b01\" t=\"0\" zOffset=\"0\" orientation=\"none\" "
      "length=\"10\" width=\"2\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");  // s is missing

  NiceMock<FakeOdRoad> mockRoad;

  ASSERT_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObjectssWithInvalidS_ThrowsException)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b01\" s=\"-100\" t=\"10\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"10\" width=\"2\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");  // s is negative

  NiceMock<FakeOdRoad> mockRoad;

  ASSERT_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObjectssWithInvalidDimensions_ThrowsException)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b01\" s=\"100\" t=\"10\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"-1.0\" width=\"-1.0\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");  // length, width = 0

  NiceMock<FakeOdRoad> mockRoad;

  ASSERT_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad), std::runtime_error);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObjectsWithZeroLength_IsIgnored)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b01\" s=\"100\" t=\"10\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"0.0\" width=\"2.0\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  EXPECT_CALL(mockRoad, AddRoadObject(_)).Times(0);
  ASSERT_NO_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad));

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObjectsWithZeroWidth_IsIgnored)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"barrier\" name=\"obstacle\" id=\"b01\" s=\"100\" t=\"10\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"none\" length=\"2.0\" width=\"0.0\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  EXPECT_CALL(mockRoad, AddRoadObject(_)).Times(0);
  ASSERT_NO_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad));

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseRepeatWithNoOverridingOfOptionalParameters)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <repeat s=\"100\" length=\"200\" distance=\"50\"/>"
      "</root>");

  RoadObjectSpecification object;
  object.type = RoadObjectType::obstacle;
  object.name = "Leitplanke";
  object.id = "";
  object.s = 0_m;
  object.t = 1_m;
  object.zOffset = 0_m;
  object.validLength = 0_m;
  object.orientation = RoadElementOrientation::negative;
  object.length = 2.0_m;
  object.width = 1.0_m;
  object.height = 0_m;
  object.hdg = 0_rad;
  object.pitch = 0_rad;
  object.roll = 0_rad;
  object.radius = 0_m;
  std::vector<RoadObjectSpecification> objectRepitions;

  objectRepitions = SceneryImporter::ParseObjectRepeat(documentRoot, object);

  for (const auto& object : objectRepitions)
  {
    ASSERT_EQ(object.t.value(), 1);
    ASSERT_EQ(object.length.value(), 2.0);
    ASSERT_EQ(object.width.value(), 1.0);
    ASSERT_EQ(object.height.value(), 0);
    ASSERT_EQ(object.zOffset.value(), 0);
    ASSERT_EQ(object.continuous, false);
  }

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseRepeatTestOverridingOfObjectRepeat)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <repeat s=\"100\" length=\"200\" distance=\"50\"/>"
      "</root>");

  RoadObjectSpecification object;
  object.s = 0_m;
  object.radius = 0_m;
  object.length = 2.0_m;
  object.width = 1.0_m;
  std::vector<RoadObjectSpecification> objectRepitions;

  objectRepitions = SceneryImporter::ParseObjectRepeat(documentRoot, object);

  int distance = 50;
  int sStart = 100;
  int count = 0;
  for (const auto& object : objectRepitions)
  {
    ASSERT_EQ(object.s.value(), (sStart + count++ * distance));
  }

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseRepeatOverridesAllOptionalParameters)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <repeat s=\"100\" length=\"200\" distance=\"50\" tStart=\"1.5\" tEnd=\"2\" widthStart=\"2\" widthEnd=\"3\" "
      "heightStart=\"1\" heightEnd=\"2\" zOffsetStart=\"0\" zOffsetEnd=\"0.1\" />"
      "</root>");

  RoadObjectSpecification object;
  object.type = RoadObjectType::obstacle;
  object.name = "Leitplanke";
  object.id = "";
  object.s = 0_m;
  object.t = 1_m;
  object.zOffset = 0.1_m;
  object.validLength = 0_m;
  object.orientation = RoadElementOrientation::negative;
  object.length = 5_m;
  object.width = 5_m;
  object.height = 5_m;
  object.hdg = 0_rad;
  object.pitch = 0_rad;
  object.roll = 0_rad;
  object.radius = 0_m;
  std::vector<RoadObjectSpecification> objectRepitions;

  objectRepitions = SceneryImporter::ParseObjectRepeat(documentRoot, object);

  RoadObjectSpecification firstObject = objectRepitions.front();
  EXPECT_NEAR(firstObject.t.value(), 1.5, 1e-4);
  EXPECT_NEAR(firstObject.s.value(), 100, 1e-4);
  EXPECT_NEAR(firstObject.width.value(), 2, 1e-4);
  EXPECT_NEAR(firstObject.height.value(), 1, 1e-4);
  EXPECT_NEAR(firstObject.zOffset.value(), 0, 1e-4);

  ASSERT_EQ(firstObject.type, RoadObjectType::obstacle);
  ASSERT_EQ(firstObject.length.value(), 5);
  ASSERT_EQ(firstObject.hdg.value(), 0);
  ASSERT_EQ(firstObject.continuous, false);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseRepeatWithDistanceZero)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <repeat s=\"100\" length=\"200\" distance=\"0\"/>"
      "</root>");

  RoadObjectSpecification object;
  object.type = RoadObjectType::obstacle;
  object.name = "Leitplanke";
  object.id = "";
  object.s = 0_m;
  object.t = 1_m;
  object.zOffset = 0_m;
  object.validLength = 0_m;
  object.orientation = RoadElementOrientation::negative;
  object.length = 2.0_m;
  object.width = 1.0_m;
  object.height = 0_m;
  object.hdg = 0_rad;
  object.pitch = 0_rad;
  object.roll = 0_rad;
  object.radius = 0_m;
  std::vector<RoadObjectSpecification> objectRepitions;

  objectRepitions = SceneryImporter::ParseObjectRepeat(documentRoot, object);

  ASSERT_THAT(objectRepitions, SizeIs(1));

  RoadObjectSpecification firstObject = objectRepitions.front();

  ASSERT_EQ(firstObject.type, RoadObjectType::obstacle);
  ASSERT_EQ(firstObject.length.value(), 200);
  ASSERT_EQ(firstObject.continuous, true);

  xmlFreeDoc(documentRoot->doc);
}

TEST(SceneryImporter_UnitTests, ParseObjectsWithRadius_ReturnsCorrectValues)
{
  xmlNodePtr documentRoot = documentRootFromString(
      "<root>"
      "  <objects>"
      "    <object type=\"tree\" name=\"tree\" id=\"b01\" s=\"120\" t=\"2\" zOffset=\"0\" validLength=\"0\" "
      "orientation=\"+\" radius=\"2\" height=\"2\" hdg=\"0\" pitch=\"0\" roll=\"0\" />"
      "  </objects>"
      "</root>");

  NiceMock<FakeOdRoad> mockRoad;

  ON_CALL(mockRoad, AddRoadObject(_)).WillByDefault(Invoke(ObjectInterceptor::intercept));
  ASSERT_NO_THROW(SceneryImporter::ParseObjects(documentRoot, &mockRoad));

  EXPECT_EQ(ObjectInterceptor::object.type, RoadObjectType::tree);
  EXPECT_EQ(ObjectInterceptor::object.name, "tree");
  EXPECT_EQ(ObjectInterceptor::object.id, "b01");
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.s.value(), 120);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.t.value(), 2);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.zOffset.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.validLength.value(), 0);
  EXPECT_EQ(ObjectInterceptor::object.orientation, RoadElementOrientation::positive);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.width.value(), 4);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.length.value(), 4);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.radius.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.height.value(), 2);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.hdg.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.pitch.value(), 0);
  EXPECT_DOUBLE_EQ(ObjectInterceptor::object.roll.value(), 0);

  xmlFreeDoc(documentRoot->doc);
}
